//*****************************************************************************
//*
//*
//*		HttpFolderPlugin.cpp
//*
//*
//*****************************************************************************
//
//	This plugin allows to show the files of a folder in a html side.
//
#include	<string.h>
#include	<stdlib.h>
#include	<stdio.h>

#include	"HttpInternal.h"
#include	"HttpIconPlugin.h"
#include	"HttpFolderPlugin.h"

#define 	ENTRIES(a)		(sizeof(a)/sizeof((a)[0]))
#define 	HTTP_API		extern "C"			
#ifndef 	FP_NAME_COUNT
#define 	FP_NAME_COUNT	512								// maximum count of files on page
#endif 	
#ifndef 	FP_NAME_SIZE
#define 	FP_NAME_SIZE	28								// size of chars in the line
#endif 	
#ifndef 	FP_NAME_MAX
#define 	FP_NAME_MAX		53								// langest file name for the list
#endif 	
#ifndef 	FP_MAX_BLOCKS
#define		FP_MAX_BLOCKS	8								// count of static data blocks
#endif 	
#ifndef 	FP_STATIC_MEM
#define 	FP_STATIC_MEM	1								// use only staic memory 
#endif 	



typedef	struct
	{
	char					cName[FP_NAME_MAX];				// file name 
	unsigned char			ucSuffixPos;					// psition of the file suffix
	unsigned char 			ucNameLen;						// length of the filename
	unsigned char 			ucIsDir;						// is it a directory
	SysUtc 					uTime;							// is the file time
	SysDW	  				uSize;							// size of file
	}FolderEntry;		

typedef	struct
	{
	unsigned				uLocked;						// is the data block in use
	unsigned				uSortMode;						// is the sort mode
	unsigned				uEntryPos;						// current entry
	unsigned				uEntryCount;					// count of entries
	unsigned				uMoreEntries;					// are all entries of the foler in the list
	FolderEntry				aEntries[FP_NAME_COUNT];		// list off entries
	}FolderData;


static  const char		   *pMonthTxt[16] = {"???","Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec","???","???","???"};
static	const char			cHeader1 [  ] = "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2 Final//EN\">\n<html><head><title>Index of ";
static	const char			cHeader2 [  ] = "</title></head>\n<body><h1>&nbsp;Index of ";
static	const char			cHeader3 [  ] = "</h1><pre>    <a href=\"";
static	const char			cHeader4 [  ] = "\">Name</a>     ";    
static	const char			cHeader5 [  ] = ":Y\">Name-Type</a>";    
static	const char			cHeader6 [  ] = ":T\">Time</a>                  <a href=\"";    
static	const char			cHeader7 [  ] = ":S\">Size</a>\n<hr>";    
static	const char			cEntry1  [  ] = "<img src=\"";
static	const char			cEntry2  [  ] = "\"> <a href=\"";
static	const char			cEntry3  [  ] = "\">";
static	const char			cTail1   [  ] = "<hr></pre><address>&nbsp;AzHttp Server at ";
static	const char			cTail2   [  ] = "</address></body></html>\n";
static	const char			cMore    [  ] = "<img src=@F@1.gif> ...more files found...\n";

#if FP_STATIC_MEM
static	SysSemaphore		hBlockSem    = 0;				// semaphor for block access
static	SysSemaphore		hBlockSemNum = 0;				// semaphor for block count
static  FolderData			aBlocks[FP_MAX_BLOCKS];			// memory for static blocks
static  unsigned			uBlockNext;						// next search position for blocks
#endif 	


static	const char		   *pSuffixFilter;					// suffix of all allowed file types	
static	char				cBaseUrl[256];					// root path for the pugin 
static	int					iBaseUrlLen=0;					// length of the root path
static	int					iUserParamId=0;					// id for the user parameter
static	int					iFolderFlags=0;					// flags for the plugin
static	unsigned 			uIgnoreRoot;					// don't show plugin at "/"
static	HttpRequestCB		pOldRequest;					// next callback in chain
static	HttpDisconnectCB	pOldDisconnect;					// next disconnect in chain


//*****************************************************************************
//*
//*		CmpByName
//*
//*****************************************************************************
//	Compare function for name sorting
static int _cdecl CmpByName(const void *pPtr1,const void *pPtr2)
{
const FolderEntry  *pEntry1=(FolderEntry*)pPtr1;	
const FolderEntry  *pEntry2=(FolderEntry*)pPtr2;



	if(pEntry2->ucIsDir!=pEntry1->ucIsDir)
		{
		return pEntry2->ucIsDir-pEntry1->ucIsDir;
		}

return stricmp(pEntry1->cName,pEntry2->cName);					
}

//*****************************************************************************
//*
//*		CmpByType
//*
//*****************************************************************************
//	Compare function for type sorting
static int _cdecl CmpByType(const void *pPtr1,const void *pPtr2)
{
const FolderEntry  *pEntry1=(FolderEntry*)pPtr1;	
const FolderEntry  *pEntry2=(FolderEntry*)pPtr2;
const char		   *pName1;	
const char		   *pName2;
int					iCmp;



	if(pEntry2->ucIsDir!=pEntry1->ucIsDir)
		{
		return pEntry2->ucIsDir-pEntry1->ucIsDir;
		}

	pName1 = pEntry1->cName+pEntry1->ucSuffixPos;	
	pName2 = pEntry2->cName+pEntry2->ucSuffixPos;	
	
	
	   iCmp = stricmp(pName1,pName2);
	if(iCmp)return iCmp;


return stricmp(pEntry1->cName,pEntry2->cName);					
}

//*****************************************************************************
//*
//*		CmpBySize
//*
//*****************************************************************************
//	Compare function for name sorting
static int _cdecl CmpBySize(const void *pPtr1,const void *pPtr2)
{
const FolderEntry  *pEntry1=(FolderEntry*)pPtr1;	
const FolderEntry  *pEntry2=(FolderEntry*)pPtr2;



	if(pEntry2->ucIsDir!=pEntry1->ucIsDir)
		{
		return pEntry2->ucIsDir-pEntry1->ucIsDir;
		}

	if(pEntry2->uSize!=pEntry1->uSize)
		{
		return (pEntry2->uSize>pEntry1->uSize)? -1:1;
		}

return stricmp(pEntry1->cName,pEntry2->cName);					
}

//*****************************************************************************
//*
//*		CmpByTime
//*
//*****************************************************************************
//	Compare function for name sorting
static int _cdecl CmpByTime(const void *pPtr1,const void *pPtr2)
{
const FolderEntry  *pEntry1=(FolderEntry*)pPtr1;	
const FolderEntry  *pEntry2=(FolderEntry*)pPtr2;



	if(pEntry2->ucIsDir!=pEntry1->ucIsDir)
		{
		return pEntry2->ucIsDir-pEntry1->ucIsDir;
		}

	if(pEntry2->uTime!=pEntry1->uTime)
		{
		return (pEntry2->uTime>pEntry1->uTime)? -1:1;
		}

return stricmp(pEntry1->cName,pEntry2->cName);					
}

//*****************************************************************************
//*
//*		CmpFileSuffix
//*
//*****************************************************************************
//	Compares a file suffix with a type list
//	pSuffix		: pointer to the suffix
//	pType		: pointer to the type list (i.e. "exe;txt;cpp" )
//	iNum		: size of the suffix
//	Returns 1 if a type is equal
static int CmpFileSuffix(const char *pSuffix,const char *pType,int iNum)
{
int		iLen;



	iNum = strlen(pSuffix);

	for(;;)
		{
		for(iLen=0;;iLen++)								// get length
			{
			if(pType[iLen]== 0 )break;
			if(pType[iLen]==';')break;
			}

		if(iLen==iNum && !strnicmp(pSuffix,pType,iNum))
			{
			return 1;
			}

			pType+=iLen;								// next type
		if(*pType==0)break;

		pType++;
		}


return 0;
}

//*****************************************************************************
//*
//*		AllocFolderList
//*
//*****************************************************************************
//	Allocates a new folder list
//	Returns the pointer to the list, or 0 if out of memory
static FolderData *AllocFolderList()
{
FolderData	*pList;

	#if FP_STATIC_MEM

	int		iCount;
	
	if(!SysSemaphoreLockWait(hBlockSemNum,3000))
		{
		return 0;
		}
	
	SysSemaphoreLock(hBlockSem);

	pList = 0;
	
	for(iCount=0;iCount<FP_MAX_BLOCKS;iCount++,uBlockNext++)
		{
		if(uBlockNext>=FP_MAX_BLOCKS)uBlockNext=0;
		if(aBlocks[uBlockNext].uLocked)continue;

		pList= aBlocks+uBlockNext;
		pList->uLocked     = 1;

		break;		
		}

	SysSemaphoreUnlock(hBlockSem);

	#else

	pList = new FolderData;	

	#endif 	


return pList;
}

//*****************************************************************************
//*
//*		DeallocFolderList
//*
//*****************************************************************************
//	Deallocates a folder list
//	pList	: is the pointer to the list
static void DeallocFolderList(FolderData *pList)
{

	if(!pList)return;


	#if FP_STATIC_MEM

	pList->uLocked     = 0;
	SysSemaphoreUnlock(hBlockSemNum);

	#else

	delete pList;	

	#endif 	

}

//*****************************************************************************
//*
//*		HttpFolderData
//*
//*****************************************************************************
//	Returns the html data of the folder
//	pBuffer		: pointer to the tranmit buffer
//	iSize		: is the size of the buffer (0=means dissconnect)
//	iOffset		: is the offest in the response
//	Returns the size of the stored data or an error number below zero
static int _cdecl HttpFolderData(HttpConnection hConn,char *pBuffer,int iSize,int iOffset) 
{
char			cTemp[12];
FolderEntry	   *pEntry;	
FolderData	   *pList;	
const char	   *pUrl;
const char	   *pHost;
const char	   *pType;
SysTimeStruct	sTime;
int				iLen;
int				iNum;
int				iMax;
int				iCnt;
int				iAdd;



	pList = (FolderData*)HttpGetUserPointer(hConn,iUserParamId);
	pUrl  = HttpGetPath(hConn);
	iMax  = strlen(pUrl);


	if(!pList)												// all data transmitted
		{
		return 0;
		}


	if(iOffset==0)											// beginn of the file
		{
		if(iSize<256*2+sizeof(cHeader1)+sizeof(cHeader2)+sizeof(cHeader3))
			{
			if(pList)
				{
				HttpSetUserPointer(hConn,iUserParamId,0);
				DeallocFolderList(pList);
				}

			return -1000;
			}

				memcpy(pBuffer     ,cHeader1,sizeof(cHeader1)-1);	
		iLen  = sizeof(cHeader1)-1;
		iLen += HttpEncodeString(pUrl,pBuffer+iLen,iSize-iLen,iMax);
		        memcpy(pBuffer+iLen,cHeader2,sizeof(cHeader2)-1);	
		iLen += sizeof(cHeader2)-1;
		iLen += HttpEncodeHtml(pUrl,pBuffer+iLen,iSize-iLen,iMax);
		        memcpy(pBuffer+iLen,cHeader3,sizeof(cHeader3)-1);	
		iLen += sizeof(cHeader3)-1;
		iLen += HttpEncodeString(pUrl,pBuffer+iLen,iSize-iLen,iMax);

		if(pList->uSortMode==0 && pList->uSortMode!='Y')
			{
			memcpy(pBuffer+iLen,cHeader5,sizeof(cHeader5)-1);	
			iLen += sizeof(cHeader5)-1;
			}
		else{
			memcpy(pBuffer+iLen,cHeader4,sizeof(cHeader4)-1);	
			iLen += sizeof(cHeader4)-1;
			}
		
		iAdd  = FP_NAME_SIZE-5;
		memset(pBuffer+iLen,' ',iAdd);
		iLen += iAdd;
		memcpy(pBuffer+iLen,"<a href=\"",9);
		iLen += 9;

		iLen += HttpEncodeString(pUrl,pBuffer+iLen,iSize-iLen,iMax);
		        memcpy(pBuffer+iLen,cHeader6,sizeof(cHeader6)-1);	
		iLen += sizeof(cHeader6)-1;
		iLen += HttpEncodeString(pUrl,pBuffer+iLen,iSize-iLen,iMax);
		        memcpy(pBuffer+iLen,cHeader7,sizeof(cHeader7)-1);	
		iLen += sizeof(cHeader7)-1;
		}
	else{
		iLen  = 0;
		}


	while(iLen+1024<iSize)									// output files
		{
		if(pList->uEntryPos>=pList->uEntryCount)
			{
			if(pList->uMoreEntries)
				{
				memcpy(pBuffer+iLen,cMore,sizeof(cMore)-1);	
				iLen+=sizeof(cMore)-1;
				}

			HttpSetUserPointer(hConn,iUserParamId,0);
			DeallocFolderList(pList);

			HttpSetUserPointer(hConn,iUserParamId,0);
			pHost = HttpGetHost(hConn);
			memcpy(pBuffer+iLen,cTail1,sizeof(cTail1)-1);	
			iLen += sizeof(cTail1)-1;
			iLen += HttpEncodeString(pHost,pBuffer+iLen,iSize-iLen,strlen(pHost));
			memcpy(pBuffer+iLen,cTail2,sizeof(cTail2)-1);	
			iLen += sizeof(cTail2)-1;

			return iLen;
			}

		
		pEntry  = pList->aEntries;
		pEntry += pList->uEntryPos;
		pType   = pEntry->cName;
		
		for(iNum=pEntry->ucNameLen;iNum>0;iNum--)			// find suffix
			{
			if(pType[iNum-1]!='.')continue;
			
			pType +=iNum;
			iNum   = pEntry->ucNameLen-iNum;

			break;
			}

		if(iFolderFlags&HTTP_FOLDER_ADDSPACE)
			{
			pBuffer[iLen] = ' ';
			iLen++;
			}

		        memcpy(pBuffer+iLen,cEntry1,sizeof(cEntry1)-1);	
		iLen += sizeof(cEntry1)-1;
		iLen += HttpGetIconPath(hConn,pType,pBuffer+iLen,iSize-iLen,pEntry->ucIsDir);
		        memcpy(pBuffer+iLen,cEntry2,sizeof(cEntry2)-1);	
		iLen += sizeof(cEntry2)-1;
		iLen += HttpEncodeString(pUrl,pBuffer+iLen,iSize-iLen,iMax);
				
		if(iMax>0 && pUrl[iMax-1]!='/')
			{
			pBuffer[iLen]='/';
			iLen++;
			}


		iCnt  = pEntry->ucNameLen;
		iLen += HttpEncodeHtml(pEntry->cName,pBuffer+iLen,iSize-iLen,iCnt);
		        memcpy(pBuffer+iLen,cEntry3,sizeof(cEntry3)-1);	
		iLen += sizeof(cEntry3)-1;


		if(iCnt>FP_NAME_SIZE)								// set filename
			{
			memcpy(pEntry->cName+FP_NAME_SIZE-3,"...",4);
			iCnt=FP_NAME_SIZE;
			}
		else{
			memset(pEntry->cName+iCnt,' ',FP_NAME_SIZE-iCnt);
			}


		SysTimeToStruct(&sTime,pEntry->uTime,1);
		iLen += HttpEncodeHtml(pEntry->cName,pBuffer+iLen,iSize-iLen,iCnt);
		memcpy(pBuffer+iLen,"</a>",4);
		iLen += 4;	

		if(iCnt<FP_NAME_SIZE)
			{
			memset(pBuffer+iLen,' ',FP_NAME_SIZE-iCnt);
			iLen += FP_NAME_SIZE-iCnt;
			}
		
	
		if(pEntry->ucIsDir==0)								// add size
			{
			iLen += sprintf(pBuffer+iLen,"   %04u-%s-%02u %02u:%02u ",sTime.usYear,pMonthTxt[sTime.ucMonth&0x0F],sTime.ucDay,sTime.ucMinute,sTime.ucMinute);	

			sprintf(cTemp,"%10u",pEntry->uSize);
			pBuffer[iLen] =  cTemp[0];	                iLen++;
			pBuffer[iLen] = (cTemp[0]==' ')? ' ':'.';	iLen++;
			pBuffer[iLen] =  cTemp[1];	                iLen++;
			pBuffer[iLen] =  cTemp[2];	                iLen++;
			pBuffer[iLen] =  cTemp[3];	                iLen++;
			pBuffer[iLen] = (cTemp[3]==' ')? ' ':'.';	iLen++;
			pBuffer[iLen] =  cTemp[4];	                iLen++;
			pBuffer[iLen] =  cTemp[5];	                iLen++;
			pBuffer[iLen] =  cTemp[6];	                iLen++;
			pBuffer[iLen] = (cTemp[6]==' ')? ' ':'.';	iLen++;
			pBuffer[iLen] =  cTemp[7];	                iLen++;
			pBuffer[iLen] =  cTemp[8];	                iLen++;
			pBuffer[iLen] =  cTemp[9];	                iLen++;
			}

		
		pBuffer[iLen] =  '\n';
		iLen++;

		pList->uEntryPos++;
		}



return iLen;
}

//*****************************************************************************
//*
//*		HttpFolderTrans
//*
//*****************************************************************************
//	Callback for a http request
//	hConn		: handle of the http connection	
//	pPath		: is the folder path
//	pParams		: is the parameter of the request 
//	iRoot		: is the path a root path
//	Returns a http error (>=100) or HTTP_RESP_???/HTTP_POST_??? or 999 for an
//	internal error
HTTP_API int HttpFolderTrans(HttpConnection hConn,const char *pPath,const char *pParams,int iRoot)
{
HttpData	   *pData=(HttpData*)hConn;
FolderData	   *pList;	
FolderEntry	   *pEntry;	
char			cTemp[260];
SysFindStruct	sData;
SysFind			hFind;
int				iLen;
int				iNum;



	SysThreadCheck();
	SysWatchdogTrigger(pData->hWatchdog);


	   iLen = strlen(pPath);
	if(iLen>sizeof(cTemp)-4)
		{
		return 999;
		}
	
	memcpy(cTemp,pPath,iLen);
	if(iLen>0 && cTemp[iLen-1]!='/' && cTemp[iLen-1]!='\\')
		{
		cTemp[iLen]='/';
		iLen++;
		}

	cTemp[iLen  ]='*';
	cTemp[iLen+1]= 0;


		hFind = SysFindFirst(cTemp,&sData);					// finds "."
	if(!hFind)
		{
		return 999;
		}

		pList = AllocFolderList();
	if(!pList)
		{
		SysFindClose(hFind);
		return 503;											// out of resources
		}


	pEntry				= pList->aEntries;
	pList->uSortMode	= *pParams;
	pList->uMoreEntries = 0;
	pList->uEntryCount  = 0;
	pList->uEntryPos	= 0;
	
	for(;;)													// read directory
		{
		SysThreadCheck();
		SysWatchdogTrigger(pData->hWatchdog);

		iLen = strlen(sData.cName);

		if(iLen<FP_NAME_MAX)
		if(iLen!=1 || sData.cName[0]!='.')
			{
			for(iNum=iLen;;iNum--)							// find suffix
				{
				if(iNum<=0){iNum=iLen;break;}
				if(sData.cName[iNum-1]=='.')break;
				}

			pEntry->ucSuffixPos	=  iNum;
			pEntry->ucNameLen	=  iLen;
			pEntry->uSize		=  sData.dwSize;
			pEntry->uTime		=  sData.dwTimeWrite;
			pEntry->ucIsDir		= (sData.dwAttr&SYS_DIR)? 1:0;
			memcpy(pEntry->cName,sData.cName,iLen+1);

			if((iRoot && iFolderFlags&HTTP_FOLDER_INGNOREROOT ) &&
			   (pEntry->ucIsDir       && pEntry->ucNameLen==2 )	&&
			   (pEntry->cName[0]=='.' && pEntry->cName[1]=='.'))
				{
				iNum = 0;
				}
			else if(pEntry->ucIsDir)
				{
				pList->uEntryCount++;
				pEntry++;
				}
			else if(pSuffixFilter==0)
				{
				pList->uEntryCount++;
				pEntry++;
				}
			else if(*pSuffixFilter!='*' &&  CmpFileSuffix(sData.cName+iNum,pSuffixFilter  ,iLen-iNum))
				{
				pList->uEntryCount++;
				pEntry++;
				}
			else if(*pSuffixFilter=='*' && !CmpFileSuffix(sData.cName+iNum,pSuffixFilter+1,iLen-iNum))
				{
				pList->uEntryCount++;
				pEntry++;
				}
			}

		if(SysFindNext(hFind,&sData)<0)
			{
			break;
			}
		
		if(pList->uEntryPos>=FP_NAME_COUNT)
			{
			pList->uMoreEntries = 1;
			break;
			}
		}

	

	switch(*pParams)										// sort all entries
		{
	case 'Y':	qsort(pList->aEntries,pList->uEntryCount,sizeof(pList->aEntries[0]),CmpByType);	break;
	case 'T':	qsort(pList->aEntries,pList->uEntryCount,sizeof(pList->aEntries[0]),CmpByTime);	break;
	case 'S':	qsort(pList->aEntries,pList->uEntryCount,sizeof(pList->aEntries[0]),CmpBySize);	break;
	default:	qsort(pList->aEntries,pList->uEntryCount,sizeof(pList->aEntries[0]),CmpByName);	break;
		}


	SysFindClose(hFind);	
	HttpSetUserPointer(hConn,iUserParamId,pList);
	HttpSetCacheMode  (hConn,HTTP_CACHE_NOCACHE);


return HttpTransmitCallback(hConn,HttpFolderData);
}

//*****************************************************************************
//*
//*		HttpFolderRequest
//*
//*****************************************************************************
//	Callback for a http request
//	hConn		: handle of the http connection	
//	iCommand	: is the type of the request command (HTTP_CMD_???)
//	pHost		: is the domain in the request (i.e. "www.myaddress.com")
//	pUrl		: is the url of the request (i.e. "/folder/index.html")
//	pParams		: is the parameter of the request 
//	iUrlLen		: is the length of the domain
//	iParamLen	: is the length of the parameter
//	Retruns a http error (>=100) or HTTP_RESP_???/HTTP_POST_??? 
static int _cdecl HttpFolderRequest(HttpConnection hConn,int iCommand,const char *pHost,const char *pUrl,const char *pParams,int iUrlLen,int iParamLen)
{
const char     *pPath;
FolderData	   *pList;	
int				iAttr;
int				iRoot;
int				iRet;



	pList = (FolderData*)HttpGetUserPointer(hConn,iUserParamId);

	if(pList)												// close old handle
		{
		HttpSetUserPointer(hConn,iUserParamId,0);
		DeallocFolderList(pList);
		}

	if(iCommand!=HTTP_CMD_GET )								// check http command
	if(iCommand!=HTTP_CMD_HEAD)
		{
		goto NextPlugin;
		}

	iRoot = 0;
	
	if(iBaseUrlLen)											// check url of path
		{
		if(strnicmp(pUrl,cBaseUrl,iBaseUrlLen))
			{
			goto NextPlugin;
			}
		
		if(iUrlLen==iBaseUrlLen)
			{
			iRoot = 1;
			}
		}
	else{
		if(iUrlLen== 1 && pUrl[0]=='/')
			{
			iRoot = 1;
			}
		}

	if(uIgnoreRoot && pUrl[1]==0)							// ignore root path
		{
		goto NextPlugin;
		}

	pPath = HttpGetFile(hConn);
	iAttr = SysGetAttributes(pPath);

	if(iAttr<0 || !(iAttr&SYS_DIR))							// is it a folder
		{
		goto NextPlugin;
		}	
	
	
	   iRet = HttpFolderTrans(hConn,pPath,pParams,iRoot);
	if(iRet!=999)return iRet;


NextPlugin:


	if(pOldRequest)											// is an other plugin active	
		{
		return pOldRequest(hConn,iCommand,pHost,pUrl,pParams,iUrlLen,iParamLen);
		}


return 0;	
} 
 

//*****************************************************************************
//*
//*		HttpFolderDisconnect
//*
//*****************************************************************************
//	Callback for dissconnecting 
//	hConn		: handle of the http connection	
//	pIpAddr		: is the ip address of the client
//	usPort		: is the tcp port of the connection
//	iReason		: is the dissconnect reason (see HTTP_MSG_???)
static void _cdecl HttpFolderDisconnect(HttpConnection hConn,const char *pIpAddr,unsigned short usPort,int iReason) 
{
FolderData	   *pList;	


	   pList = (FolderData*)HttpGetUserPointer(hConn,iUserParamId);
	if(pList)
		{
		HttpSetUserPointer(hConn,iUserParamId,0);
		DeallocFolderList(pList);
		}

	if(pOldDisconnect)
		{
		pOldDisconnect(hConn,pIpAddr,usPort,iReason);
		}

}


//*****************************************************************************
//*
//*		HttpRegisterFolder
//*
//*****************************************************************************
//	Registers the folder plugin. This function must called directly 
//	before HttpDaenon(...) is called.
//	pConfig		: config data of the server	
//	pBaseUrl	: is the base url for which the plugin is used
//					0 = means every folder
//	uFlags		: several flags for the plugin				
//					HTTP_FOLDER_INGNOREROOT	: no folder view for the root path 				
//					HTTP_FOLDER_ADDSPACE	: add a space at the begin of a entry		
//	pTypeList	: is a list of viable file types
//	pTypeFilter	: list of suffixes of all allowed or not allowed file types	
//					"html;htm;txt"	--> show       "*.html" "*.htm" and "*.txt"
//					"*html;htm;txt"	--> don't show "*.html" "*.htm" and "*.txt"
//	Returns 0 if the plugin was installed or an error code
//					1 = no more user parameters
//					2 = base url is to long
HTTP_API int HttpRegisterFolder(HttpConfigData *pConfig,const char *pBaseUrl,unsigned uFlags,const char *pTypeFilter)  
{

	
		iUserParamId = HttpReserveUserParams(pConfig,1);
	if(!iUserParamId)return 1;

	#if FP_STATIC_MEM
	if(!hBlockSem   )hBlockSem   =SysSemaphoreCreate();
	if(!hBlockSemNum)hBlockSemNum=SysSemaphoreCreateEx(FP_MAX_BLOCKS);
	#endif 	
	
	if(pBaseUrl)
		{
		iBaseUrlLen = strlen(pBaseUrl);
		if(iBaseUrlLen>=sizeof(cBaseUrl)-3)return 2;
		memcpy(cBaseUrl,pBaseUrl,iBaseUrlLen);
		
		if(iBaseUrlLen>0 && cBaseUrl[iBaseUrlLen-1]!='/')
			{
			cBaseUrl[iBaseUrlLen] = '/';
			iBaseUrlLen++;
			}
		}
	else{
		iBaseUrlLen = 0;
		}


	iFolderFlags			= uFlags;						// flags for the plugin
	pSuffixFilter			= pTypeFilter;					// suffix of all allowed file types	
	pOldRequest				= pConfig->pCbRequest;			// next callback in chain
	pOldDisconnect			= pConfig->pCbDisconnect;		// next disconnect in chain

	pConfig->pCbRequest		= HttpFolderRequest;	
	pConfig->pCbDisconnect	= HttpFolderDisconnect;

	

return 0;
}
 

//*****************************************************************************
//*
//*		HttpInitFolder
//*
//*****************************************************************************
//	Initalizes the folder plugin for external use with HttpFolderTrans(...). 
//	pBaseUrl	: is the base url for which the plugin is used
//					0 = means every folder
//	uFlags		: several flags for the plugin				
//					HTTP_FOLDER_INGNOREROOT	: no folder view for the root path 				
//					HTTP_FOLDER_ADDSPACE	: add a space at the begin of a entry		
//	pTypeList	: is a list of viable file types
//	pTypeFilter	: list of suffixes of all allowed or not allowed file types	
//					"html;htm;txt"	--> show       "*.html" "*.htm" and "*.txt"
//					"*html;htm;txt"	--> don't show "*.html" "*.htm" and "*.txt"
//	iUserId		: user id for internal use
//	Returns 0 if the plugin was installed or an error code
//					1 = wrong user id
//					2 = base url is to long
HTTP_API int HttpInitFolder(const char *pBaseUrl,unsigned uFlags,const char *pTypeFilter,int iUserId)  
{


	if((unsigned)iUserId>=HTTP_USER_PARAMS)return 1;

	iUserParamId = iUserId;

	#if FP_STATIC_MEM
	if(!hBlockSem   )hBlockSem   =SysSemaphoreCreate();
	if(!hBlockSemNum)hBlockSemNum=SysSemaphoreCreateEx(FP_MAX_BLOCKS);
	#endif 	


	if(pBaseUrl)
		{
		iBaseUrlLen = strlen(pBaseUrl);
		if(iBaseUrlLen>=sizeof(cBaseUrl)-3)return 2;
		memcpy(cBaseUrl,pBaseUrl,iBaseUrlLen);
		
		if(iBaseUrlLen>0 && cBaseUrl[iBaseUrlLen-1]!='/')
			{
			cBaseUrl[iBaseUrlLen] = '/';
			iBaseUrlLen++;
			}
		}
	else{
		iBaseUrlLen = 0;
		}



	iFolderFlags			= uFlags;						// flags for the plugin
	pSuffixFilter			= pTypeFilter;					// suffix of all allowed file types	
	pOldRequest				= 0;							// next callback in chain
	pOldDisconnect			= 0;							// next disconnect in chain

	

return 0;
}

