//*****************************************************************************
//*
//*
//*		HttpInternal.h
//*
//*
//*****************************************************************************
#ifndef 	__HTTPINTERNAL_H__
#define 	__HTTPINTERNAL_H__

#include	"System/SocketUtility.h"
#include	"System/System.h"
#include	"HttpServer.h"

#if        _MSC_VER >= 1000
#pragma 	warning( disable :4996 )
#endif

#define 	HTTP_PRINT				0						// print in and output data
#ifndef 	HTTP_USER_PARAMS
#define 	HTTP_USER_PARAMS		16						// count of user parameter
#endif
#ifndef 	HTTP_STATIC_MEMORY
#define 	HTTP_STATIC_MEMORY		0						// use the http server only static memory
#endif
#ifndef 	HTTP_STATIC_COUNT
#define 	HTTP_STATIC_COUNT		16						// cont of static memory
#endif
#ifndef 	HTTP_HAVE_AUTH
#define 	HTTP_HAVE_AUTH			0						// supports the http server logon
#endif
#define 	HTTP_INPUT_SIZE			2048					// size of the input buffer
#define 	HTTP_OUTPUT_SIZE		16384					// size of the input buffer
#define 	HTTP_MAX_PASSWD			28						// size of maximum username
#define 	HTTP_MAX_USER			28						// size of maximum username
#define 	HTTP_MAX_PATH			256						// size of maximum request paths
#define 	HTTP_MAX_PARAM			256						// size of maximum request parameters
#define 	HTTP_MAX_COOKIE			512						// size of maximum cookie data
#define 	HTTP_MAX_DOMAIN			64						// size of maximum domain
#define 	HTTP_MAX_BROWSER		256						// size of maximum browser name
#define 	HTTP_USER_CACHE			16						// maximum count of user cache
#define 	HTTP_REQU_TIMEOUT		8000					// timeout for incomming data
#define 	HTTP_RECV_TIMEOUT		300000					// timeout for next request
#define 	HTTP_STACK_SIZE			0x10000					// stack size for conection thread
#define 	HTTP_MEMORY_MAX			0x40000					// maximum memory size for the http server
#define 	HTTP_WDOG_TIME			30000					// watchdog timeout for the threads
#define 	HTTP_SERVER_STRING		"AzHttp/1.05"			// is the version string of the server


#if			HTTP_PRINT
#define 	HTTP_LOG_REQU(a)		printf a;
#define 	HTTP_LOG_RESP(a)		printf a;
#define 	HTTP_LOG_CONN(a)		printf a;
#define 	HTTP_LOG_AUTH(a)		printf a;
#else
#define 	HTTP_LOG_REQU(a)
#define 	HTTP_LOG_RESP(a)
#define 	HTTP_LOG_CONN(a)
#define 	HTTP_LOG_AUTH(a)
#endif

#ifdef		linux
#ifndef 	stricmp
#define 	stricmp			strcasecmp
#endif
#ifndef 	strnicmp
#define 	strnicmp		strncasecmp
#endif
#endif


typedef struct
	{
	SysDW				dwLastAccess;						// last access time of the user
	unsigned			uIndexOffset;						// entry possition in the index
	char				cUserName [HTTP_MAX_USER  ];		// is the user name
	char				cPassWord [HTTP_MAX_PASSWD];		// is the password off the user
	char				cAuthNonce[32];						// is the nonce value for the authorization
	}HttpUserCache;


typedef struct
	{
	TcpHandle			hServer;							// socket handle for server
	SysSemaphore		hSemaphore;							// semapore for the server
    unsigned short 		usTcpPort;							// is the tcp port for the server
	unsigned			uUserOffset;						// start offest for user parameters
	int					iDefaultExtLen;						// length of the default file name
	char				cDefaultExt[ 32];					// default file name for access to paths ("/index.html")
	char				cFilePath  [256];					// path for server files with '/' at the end
	int					iFilePathLen;						// length of the path in chars
	int					iMaxConnections;					// maximum count of client connections
	int	volatile		iCurConnections;					// current count of client connections
	int volatile		iShutdown;							// shutdown the server
	int volatile		iRunning;							// is the server running
	int					iPriority;							// taks priority 0-256
	HttpRequestCB		pCbRequest;							// callback for a client request
	HttpConnectCB   	pCbConnect;							// callback for client connection
	HttpPasswordCB		pCbPassword;						// callback for password returning
	HttpDisconnectCB	pCbDisconnect;						// callback for client disconnection

	#if HTTP_HAVE_AUTH
	char				cAuthOpaque[ 32];					// is the opaque value for the authorization
	char				cAuthRealm [ 64];					// is the realm value for the authorization
	unsigned			uUserNext;							// next entry in the user cache to insert
	unsigned			uUserIndexSize;						// count of index entries
	unsigned			aUserIndex [HTTP_USER_CACHE];		// is the user cache index
	HttpUserCache		aUserCache [HTTP_USER_CACHE];		// is the user cache
	#endif

	}HttpServerData;


typedef struct
	{
	HttpServerData	   *pServer;							// pointer to the server data
	TcpHandle			hConnection;						// socket handle for input
	SysFile				hFile;								// is the current file
	SysWdog				hWatchdog;							// wachdog handel fr connection thread
	char				cConnIp[64];						// ip addres of the client
	unsigned short 		usConnPort;							// tcp port of the client

	unsigned			uUserOffset;						// start offest for user parameters
	void			   *pUserParams [HTTP_USER_PARAMS];		// serveral user parameters
	char				cRecvBuffer [HTTP_INPUT_SIZE ];		// input buffer
	char				cSendBuffer [HTTP_OUTPUT_SIZE];		// output puffer
	char				cAuthUser   [HTTP_MAX_USER   ];		// is the username for the authorization
	unsigned			uSendAlloc;							// allocated size in the send buffer
	unsigned			uRecvSize;							// bytes in input buffer
	unsigned			uRecvTimeout;						// timeout for data receiving
	unsigned			uRequTimeout;						// timeout for incomming request data
	unsigned			uTransmitActiv;						// is currently a request handeld

	unsigned			uRequLast;							// time of last request
	unsigned			uRequCommand;						// is the command of the request
	unsigned			uRequKeepAlive;						// don't close connection after request
	unsigned			uRequContentLen;					// length off the content
	unsigned			uRequPathLength;					// length off the request path
	unsigned			uRequParamLength;					// length off the request param
	unsigned			uRequHttpVersion;					// is the http version 0xXXYY
	char				cRequFile   [HTTP_MAX_PATH   ];		// path oft used file request
	char				cRequPath   [HTTP_MAX_PATH   ];		// path oft the request
	char				cRequParam  [HTTP_MAX_PARAM  ];		// parameter of the request
	char				cRequCookie [HTTP_MAX_COOKIE ];		// cookies of the request
	char				cRequBrowser[HTTP_MAX_BROWSER];		// name of the browser
	char				cRequHost   [64];					// the name off the request
	char			   *pRequType;							// ponter to the filetype in the url

	HttpDataCB			pResponseCb;						// callback for response
	int					iResponseLen;						// length of the response (-1=chunked)
	SysDW				uResponseTime;						// time for the response (0=none)
	unsigned 			uResponseCache;						// cache settings (see. HTTP_CACHE_???)
	char				cResponseType[64];					// content-type for response (""=none)
	char				cResponseCookie[HTTP_MAX_COOKIE ];	// add cookies for the reqonse

	int					iMemoryMax;							// is the maximum size of the memory
	int					iMemoryPos;							// is the position in the memory
	char			   *pMemoryPtr;							// pointer for the response memory

	#if HTTP_HAVE_AUTH
	unsigned			uIgnoreLogin;						// ignores all password checks login
	char				cAuthNonce  [32];					// is the nonce value for the authorization
	#endif
	}HttpData;


//*****************************************************************************
//*
//*		MallocFunctions
//*
//*****************************************************************************

extern        void *hHttpHeap;
static inline void *HttpAlloc(unsigned uSize){return SysHeapAlloc(hHttpHeap,uSize);}
static inline void  HttpFree (void    *pPtr ){       SysHeapFree (hHttpHeap,pPtr );}


#endif
