//*****************************************************************************
//*
//*
//*		HttpScriptPlugin.cpp
//*
//*
//*****************************************************************************
//
//	This plugin allows to load html script files (*.htmx).
//	Tags in the script ("@@<name>@@") files will be replaced 
//	by data from a callback funktion.
//
#include	<string.h>
#include	<stdlib.h>
#include	<stdio.h>

#include	"HttpInternal.h"
#include	"HttpIconPlugin.h"
#include	"HttpScriptPlugin.h"

#define 	ENTRIES(a)		(sizeof(a)/sizeof((a)[0]))
#define 	HTTP_API		extern "C"			
#define 	SP_ALLOC		1024							// alloc size 
#define 	SP_TAG_MAX		64




static	char				cSuffix[8];						// suffix for the script
static	int					iSuffixLen=0;					// length of the suffix
static	unsigned 			uSuffixMask=0;					// mask for fast suffix compare 
static	unsigned 			uSuffixData=0;					// data for fast suffix compare 
static	int					iUserParamId=0;					// id for the user parameter
static	HttpScriptCB		pScriptRequest;					// callback for script data
static	HttpRequestCB		pOldRequest;					// next callback in chain



//*****************************************************************************
//*
//*		HttpScriptData
//*
//*****************************************************************************
//	Returns the html data of the folder
//	pBuffer		: pointer to the tranmit buffer
//	iSize		: is the size of the buffer (0=means dissconnect)
//	iOffset		: is the offest in the response
//	Returns the size of the stored data or an error number below zero
static int _cdecl HttpScriptData(HttpConnection hConn,char *pBuffer,int iSize,int iOffset) 
{
HttpData	   *pData=(HttpData*)hConn;
char		   *pTagPtr;	
char		   *pInput;	
SysFile			hFile;	 
int				iOut;
int				iMax;
int				iNum;
int				iLen;
int				iPos;
int				iAdd;



	hFile = (SysFile)HttpGetUserPointer(hConn,iUserParamId+2);
	iLen  =          HttpGetUserValue  (hConn,iUserParamId+1);
	iPos  =          HttpGetUserValue  (hConn,iUserParamId);


	if(iSize==0)											// trannsmision done ?
		{
		SysClose(hFile);
		pData->uSendAlloc = 0;
		HttpSetUserPointer(hConn,iUserParamId  ,0);
		HttpSetUserPointer(hConn,iUserParamId+1,0);
		HttpSetUserPointer(hConn,iUserParamId+2,0);
		return 0;
		}



	pInput	= pData->cSendBuffer+sizeof(pData->cSendBuffer)-SP_ALLOC;
	iMax	= iSize-1024;
	iOut	= 0;


	if(iLen>0 && pInput[iPos]=='@')							// recall of a tag
	if(pInput[iPos+1]=='*' && pInput[iPos+2]>='0')
		{
		for(iAdd=2;iAdd<SP_TAG_MAX && iAdd<iLen;iAdd++)
			{
			if(pInput[iPos+iAdd  ]!= 0 )continue;
			if(pInput[iPos+iAdd+1]!='@')break;

			iNum = pScriptRequest(hConn,pBuffer,iSize,pInput+iPos+1);

			if(iNum<0)return -1;
			if(iNum&HTTP_SCRIPT_RECALL)
				{
				return iNum&~HTTP_SCRIPT_RECALL;
				}

			iPos +=  iAdd+2;
			iOut  =  iNum; 
			}
		}
		


	while(iOut<iMax)
		{
		if(iPos>=iLen-SP_TAG_MAX)						
			{
			if(iLen==SP_ALLOC || !iOffset)				// load next part ot the script
				{
				memcpy(pInput-SP_TAG_MAX,pInput+SP_ALLOC-SP_TAG_MAX,SP_TAG_MAX);

				if(iOffset)iPos-=SP_ALLOC;
				else	   iOffset=1;	

				iLen = SysRead(hFile,pInput,SP_ALLOC);

				if(iLen<0       )iLen = 0;
				if(iLen<SP_ALLOC)pInput[iLen]=0;

				HttpSetUserValue(hConn,iUserParamId+1,iLen);
				}

			if(iPos>=iLen)
				{
				break;	
				}
			}

														// seach for a tag
		if(pInput[iPos]=='@' && pInput[iPos+1]=='@' && pInput[iPos+2]>='0')							
			{
			pTagPtr = pInput+iPos+2;

			for(iAdd=0;iAdd<SP_TAG_MAX-4;iAdd++)
				{
				if(pTagPtr[iAdd]=='@')break;
				if(pTagPtr[iAdd]< '0')break;
				}
														// check end of tag
			if(pTagPtr[iAdd]=='@' && pTagPtr[iAdd+1]=='@')
				{
				pTagPtr[iAdd] =  0;
				pTagPtr[ -1 ] = '*';
				
				iNum = pScriptRequest(hConn,pBuffer+iOut,iSize-iOut,pTagPtr);

				if(iNum<0)return -1;
				if(iNum&HTTP_SCRIPT_RECALL)
					{
					iNum &= ~HTTP_SCRIPT_RECALL;
					iOut +=  iNum; 
					break;
					}
				else{
					iPos +=  iAdd+4;
					iOut +=  iNum; 
					}
				
				continue;
				}
			}


		pBuffer[iOut] = pInput[iPos];
		iOut++;
		iPos++;
		}


	if(!iOut)											// trannsmision done ?
		{
		SysClose(hFile);
		pData->uSendAlloc = 0;
		HttpSetUserPointer(hConn,iUserParamId  ,0);
		HttpSetUserPointer(hConn,iUserParamId+1,0);
		HttpSetUserPointer(hConn,iUserParamId+2,0);
		return 0;
		}


	HttpSetUserValue(hConn,iUserParamId,iPos);



return iOut;
}

//*****************************************************************************
//*
//*		HttpScriptTrans
//*
//*****************************************************************************
//	Callback for a http request
//	hConn		: handle of the http connection	
//	pPath		: is the path for the script file
//	Returns a http error (>=100) or HTTP_RESP_???/HTTP_POST_???
HTTP_API int HttpScriptTrans(HttpConnection hConn,const char *pPath)
{
HttpData	   *pData=(HttpData*)hConn;
SysFile			hFile;
int				iErr;



	SysThreadCheck();
	SysWatchdogTrigger(pData->hWatchdog);


		hFile = SysOpenEx(pPath,SYS_READ,&iErr);
	if(!hFile)
		{
		return (iErr==SYS_ERR_ACCESS_DENIED)? 403:404;
		}


	HttpSetUserValue  (hConn,iUserParamId  ,0);
	HttpSetUserValue  (hConn,iUserParamId+1,0);
	HttpSetUserPointer(hConn,iUserParamId+2,hFile);
	HttpSetCacheMode  (hConn,HTTP_CACHE_NOCACHE);
	pData->uSendAlloc = SP_ALLOC+SP_TAG_MAX;


	SysThreadCheck();
	SysWatchdogTrigger(pData->hWatchdog);



return HttpTransmitCallback(hConn,HttpScriptData);
}

//*****************************************************************************
//*
//*		HttpScriptRequest
//*
//*****************************************************************************
//	Callback for a http request
//	hConn		: handle of the http connection	
//	iCommand	: is the type of the request command (HTTP_CMD_???)
//	pHost		: is the domain in the request (i.e. "www.myaddress.com")
//	pUrl		: is the url of the request (i.e. "/folder/index.html")
//	pParams		: is the parameter of the request 
//	iUrlLen		: is the length of the domain
//	iParamLen	: is the length of the parameter
//	Retruns a http error (>=100) or HTTP_RESP_???/HTTP_POST_??? 
static int _cdecl HttpScriptRequest(HttpConnection hConn,int iCommand,const char *pHost,const char *pUrl,const char *pParams,int iUrlLen,int iParamLen)
{
const char     *pPath;
unsigned		uValue;
int				iAttr;
int				iRet;



	if(iCommand!=HTTP_CMD_GET )								// check http command
	if(iCommand!=HTTP_CMD_HEAD)
		{
		goto NextPlugin;
		}

	if(iUrlLen<=iSuffixLen || pUrl[iUrlLen-iSuffixLen-1]!='.')
		{
		goto NextPlugin;
		}


	if(iUrlLen>=iSuffixLen)									// fast suffix compare
		{
		uValue  = *(unsigned*)(pUrl+iUrlLen-4);
		uValue ^= uSuffixData;
		uValue &= uSuffixMask;

		if(uValue)goto NextPlugin;

		if(iSuffixLen>4 && stricmp(pUrl+iUrlLen-iSuffixLen,cSuffix))
			{
			goto NextPlugin;
			}
		}
	

	pPath = HttpGetFile(hConn);
	iAttr = SysGetAttributes(pPath);

	if(iAttr<0 || (iAttr&SYS_DIR))							// is it a folder
		{
		goto NextPlugin;
		}	
	
	
	return iRet = HttpScriptTrans(hConn,pPath);


NextPlugin:


	if(pOldRequest)											// is an other plugin active	
		{
		return pOldRequest(hConn,iCommand,pHost,pUrl,pParams,iUrlLen,iParamLen);
		}


return 0;	
} 
 



//*****************************************************************************
//*
//*		HttpRegisterScript
//*
//*****************************************************************************
//	Registers the folder plugin. This function must called directly 
//	before HttpDaenon(...) is called.
//	pConfig		: config data of the server	
//	pCbScript	: callback for tag data
//	pSuffix		: is the suffix of the script files
//					0 = means the "*.htmx" suffix
//	Returns 0 if the plugin was installed or an error code
//					1 = no more user parameters
//					2 = suffix is to long
HTTP_API int HttpRegisterScript(HttpConfigData *pConfig,HttpScriptCB pCbScript,const char *pSuffix)  
{
char	cMask[sizeof(cSuffix)+4]={0,0,0,0};
int		iLen;
int		iPos;


	
		iUserParamId = HttpReserveUserParams(pConfig,3);
	if(!iUserParamId)return 1;

	
	if(!pSuffix)pSuffix="htmx";
	if(*pSuffix=='.')pSuffix++;


	   iLen = strlen(pSuffix);
	if(iLen>=sizeof(cSuffix)-1)return 2;

	cSuffix[0] = '.';
	memcpy(cSuffix,pSuffix,iLen+1);


	for(iPos=0;iPos<=iLen;iPos++)							// convert to lower case
		{
		cMask[iPos+4]  = ~('a'-'A');

		if(cSuffix[iPos]<'a')continue;
		if(cSuffix[iPos]>'z')continue;

		cSuffix[iPos] -= 'a'-'A';
		}

	iSuffixLen	= iLen;										// length of the suffix
	uSuffixMask	= *(unsigned*)(cMask+4+iSuffixLen-4);		// mask for fast suffix compare 
	uSuffixData	= *(unsigned*)(cSuffix+iSuffixLen-4);		// data for fast suffix compare 
	
	pOldRequest			= pConfig->pCbRequest;				// next callback in chain
	pScriptRequest		= pCbScript;
	pConfig->pCbRequest	= HttpScriptRequest;	

	

return 0;
}
 

//*****************************************************************************
//*
//*		HttpInitScript
//*
//*****************************************************************************
//	Initalizes the script plugin for external use with HttpScriptTrans(...). 
//	pCbScript	: callback for tag data
//	iUserId		: user id for internal use (three id's are used)
//	Returns 0 if the plugin was installed or an error code
//					1 = wrong user id
HTTP_API int HttpInitScript(HttpScriptCB pCbScript,int iUserId)  
{


	if((unsigned)iUserId>=HTTP_USER_PARAMS)return 1;

	pScriptRequest	= pScriptRequest;
	iUserParamId	= iUserId;
	

return 0;
}

