//*****************************************************************************
//*
//*
//*		HttpServer.cpp
//*
//*
//*****************************************************************************
#include	<string.h>
#include	<stdlib.h>
#include	<stdio.h>

#include	"HttpInternal.h"

#if			HTTP_HAVE_AUTH
#include	"HttpMd5.h"
#endif


#define 	ENTRIES(a)	(sizeof(a)/sizeof((a)[0]))


typedef struct
	{
	int			iCode;
	const char *pName;
	}ErrorCodes;


		void           *hHttpHeap		= 0;				// heap for the http server
static	SysSemaphore	hGlobalSem		= 0;				// global semaphore
static	SysSemaphore	hCloseSem		= 0;				// semaphore for waiting until next closing
static	int volatile	iCloseFlag		= 0;				// the listen thread waits for closing
static	int volatile	iServerRunning	= 0;				// is the daemon started
static	HttpServerData	sHttpServer;						// instance of the http server
static	const char	   *pHttpCommands[] = 					// all http commands
			{
			"GET",
			"HEAD",
			"POST",
			"PUT",
			"DELETE",
			"TRACE",
			"OPTIONS",
			"CONNECT",
			"LINK",
			"UNLINK"
			};

 static	ErrorCodes		aErrorCodes[] =
			{
			100,"Continue",
			101,"Switching Protocols",
			200,"OK",
			201,"Created",
			202,"Accepted",
			203,"Non-Authoritative Information",
			204,"No Content",
			205,"Reset Content",
			206,"Partial Content",
			207,"Multi-Status",
			300,"Multiple Choices",
			301,"Moved Permanently",
			302,"Moved Temporarily",
			303,"See Other",
			304,"Not Modified",
			305,"Use Proxy",
			307,"Temporary Redirect",
			400,"Bad Request",
			401,"Unauthorized",
			402,"Payment Required",
			403,"Forbidden",
			404,"Not Found",
			405,"Method Not Allowed",
			406,"Not Acceptable",
			407,"Proxy Authentication Required",
			408,"Request Time-out",
			409,"Conflict",
			410,"Gone",
			411,"Length Required",
			412,"Precondition Failed",
			413,"request entity too large",
			414,"Request-URI Too Large",
			415,"Unsupported Media Type",
			416,"Requested range not satisfiable",
			417,"Expectation Failed",
			421,"There are too many connections from your internet address",
			422,"Unprocessable Entity",
			423,"Locked",
			424,"Failed Dependency",
			425,"Unordered Collection",
			426,"Upgrade Required",
			500,"Internal Server Error",
			501,"Not Implemented",
			502,"Bad Gateway",
			503,"Service Unavailable",
			504,"Gateway Time-out",
			505,"HTTP Version not supported",
			506,"Variant Also Negotiates",
			507,"Insufficient Storage",
			509,"Bandwidth Limit Exceeded",
			510,"Not Extended",
			};

//*****************************************************************************
//*
//*		Memory Allocation
//*
//*****************************************************************************
#if HTTP_STATIC_MEMORY

	static	HttpData		aConnections[HTTP_STATIC_COUNT];
	static	char			aConnAlloc  [HTTP_STATIC_COUNT];

	static inline HttpData *AllocHttpData()
		{
		unsigned  uPos;

		SysSemaphoreLock(hGlobalSem);

		for(uPos=0;uPos<HTTP_STATIC_COUNT;uPos++)
			{
			if(aConnAlloc[uPos])continue;
			aConnAlloc[uPos]=1;
			break;
			}

		SysSemaphoreUnlock(hGlobalSem);

		if(uPos>=HTTP_STATIC_COUNT)return 0;

		return aConnections+uPos;
		}

	static inline void FreeHttpData(HttpData *pData)
		{
		unsigned	uPos;

		uPos = pData-aConnections;

		if(uPos<HTTP_STATIC_COUNT)
			{
			aConnAlloc[uPos] = 0;
			}
		}
#else

	typedef struct HttpEntryPtr
		{
		HttpData		sHttpData;							// http client data
		HttpEntryPtr   *pNextClient;						// pointer to the next client
		HttpEntryPtr   *pPrevClient;						// pointer to the previous client
		}HttpEntry;

	static	HttpEntry  *pFirstClient;	

	static inline HttpData *AllocHttpData()
		{
		HttpEntry   *pClient;	

		SysSemaphoreLock(hGlobalSem);

			pClient = (HttpEntry*)HttpAlloc(sizeof(HttpEntry));
		if(!pClient)
			{
			SysSemaphoreUnlock(hGlobalSem);
			return 0;
			}

		if(pFirstClient)pFirstClient->pPrevClient=pClient;

		pClient->pNextClient	= pFirstClient;
		pClient->pPrevClient	= 0;
		pFirstClient			= pClient;

		SysSemaphoreUnlock(hGlobalSem);

		return &pClient->sHttpData;
		}

	static inline void FreeHttpData(HttpData *pData)
		{
		HttpEntry   *pClient=(HttpEntry*)pData;	
		
		if(!pData)return;

		SysSemaphoreLock(hGlobalSem);
		
		if(pClient->pNextClient)
			{
			pClient->pNextClient->pPrevClient = pClient->pPrevClient;
			}

		if(pClient->pPrevClient)
			{
			pClient->pPrevClient->pNextClient = pClient->pNextClient;
			}
		
		HttpFree(pClient);

		SysSemaphoreUnlock(hGlobalSem);
		}

#endif

//*****************************************************************************
//*
//*		Rand32
//*
//*****************************************************************************
//	Gets a random 32 bit value
#if HTTP_HAVE_AUTH
inline unsigned Rand32()
{
unsigned	uVal;

	uVal	= rand();
	uVal   ^= rand()<<9;
	uVal   ^= rand()<<17;
	uVal   ^= SysTickCount();


return uVal;
}
#endif

//*****************************************************************************
//*
//*		CreateRand128
//*
//*****************************************************************************
//	Creates a random 128 bit string
//	pOutput		: 32 byte buffer
#if HTTP_HAVE_AUTH
void CreateRand128(char *pOutput)
{
unsigned		aRand[4];
unsigned char  *pData;
int				iVal;
int				iPos;


	aRand[0] = Rand32()^SysTimeGet();
	aRand[1] = SysTickCount();
	aRand[2] = Rand32();
	aRand[3] = Rand32();
	pData	 = (unsigned char*)aRand;

	for(iPos=0;iPos<32;iPos++)
		{
		iVal = pData[iPos/2];

		if(iPos&1)iVal>>=0x04;
		else	  iVal &=0x0F;

		pOutput[iPos] = "0123456789abcdef"[iVal];
		}

}
#endif

//*****************************************************************************
//*
//*		DecodeChar
//*
//*****************************************************************************
//	Translate a "%XX" statement into the original char
//	pText		: text behind the '%'char
//	cChar		: here the original char will be stored
//	Returns the count or procited chars
static inline int DecodeChar(const char *pText,char &cChar)
{

	     if(*pText>='0' && *pText<='9'){cChar=*pText-'0'   ;pText++;}
	else if(*pText>='A' && *pText<='F'){cChar=*pText-'A'+10;pText++;}
	else if(*pText>='a' && *pText<='f'){cChar=*pText-'a'+10;pText++;}
	else                               {cChar= '%';return 0;}

	     if(*pText>='0' && *pText<='9'){cChar=(cChar<<4)|(*pText-'0'   );return 2;}
	else if(*pText>='A' && *pText<='F'){cChar=(cChar<<4)|(*pText-'A'+10);return 2;}
	else if(*pText>='a' && *pText<='f'){cChar=(cChar<<4)|(*pText-'a'+10);return 2;}

return 1;
}

//*****************************************************************************
//*
//*		DecodeString
//*
//*****************************************************************************
//	Translate a text with "%XX" statements into the original string
//	pText		: is the text with the "%XX" statements
//	pTarget		: is the buffer where the original string will be stored
//	iSize		: is the size of the buffer
//	iLen		: is the length of the text
//	Returns the count stored chars chars
static int DecodeString(const char *pText,char *pTarget,int iSize,int iLen=0x7FFFFFFF)
{
char	cChar;
int		iPos;
int		iOff;


	if(iSize<=0)return 0;

	iOff = 0;

	for(iPos=0;iPos<iLen;iPos++)							// decode string
		{
		   cChar = pText[iPos];
		if(cChar==0)break;

		if(cChar=='%')
			{
			iPos+=DecodeChar(pText+iPos+1,cChar);
			}

		pTarget[iOff] = cChar;
		iOff++;

		if(iOff>=iSize)										// buffer overflow
			{
			pTarget[iSize-1]=0;
			return iSize;
			}
		}


	pTarget[iOff] = 0;



return iOff;
}

//*****************************************************************************
//*
//*		HttpEncodeString
//*
//*****************************************************************************
//	Translate a text to a http string with "%XX" statements
//	pText		: is the text tatements
//	pTarget		: is the buffer where the http string will be stored
//	iSize		: is the size of the buffer
//	iLen		: is the length of the text
//	Returns the count stored chars chars
HTTP_API int HttpEncodeString(const char *pText,char *pTarget,int iSize,int iLen)
{
char	cChar;
int		iPos;
int		iOff;


	if(iSize<=0)return 0;

	iOff = 0;

	for(iPos=0;iPos<iLen;iPos++)							// decode string
		{
		   cChar = pText[iPos];
		if(cChar==0)break;

		if(cChar<=' ' || cChar=='%')
			{
			if(iSize<4)break;

			pTarget[iOff]='%';iOff++;
			pTarget[iOff]="0123456789ABCDEF"[(cChar>>4)&0x0F];iOff++;
			pTarget[iOff]="0123456789ABCDEF"[(cChar   )&0x0F];iOff++;

			continue;
			}

		if(iSize<2)break;

		pTarget[iOff] = cChar;
		iOff++;
		}


	pTarget[iOff] = 0;



return iOff;
}

//*****************************************************************************
//*
//*		HttpEncodeHtml
//*
//*****************************************************************************
//	Translate a text to a html string with "&#XX;" statements
//	pText		: is the text tatements
//	pTarget		: is the buffer where the http string will be stored
//	iSize		: is the size of the buffer
//	iLen		: is the length of the text
//	Returns the count stored chars chars
HTTP_API int HttpEncodeHtml(const char *pText,char *pTarget,int iSize,int iLen)
{
char	cChar;
int		iPos;
int		iOff;
int		iNum;


	if(iSize<=0)return 0;

	iOff = 0;
	iNum = 0;

	for(iPos=0;iPos<iLen;iPos++)							// decode string
		{
		   cChar = pText[iPos];
		if(cChar==0)break;

		if(cChar<' ' || cChar=='>' || cChar=='<' || cChar=='&')
			{
			if(iSize<6)break;
			iOff+=sprintf(pTarget+iOff,"&#%i;",cChar&0xFF);
			continue;
			}

		if(cChar==' ')
			{
			if(iPos==0 || pText[iPos-1]!=' ')
				{
				iNum=1;
				}
			else if(iNum==1)
				{
				if(iSize<6)break;
				memcpy(pTarget+iOff,"&#160;",6);
				iOff+=6;
				iNum =0;
				continue;
				}
			else{
				iNum = 1;
				}
			}

		if(iSize<2)break;

		pTarget[iOff] = cChar;
		iOff++;
		}


	pTarget[iOff] = 0;



return iOff;
}

//*****************************************************************************
//*
//*		EncodeTime
//*
//*****************************************************************************
//	Translate a system time into a time string "Sun, 06 Nov 1994 08:49:37 GMT"
//	uTime		: is the system time
//	pBuffer		: is the buffer for the time string
//	Returns the length of the time string.
static int EncodeTime(SysUtc  uTime,char *pBuffer)
{
static const char *pWeekDsys[ 8] = {"Sun","Mon","Tue","Wed","Thu","Fri","Sat","???"};
static const char *pMonthTxt[16] = {"???","Jan","Feb","Mar","Apr","May","Jun","Jul",
									"Aug","Sep","Oct","Nov","Dec","???","???","???"};

SysTimeStruct	sData;
int				iDay;
int				iLen;

	iDay = SysTimeToWeekday(uTime);
	SysTimeToStruct(&sData,uTime,0);


	iLen  = sprintf(pBuffer,"%s, %02u %s %04u %02u:%02u:%02u GMT",
					pWeekDsys[iDay&0x07],
					sData.ucDay,
					pMonthTxt[sData.ucMonth&0x0F],
					sData.usYear,
					sData.ucHour,
					sData.ucMinute,
					sData.ucSecond);

return iLen;
}

//*****************************************************************************
//*
//*		GetContentType
//*
//*****************************************************************************
//	Gets the content-type of a filetype
//	pFileType		: pointer to the filetype (i.e. "htm")
//	pContentType	: is the buffer for the content string
//	Returns the string size of the content
static int GetContentType(const char *pFileType,char *pContentType)
{

	if(!stricmp(pFileType,"html") ||
	   !stricmp(pFileType,"htm"))
		{
		memcpy(pContentType,"text/html",10);
		return 9;
		}

	if(!stricmp(pFileType,"txt"))
		{
		memcpy(pContentType,"text/txt",9);
		return 8;
		}

	if(!stricmp(pFileType,"gif"))
		{
		memcpy(pContentType,"image/gif",10);
		return 9;
		}

	if(!stricmp(pFileType,"bmp"))
		{
		memcpy(pContentType,"image/bmp",10);
		return 9;
		}

	if(!stricmp(pFileType,"png"))
		{
		memcpy(pContentType,"image/png",10);
		return 9;
		}

	if(!stricmp(pFileType,"jpeg") ||
	   !stricmp(pFileType,"jpg"))
		{
		memcpy(pContentType,"image/jpeg",11);
		return 10;
		}

	if(!stricmp(pFileType,"mpeg") ||
	   !stricmp(pFileType,"mpg"))
		{
		memcpy(pContentType,"video/mpeg",11);
		return 10;
		}

	if(!stricmp(pFileType,"avi"))
		{
		memcpy(pContentType,"video/avi",10);
		return 9;
		}

	if(!stricmp(pFileType,"wmv"))
		{
		memcpy(pContentType,"video/wmv",10);
		return 9;
		}

	if(!stricmp(pFileType,"divx"))
		{
		memcpy(pContentType,"video/divx",11);
		return 10;
		}

	if(!stricmp(pFileType,"wav"))
		{
		memcpy(pContentType,"audio/wave",11);
		return 10;
		}

	if(!stricmp(pFileType,"mp3"))
		{
		memcpy(pContentType,"audio/mpeg",11);
		return 10;
		}

	if(!stricmp(pFileType,"mp2"))
		{
		memcpy(pContentType,"audio/mpeg",11);
		return 10;
		}

	if(!stricmp(pFileType,"aac"))
		{
		memcpy(pContentType,"audio/aac",9);
		return 10;
		}

	if(!stricmp(pFileType,"ogg"))
		{
		memcpy(pContentType,"audio/ogg",9);
		return 10;
		}

	if(!stricmp(pFileType,"m4a"))
		{
		memcpy(pContentType,"audio/m4a",9);
		return 10;
		}

	if(!stricmp(pFileType,"pdf"))
		{
		memcpy(pContentType,"application/pdf",16);
		return 15;
		}

	if(!stricmp(pFileType,"zip"))
		{
		memcpy(pContentType,"application/zip",16);
		return 15;
		}

	if(!stricmp(pFileType,"tar"))
		{
		memcpy(pContentType,"application/tar",16);
		return 15;
		}

	if(!stricmp(pFileType,"gz"))
		{
		memcpy(pContentType,"application/gz",15);
		return 14;
		}

	pContentType[0]=0;



return 0;
}

//*****************************************************************************
//*
//*		CreateResponseHeader
//*
//*****************************************************************************
//	Creates a response header (i.e.: "HTTP/1.1 403 Forbidden\r\n")
//	pBuffer			: buffer for the header with a minimum size of 64 bytes
//	iErrorCode		: is the error code
//	iVersion		: is the http version 0xYYXX
//	iAddLf			: adds a "\r\n" suffix
//	Returns the size of the header.
static int CreateResponseHeader(char *pBuffer,int iErrorCode,int iVersion,int iAddLf=0)
{
const char *pName;
int			iStart;
int			iStop;
int			iHalf;
int			iCmp;
int			iLen;


	iStart	= 0;
	iStop	= ENTRIES(aErrorCodes)-1;

	for(;;)
		{
		iHalf = (iStop+iStart+1)/2;
		iCmp  =  iErrorCode-aErrorCodes[iHalf].iCode;


		if(iHalf<=iStart)
			{
			pName = (iCmp)? "Unknown error code":aErrorCodes[iHalf].pName;
			break;
			}

		if(iCmp<0)
			{
			iStop = iHalf-1;
			continue;
			}

		if(iCmp>0)
			{
			   iStart=iHalf+1;
			if(iStart>iStop)iStart=iStop;
			continue;
			}

		pName = aErrorCodes[iHalf].pName;
		break;
		}

	iLen = sprintf(pBuffer,"HTTP/%u.%u %u %s\r\n",iVersion>>8,iVersion&0xFF,iErrorCode,pName);

	memcpy(pBuffer+iLen,"Date: ",6);
	iLen+=6;
	iLen+=EncodeTime(SysTimeGet(),pBuffer+iLen);
	memcpy(pBuffer+iLen,"\r\n"  ,2);
	iLen+=2;

	if(iAddLf)
		{
		memcpy(pBuffer+iLen,"Content-Length: 0\r\n\r\n",21);
		iLen+=21;
		}


return iLen;
}

//*****************************************************************************
//*
//*		GetUserCacheEntry
//*
//*****************************************************************************
//	Seaches fpr an username in the user cache
//	pData			: buffer for the header with a minimum size of 64 bytes
//	pUserName		: is the username
//	Returns the pointer to the cache entry, or 0 if not found
#if HTTP_HAVE_AUTH
static HttpUserCache *GetUserCacheEntry(HttpServerData *pServer,const char *pUserName)
{
HttpUserCache  *pEntry;
int				iStart;
int				iStop;
int				iHalf;
int				iCmp;


	if(!*pUserName)return 0;

	SysSemaphoreLock(pServer->hSemaphore);

	iStart	= 0;
	iStop	= pServer->uUserIndexSize-1;

	for(;;)
		{
		iHalf  = (iStop+iStart+1)/2;
		pEntry = pServer->aUserCache+pServer->aUserIndex[iHalf];
		iCmp   = strcmp(pUserName,pEntry->cUserName);


		if(iHalf<=iStart)
			{
			if(iCmp)pEntry=0;
			break;
			}

		if(iCmp<0)
			{
			iStop = iHalf-1;
			continue;
			}

		if(iCmp>0)
			{
			   iStart=iHalf+1;
			if(iStart>iStop)iStart=iStop;
			continue;
			}

		break;
		}

	SysSemaphoreUnlock(pServer->hSemaphore);



return pEntry;
}
#endif

//*****************************************************************************
//*
//*		SetUserCacheEntry
//*
//*****************************************************************************
//	Seaches for an username in the user cache
//	pData			: buffer for the header with a minimum size of 64 bytes
//	pUserName		: is the username
//	pPassWord		: Ii the password for the user
//	Returns the pointer to the cache entry, or 0 if not found
#if HTTP_HAVE_AUTH
static HttpUserCache *SetUserCacheEntry(HttpServerData *pServer,const char *pUserName,const char *pNonce,const char *pPassWord)
{
HttpUserCache  *pEntry;
HttpUserCache  *pInsert;
unsigned		uDelete;
unsigned		uOffset;
unsigned		uLast;
SysDW			uTime;
int				iStart;
int				iStop;
int				iHalf;
int				iCmp;
int				iPos;


	SysSemaphoreLock(pServer->hSemaphore);

	uOffset = pServer->uUserNext;

	if(uOffset>HTTP_USER_CACHE)
		{
		uOffset=0;
		}

	uLast   = uOffset;
	uTime	= 0xFFFFFFFF;


	for(iPos=0;iPos<HTTP_USER_CACHE;iPos++)					// find empty slot
		{
			pEntry = pServer->aUserCache+uOffset;
		if(!pEntry->cUserName[0])
			{
			pServer->uUserNext = uOffset+1;
			break;
			}

		if(pEntry->dwLastAccess<uTime || uLast==0xFFFFFFFF)
			{
			pEntry->cAuthNonce[0] = 0;
			pEntry->cUserName [0] = 0;
			pEntry->cPassWord [0] = 0;

			uDelete	= pEntry->uIndexOffset;
			uTime	= pEntry->dwLastAccess;
			uLast	= uOffset;

			if(pServer->uUserIndexSize>0)					// remove the old entry
				{
				memmove(pServer->aUserIndex+uDelete+1,
						pServer->aUserIndex+uDelete  ,
						sizeof(pServer->aUserIndex[0])*(pServer->uUserIndexSize-uDelete-1));

				pServer->uUserIndexSize--;
				}
			}

		uOffset++;

		if(uOffset>HTTP_USER_CACHE)
			{
			uOffset=0;
			}
		}

	if(iPos>=HTTP_USER_CACHE)								// take oldes entry
		{
		pEntry = pServer->aUserCache+uLast;
		}


	strncpy(pEntry->cUserName ,pUserName,sizeof(pEntry->cUserName)-1);
	strncpy(pEntry->cPassWord ,pPassWord,sizeof(pEntry->cPassWord)-1);
	memcpy (pEntry->cAuthNonce,pNonce,32);
	pEntry->dwLastAccess = SysTimeGet();


	iStart	= 0;
	iStop	= pServer->uUserIndexSize-1;

	if(pServer->uUserIndexSize==0)							// insert first entry
		{
		pServer->aUserIndex[0]  = uOffset;
		pServer->uUserIndexSize = 1;
		pEntry ->uIndexOffset   = 0;
		}
	else for(;;)											// find insert position
		{
		iHalf   = (iStop+iStart+1)/2;
		pInsert = pServer->aUserCache+pServer->aUserIndex[iHalf];
		iCmp    = strcmp(pUserName,pInsert->cUserName);


		if(iHalf<=iStart)
			{
			if(iCmp)										// insert entry in index list
				{
				if(iCmp>0)iHalf++;
				memmove(pServer->aUserIndex+iHalf+1,
						pServer->aUserIndex+iHalf,
						sizeof(pServer->aUserIndex[0])*(pServer->uUserIndexSize-iHalf));

				pEntry ->uIndexOffset = iHalf;
				pServer->aUserIndex[iHalf]=uOffset;
				pServer->uUserIndexSize++;
				break;
				}

			pEntry->cUserName[0]=0;							// should no be happen
			pEntry=0;
			break;
			}

		if(iCmp<0)
			{
			iStop = iHalf-1;
			continue;
			}

		if(iCmp>0)
			{
			   iStart=iHalf+1;
			if(iStart>iStop)iStart=iStop;
			continue;
			}

		pEntry->cUserName[0]=0;								// should no be happen
		pEntry=0;
		break;
		}




	SysSemaphoreUnlock(pServer->hSemaphore);


return pEntry;
}
#endif

//*****************************************************************************
//*
//*		DelUserCacheEntry
//*
//*****************************************************************************
//	Removes an username in the user cache
//	pData			: buffer for the header with a minimum size of 64 bytes
//	pUserName		: is the username
//	Returns 1 if the user was found and deleted or 0 if not.
#if HTTP_HAVE_AUTH
static int DelUserCacheEntry(HttpServerData *pServer,const char *pUserName)
{
HttpUserCache  *pEntry;
int				iStart;
int				iStop;
int				iHalf;
int				iCmp;
int				iPos;


	if(!*pUserName)return 0;

	SysSemaphoreLock(pServer->hSemaphore);

	iStart	= 0;
	iStop	= pServer->uUserIndexSize-1;

	for(;;)
		{
		iHalf  = (iStop+iStart+1)/2;
		pEntry = pServer->aUserCache+pServer->aUserIndex[iHalf];
		iCmp   = strcmp(pUserName,pEntry->cUserName);


		if(iHalf<=iStart)
			{
			if(iCmp)pEntry=0;
			break;
			}

		if(iCmp<0)
			{
			iStop = iHalf-1;
			continue;
			}

		if(iCmp>0)
			{
			   iStart=iHalf+1;
			if(iStart>iStop)iStart=iStop;
			continue;
			}

		break;
		}

	if(pEntry)
		{
		iPos = pEntry->uIndexOffset;

		memmove(pServer->aUserIndex+iPos,
				pServer->aUserIndex+iPos+1,
				sizeof(pServer->aUserIndex[0])*(pServer->uUserIndexSize-iPos));

		pEntry->cUserName [0] = 0;
		pEntry->cPassWord [0] = 0;
		pEntry->cAuthNonce[0] = 0;
		pServer->uUserIndexSize--;
		}


	SysSemaphoreUnlock(pServer->hSemaphore);



return (pEntry)? 1:0;
}
#endif


//*****************************************************************************
//*
//*		ScanAuthCmd
//*
//*****************************************************************************
//	Scanns a line with an authorization command
//		"Digest username="...", realm="xxx@yyy.zzz", nonce="a5...",
//			uri="/...", response="fed...", opaque="b4..."
//	pData		: pointer to the connection data
//	pLine		: pointer to the line
//	iLen		: is the length of the line
//	Returns 0 if the line was ok or an http error code
//				1 = means goto callback response
int ScanAuthCmd(HttpData *pData,const char *pLine,int iLen)
{
#if HTTP_HAVE_AUTH

Md5Context		sMd5;
HttpUserCache  *pEntry;
unsigned char	aData[16];
char			cHal1[36];
char			cHal2[36];
char			cResp[36];
char			cPass[HTTP_MAX_PASSWD];
const char     *pRealm;
const char     *pPass;
const char     *pUser;
const char     *pResp;
const char     *pMeth;
const char     *pUri;
int				iFrom;
int				iRealm;
int				iPass;
int				iUser;
int				iResp;
int				iMeth;
int				iUri;
int				iPos;
int				iVal;
int				iEnd;
int				iRet;




	if(pData->uIgnoreLogin)								// don't check the password
		{
		return 0;
		}

	if(pLine[0]==' ')
		{
		pLine++;
		iLen--;
		}

	if(strnicmp(pLine,"Digest ",7))						// check authorization mode
		{
		return 411;
		}

	iPos  = 7;
	iUser = 0;
	iResp = 0;
	iUri  = 0;
	pUser = "";
	pResp = "";
	pUri  = "";



	while(pLine[iPos])									// authorization parameters
		{
		if(pLine[iPos]==' ' || pLine[iPos]==',')
			{
			iPos++;
			continue;
			}

		iVal = 0;

		if(!strnicmp(pLine+iPos,"username=",9))
			{
			iPos += 9;
			iVal  = 1;
			}
		else if(!strnicmp(pLine+iPos,"uri=",4))
			{
			iPos += 4;
			iVal  = 3;
			}
		else if(!strnicmp(pLine+iPos,"response=",9))
			{
			iPos += 9;
			iVal  = 2;
			}
		else{
			for(;pLine[iPos];iPos++)
				{
				if(pLine[iPos]==' ')break;
				if(pLine[iPos]==',')break;
				if(pLine[iPos]=='='){iPos++;break;}
				}
			}

		if(pLine[iPos]=='"')
			{
			iPos  += 1;
			iFrom  = iPos;

			for(;pLine[iPos];iPos++)
				{
				if(pLine[iPos]!='"')continue;
				iEnd=iPos;
				iPos++;
				break;
				}
			}
		else{
			iFrom = iPos;

			for(iPos++;pLine[iPos];iPos++)
				{
				if(pLine[iPos]==' ')break;
				if(pLine[iPos]==',')break;
				}

			iEnd = iPos;
			}


		switch(iVal)
			{
		case 1:	pUser = pLine+iFrom;	iUser=iEnd-iFrom;	break;
		case 2:	pResp = pLine+iFrom;	iResp=iEnd-iFrom;	break;
		case 3:	pUri  = pLine+iFrom;	iUri =iEnd-iFrom;	break;
			}

		}


	if(iUser>=sizeof(pData->cAuthUser))
		{
		iUser=sizeof(pData->cAuthUser)-1;
		}


	memcpy(pData->cAuthUser,pUser,iUser);
	pData->cAuthUser[iUser]=0;
	pEntry = GetUserCacheEntry(pData->pServer,pData->cAuthUser);


	if(!pData->cAuthNonce[0])							// no user nonce
		{
		if(!pEntry)
			return 401;

		memcpy(pData->cAuthNonce,pEntry->cAuthNonce,32);
		}

	if(pEntry)
		{
		pPass 	= pEntry->cPassWord;
		}
	else if(pData->pServer->pCbPassword)				// get password via callback
		{
		   iRet	= pData->pServer->pCbPassword(pData,pUser,cPass,sizeof(cPass));
		if(iRet==-1)return 0;
		if(iRet!= 0)return iRet;						// no password check

		pPass	= cPass;
		}
	else{
		return 403;
		}

	iPass 	= strlen(pPass);
	pRealm	= pData->pServer->cAuthRealm;
	iRealm	= strlen(pRealm);
	pMeth	= pHttpCommands[pData->uRequCommand];
	iMeth	= strlen(pMeth);


	Md5Init  (&sMd5);									// calculate md5 value
	Md5Update(&sMd5,pUser ,iUser );
	Md5Update(&sMd5,":"   ,1     );
	Md5Update(&sMd5,pRealm,iRealm);
	Md5Update(&sMd5,":"   ,1     );
	Md5Update(&sMd5,pPass ,iPass );
	Md5Final (&sMd5,aData        );
	Md5ToString(cHal1,aData);

	Md5Init  (&sMd5);
	Md5Update(&sMd5,pMeth ,iMeth );
	Md5Update(&sMd5,":"   , 1    );
	Md5Update(&sMd5,pUri  ,iUri  );
	Md5Final (&sMd5,aData        );
	Md5ToString(cHal2,aData);

	Md5Init  (&sMd5);
	Md5Update(&sMd5,cHal1            ,32);
	Md5Update(&sMd5,":"              , 1);
	Md5Update(&sMd5,pData->cAuthNonce,32);
	Md5Update(&sMd5,":"              , 1);
	Md5Update(&sMd5,cHal2            ,32);
	Md5Final (&sMd5,aData               );
	Md5ToString(cResp,aData);


	HTTP_LOG_AUTH(("\n\nAuth:\n"))
	HTTP_LOG_AUTH(("\turl  :\t%.*s\n",iUri  ,pUri  ))
	HTTP_LOG_AUTH(("\tuser :\t%.*s\n",iUser ,pUser ))
	HTTP_LOG_AUTH(("\tpass :\t%.*s\n",iPass ,pPass ))
	HTTP_LOG_AUTH(("\tmeth :\t%.*s\n",iMeth ,pMeth ))
	HTTP_LOG_AUTH(("\trealm:\t%.*s\n",iRealm,pRealm))
	HTTP_LOG_AUTH(("\tresp1:\t%.*s\n",32    ,cResp ))
	HTTP_LOG_AUTH(("\tresp2:\t%.*s\n",32    ,pResp ))
	HTTP_LOG_AUTH(("\tnonce:\t%.*s\n",32    ,pData->cAuthNonce))
	HTTP_LOG_AUTH(("\n"));


	if(memcmp(pResp,cResp,32))							// ckeck the password
		{
		return 401;
		}


	if(!pEntry)
		{
		pEntry = SetUserCacheEntry(pData->pServer,pData->cAuthUser,pData->cAuthNonce,pPass);
		}
	else{
		SysSemaphoreLock(pData->pServer->hSemaphore);

		if(!strcmp(pEntry->cUserName,pData->cAuthUser))
			{
			memcpy(pEntry->cAuthNonce,pData->cAuthNonce,32);
			pEntry->dwLastAccess = SysTimeGet();
			}

		SysSemaphoreUnlock(pData->pServer->hSemaphore);
		}


#endif

return 0;
}

//*****************************************************************************
//*
//*		ScanRequestCmd
//*
//*****************************************************************************
//	Scanns the first line of a http request
//	pData		: pointer to the connection data
//	pLine		: pointer to the line
//	iLen		: is the length of the line
//	Returns 0 if the line was ok or an http error code
static int ScanRequestCmd(HttpData *pData,const char *pLine,int iLen)
{
const char *pCommand;
char	   *pTarget;
unsigned   *pLength;
char		cChar;
int			iMax;
int			iPos;
int			iCmd;
int			iRet;
int			iOff;



	HTTP_LOG_REQU(("\n>> %s\n",pLine))

	pData->uRequHttpVersion	= 0;
	pData->uRequContentLen	= 0;
	pData->uRequKeepAlive	= 0;
	pData->cRequCookie[0]	= 0;
	pData->cRequParam [0]	= 0;
	pData->cRequPath  [0]	= 0;
	pData->cRequHost  [0]	= 0;
	pData->cRequFile  [0]	= 0;


	for(iPos=0;;iPos++)										// find end of command string
		{
		if(iPos>=iLen)										// wrong format
			{
			pData->uRequCommand = HTTP_CMD_NOTSOPPORTED;
			return 400;
			}

		if(pLine[iPos]==' ')break;
		}

	for(iCmd=0;;iCmd++)										// find command value
		{
		if(iCmd>=ENTRIES(pHttpCommands))					// command not implemented
			{
			iRet=501;
			break;
			}

		pCommand = pHttpCommands[iCmd];
		if(memcmp(pLine,pCommand,iPos))continue;
		if(pCommand[iPos]==0)break;
		}


	pData->uRequParamLength	= 0;
	pData->uRequPathLength	= 0;
	pData->uRequCommand		= iCmd;
	pLength = &pData->uRequPathLength;
	pTarget =  pData->cRequPath;
	iMax    =  HTTP_MAX_PATH;
	iOff	=  0;


	for(iPos++;;iPos++)										// decode path
		{
		if(iPos>=iLen)										// wrong format
			{
			pTarget[iOff]=0;
			return 400;
			}

		cChar = pLine[iPos];

		if(cChar==' ')										// end of url
			{
			pLength[0]=iOff;
			pTarget[iOff]=0;
			iPos++;
			break;
			}

		if(cChar=='%')										// special hexchar
			{
			iPos+=DecodeChar(pLine+iPos+1,cChar);
			}

		if(cChar==':' || cChar=='?')						// found parameter
			{
			pTarget[iOff] = 0;
			pLength[0] = iOff;
			pLength = &pData->uRequParamLength;
			pTarget =  pData->cRequParam;
			iMax    =  HTTP_MAX_PARAM;
			iOff    =  0;
			continue;
			}

		if(iOff>=HTTP_MAX_PATH-1)							// request URL too long
			{
			pTarget[iOff] = 0;
			return 414;
			}

		pTarget[iOff] = cChar;
		iOff++;
		}



	if(memcmp(pLine+iPos,"HTTP/",5))						// get http version
		{
		return 400;
		}

	iPos += 5;


	if(!memcmp(pLine+iPos,"1.1",3))
		{
		pData->uRequHttpVersion = 0x0101;
		pData->uRequKeepAlive	= 1;
		iPos += 3;
		}
	else if(!memcmp(pLine+iPos,"1.0",3))
		{
		pData->uRequHttpVersion = 0x0100;
		iPos += 3;
		}
	else if(pLine[iPos]>='0' && pLine[iPos]<='9')			// http version not supported:
		{
		return 505;
		}
	else{
		return 400;
		}


	if(pLine[iPos]==0)
		{
		return 0;
		}


	if(pLine[iPos]>='0' && pLine[iPos]<='9')				// http version not supported:
		{
		return 505;
		}


return 400;
}


//*****************************************************************************
//*
//*		ScanRequestParam
//*
//*****************************************************************************
//	Scanns a parameter line of a http request
//	pData		: pointer to the connection data
//	pLine		: pointer to the line
//	iLen		: is the length of the line
//	Returns 0 if the line was ok or an http error code
static int ScanRequestParam(HttpData *pData,const char *pLine,int iLen)
{
char		cTemp[64];
int			iPos;
int			iEnd;


	HTTP_LOG_REQU((">>(%04i/%04i) %s \n",iLen,pLine-pData->cRecvBuffer,pLine))


	if(!strnicmp(pLine,"Host:",5))							// Get requested host
		{
		iPos = (pLine[5]==' ')? 6:5;
		iEnd =  DecodeString(pLine+iPos,pData->cRequHost,sizeof(pData->cRequHost),iLen-iPos);

		if(iEnd>=sizeof(pData->cRequHost))
			{
			return 413;										// request entity too large
		 	}

		return 0;
		}


	if(!strnicmp(pLine,"Cookie:",7))						// Store the cookies
		{
		iPos = (pLine[7]==' ')? 8:7;
		iEnd =  DecodeString(pLine+iPos,pData->cRequCookie,sizeof(pData->cRequCookie),iLen-iPos);

		if(iEnd>=sizeof(pData->cRequCookie))
			{
			return 413;										// request entity too large
		 	}

		return 0;
		}


	if(!strnicmp(pLine,"User-Agent:",11))					// Store the cookies
		{
		iPos = (pLine[11]==' ')? 12:11;
		iEnd =  DecodeString(pLine+iPos,pData->cRequBrowser,sizeof(pData->cRequBrowser),iLen-iPos);

		return 0;
		}


	if(!strnicmp(pLine,"Connection:",11))					// Get Keep-Alive mode
		{
		iPos = (pLine[11]==' ')? 12:11;
		iEnd =  DecodeString(pLine+iPos,cTemp,sizeof(cTemp),iLen-iPos);

		if(iEnd>=sizeof(cTemp))
			{
			return 413;										// request entity too large
		 	}

		if(!stricmp(cTemp,"keep-alive"))
			{
			pData->uRequKeepAlive = 1;
			}
		else if(!stricmp(cTemp,"close"))
			{
			pData->uRequKeepAlive = 0;
			}

		return 0;
		}


	if(!strnicmp(pLine,"Content-Length:",15))				// Get Content-Length
		{
		iPos = (pLine[15]==' ')? 16:15;

		if(pLine[iPos]<'0' || pLine[iPos]>'9')
			{
			return 400;
			}

		pData->uRequContentLen = atoi(pLine+iPos);

		return 0;
		}

	if(!strnicmp(pLine,"Transfer-Encoding:",18))			// Get Transfer-Encoding
		{
		iPos = (pLine[18]==' ')? 19:18;

		if(strnicmp(pLine+iPos,"chunked",7))
			{
			return 411;
			}

		pData->uRequContentLen = 0xFFFFFFFF;

		return 0;
		}

	if(!strnicmp(pLine,"Authorization:",14))				// Get Authorization
		{
		return ScanAuthCmd(pData,pLine+14,iLen-14);
		}



return 0;
}

//*****************************************************************************
//*
//*		CallbackReadFile
//*
//*****************************************************************************
//	Callback for file transmisssion
//	hConn	: pointer to the connection data
//	pBuffer	: pointer to input puffer
//	iSize	: size of the input buffer
//	iOffset	: is the offset in the file
//	Returns the count of stored bytes, 0 at the file end or below 0 at an abort.
static int _cdecl CallbackReadFile(HttpConnection hConn,char *pBuffer,int iSize,int iOffset)
{
HttpData   *pData=(HttpData*)hConn;
int			iLen;


	   iLen = SysRead(pData->hFile,pBuffer,iSize);
	if(iLen>0)return iLen;

	SysClose(pData->hFile);
	pData->hFile = 0;


return iLen;
}

//*****************************************************************************
//*
//*		CallbackReadMem
//*
//*****************************************************************************
//	Callback for memory transmisssion
//	hConn	: pointer to the connection data
//	pBuffer	: pointer to input puffer
//	iSize	: size of the input buffer
//	iOffset	: is the offset in the memory
//	Returns the count of stored bytes, 0 at the file end or below 0 at an abort.
static int _cdecl CallbackReadMem(HttpConnection hConn,char *pBuffer,int iSize,int iOffset)
{
HttpData   *pData=(HttpData*)hConn;
int			iLen;



	   iLen = pData->iResponseLen-pData->iMemoryPos;
	if(iLen<0)return 0;
	if(iLen>iSize)iLen=iSize;

	memcpy(pBuffer,pData->pMemoryPtr+pData->iMemoryPos,iLen);
	pData->iMemoryPos += iLen;


return iLen;
}

//*****************************************************************************
//*
//*		CallbackWriteMem
//*
//*****************************************************************************
//	Callback for memory transmisssion
//	hConn	: pointer to the connection data
//	pBuffer	: pointer to input puffer
//	iSize	: size of the input buffer
//	iOffset	: is the offset in the memory
//	Returns the count of stored bytes, 0 at the file end or below 0 at an abort.
static int _cdecl CallbackWriteMem(HttpConnection hConn,char *pBuffer,int iSize,int iOffset)
{
HttpData   *pData=(HttpData*)hConn;
int			iLen;



	   iLen = pData->iMemoryMax-pData->iMemoryPos;
	if(iLen<0)return 0;
	if(iLen<iSize)return 413;							// request entity too large:

	memcpy(pData->pMemoryPtr+pData->iMemoryPos,pBuffer,iSize);
	pData->iMemoryPos += iSize;



return iLen;
}



//*****************************************************************************
//*
//*		DefaultResponse
//*
//*****************************************************************************
//	Creates a default get response
//	pData		: pointer to the connection data
//	iCommand	: is the command of the request (see HTTP_CMD_???)
//	Returns the error code for the request
static int DefaultResponse(HttpData *pData,int iCommand)
{
SysFindStruct	sInfo;
SysFile			hFile;
int 			iError;
int				iAttr;
int				iSize;
int				iLen;



	   iAttr = SysGetAttributes(pData->cRequFile);
	if(iAttr<0){iError=iAttr;goto Error;}
	if(iAttr&SYS_DIR)									// is the filename a path
		{
		iLen = strlen(pData->cRequFile);

		if(iLen+pData->pServer->iDefaultExtLen>=sizeof(pData->cRequFile))
			{
			return 413;									// request entity too large
			}

		memcpy(pData->cRequFile+iLen,pData->pServer->cDefaultExt,pData->pServer->iDefaultExtLen+1);

		   iAttr = SysGetAttributes(pData->cRequFile);
		if(iAttr<0){iError=iAttr;goto Error;}
		if(iAttr&SYS_DIR)								// is the filename a path
			{
			return 403;
			}
		}


	SysThreadCheck();
	SysWatchdogTrigger(pData->hWatchdog);


		hFile = SysOpenEx(pData->cRequFile,SYS_READ|SYS_SHARE_RD,&iError);
	if(!hFile)goto Error;

	if(SysGetFileInfo(hFile,&sInfo)>=0)					// get modify time
		{
		pData->uResponseTime = sInfo.dwTimeWrite;
		}


	   iSize = SysGetFileSize(hFile);
	if(iSize<=0)										// use chunked output
		{
		pData->iResponseLen = -1;
		}
	else{
		pData->iResponseLen = iSize;
		}

	pData->pResponseCb  = CallbackReadFile;
	pData->hFile		= hFile;


	return 1;

Error:

	switch(iError)
		{
	case SYS_ERR_INVALID_FILENAME:
	case SYS_ERR_FILE_NOT_FOUND:
	case SYS_ERR_PATH_NOT_FOUND:		return 404;			// Not found
		}


return 403;												// Forbidden
}

//*****************************************************************************
//*
//*		GetPostData
//*
//*****************************************************************************
//	Gets the data of a post request
//	pData		: pointer to the connection data
//	pInput		: pointer to the current input buffer
//	iSize		: size of the bytes in the input buffer
//	pRead		: here the count off readed bytes from pInput is stored
//	Returns an http error code or
//					0 for a common response
//					1 for a callback response
//				   >0 for HTTP_MSG_???? messagen (use return -1-HTTP_MSG_???)
static int GetPostData(HttpData *pData,const char *pInput,int iSize,int *pRead)
{
char		cTemp;
unsigned 	uLast;
unsigned	uDiff;
unsigned	uTime;
int			iRest;
int			iAll;
int			iOff;
int			iLen;
int			iMax;
int			iPos;
int			iNum;
int			iAdd;
int			iErr;
int			iRet;


	iErr  = 0;
	uLast = SysTickCount();



	if(pData->uRequContentLen!=0xFFFFFFFF)				// user content-length
		{
		iMax = pData->uRequContentLen;

		   iLen = iSize;
		if(iLen>iMax)
			{
			iLen = iMax;
			}

		memcpy(pData->cSendBuffer,pInput,iLen);
		pData->cSendBuffer[iLen]=0;
		if(pRead)*pRead = iLen;
		iPos = iLen;

		for(;;)
			{
			while(iPos<iMax)							// get data from socket
				{
				SysThreadCheck();
				SysWatchdogTrigger(pData->hWatchdog);

				   iNum = TcpGetNext(pData->hConnection,pData->cSendBuffer+iLen,sizeof(pData->cSendBuffer)-pData->uSendAlloc-iLen-8,3000);
				if(iNum<0)
					{
					return -1-HTTP_MSG_CONNLOST;
					}

				uTime = SysTickCount();

				if(iNum==0)								// check timeout
					{
					   uDiff = uTime-uLast;
					if(uDiff<pData->uRequTimeout)
						{
						continue;
						}

					return -1-HTTP_MSG_CONNLOST;
					}

				iLen += iNum;
				iPos += iNum;
														// is the buffer 1/2 filled
				if(iLen>=(int)(sizeof(pData->cSendBuffer)-pData->uSendAlloc)/2)	
					{
					if(pData->pResponseCb)break;
					if(sizeof(pData->cSendBuffer)-pData->uSendAlloc-iLen<=12)
						{
						pData->uRequKeepAlive=0;
						return 413;						// request entity too large:
						}
					}
				}

			if(pData->pResponseCb)						// send data via callback
				{
				pData->cSendBuffer[iLen] = 0;
				iRet = pData->pResponseCb(pData,pData->cSendBuffer,iLen,iPos-iLen);
				if(iRet)iErr=iRet;
				}

			if(iLen<=0)return iErr;

			iLen = 0;
			}
		}


//******************** chunked data *******************************************


	pData->cSendBuffer[0] = '\r';
	pData->cSendBuffer[1] = '\n';

	memcpy(pData->cSendBuffer+2,pInput,iSize);

	iLen    = iSize+2;
	iNum	= iSize;
	iAll	= -2;
	iOff	= 0;
	iPos	= 0;
	iRest	= 0;


	for(;;)
		{
		if(iNum<=0)									// receive new data
			{
			SysThreadCheck();
			SysWatchdogTrigger(pData->hWatchdog);


			   iMax = sizeof(pData->cSendBuffer)-pData->uSendAlloc-iLen-8;
			if(iMax>32 && iRest<=0)iMax=32;

			   iNum = TcpGetNext(pData->hConnection,pData->cSendBuffer+iLen,iMax,3000);
			if(iNum<0)
				{
				return -1-HTTP_MSG_CONNLOST;
				}

			uTime = SysTickCount();

			if(iNum==0)								// check timeout
				{
				   uDiff = uTime-uLast;
				if(uDiff<pData->uRequTimeout)
					{
					continue;
					}

				return -1-HTTP_MSG_CONNLOST;
				}

			iLen += iNum;
			}


		if(iRest<=0)								// check chunk
			{
			iNum = iLen-iOff;
			iAdd = 0;
			
			if(iNum>=2)
			if(pData->cSendBuffer[iOff  ]=='\r')
			if(pData->cSendBuffer[iOff+1]=='\n')
				{
				iAdd = 2;

				if(iNum>=4)
				if(pData->cSendBuffer[iOff+2]=='\r')
				if(pData->cSendBuffer[iOff+3]=='\n')
					{
					iAdd = 4;
					}
				}

			iNum = iOff+iAdd;
													// exist a hex char
			if(pData->cSendBuffer[iNum]<'0' || pData->cSendBuffer[iNum]>'f'  || 
			  (pData->cSendBuffer[iNum]>'9' && pData->cSendBuffer[iNum]<'A') ||
			  (pData->cSendBuffer[iNum]>'F' && pData->cSendBuffer[iNum]<'a'))
				{
				if(pRead)*pRead=iSize;
				return 400;
				}

			for(iNum++;iNum+1<iLen;iNum++)			// seach for "\r\n" after length
				{
				if(pData->cSendBuffer[iNum  ]=='\r')
				if(pData->cSendBuffer[iNum+1]=='\n')break;

				if(pData->cSendBuffer[iNum  ]>='0' )
				if(pData->cSendBuffer[iNum  ]<='9' )continue;
				if(pData->cSendBuffer[iNum  ]>='a' )
				if(pData->cSendBuffer[iNum  ]<='f' )continue;
				if(pData->cSendBuffer[iNum  ]>='A' )
				if(pData->cSendBuffer[iNum  ]<='F' )continue;

				if(pRead)*pRead=iSize;
				return 400;
				}

			   iNum+=2;
			if(iNum>iLen)							// full chunck received ?
				{
				iNum=0;
				continue;
				}

			iRest = strtol(pData->cSendBuffer+iOff+iAdd,0,16);

			if(iRest==0)							// is it the last chunk
				{
				if(iNum+2<iLen                     ){iNum=0;continue;}
				if(pData->cSendBuffer[iNum  ]!='\r'){iNum=0;continue;}
				if(pData->cSendBuffer[iNum+1]!='\n'){iNum=0;continue;}

				pData->uRequContentLen   = iPos;


				if(pData->pResponseCb && iOff>0)	// send data via callback
					{
					pData->cSendBuffer[iOff] = 0;

					   iRet = pData->pResponseCb(pData,pData->cSendBuffer,iOff,iPos-iOff);
					if(iRet)iErr=iRet;
					}

				if(pData->pResponseCb)				// send end callback
					{
					   iRet = pData->pResponseCb(pData,pData->cSendBuffer+iOff,0,iPos);
					if(iRet)iErr=iRet;
					}

				if(pRead)
					{
					iAll += iNum-iOff+iAdd;
					*pRead= (iAll>iSize)? iSize:iAll;
					}

				return iErr;
				}

			iNum  -= iOff;
			iLen  -= iNum;
			iAll  += iNum;							// remove chunk from input data
			memmove(pData->cSendBuffer+iOff,pData->cSendBuffer+iOff+iNum,iLen-iOff);
			iNum   = iLen-iOff;
			iOff  += iRest;
			}


		iPos  += iNum;								// data received
		iAll  += iNum;
		iRest -= iNum;


		if(iRest<=0)								// end of chunk data
			{
			iPos += iRest;
			iAll += iRest;
			}

													// is the buffer less than half filled
		if(iLen<(int)(sizeof(pData->cSendBuffer)-pData->uSendAlloc)/2)		
			{
			iNum = (iRest<=0)? iLen-iOff:0;
			continue;
			}

		iNum = (iRest<0)? iLen+iRest:iLen;

		if(pData->pResponseCb)						// send data via callback
			{
			cTemp = pData->cSendBuffer[iNum];
			pData->cSendBuffer[iNum] = 0;
			iRet = pData->pResponseCb(pData,pData->cSendBuffer,iNum,iPos-iNum);
			iRet = 0;
			if(iRet)iErr=iRet;
			pData->cSendBuffer[iNum] = cTemp;

			if(iNum<iLen)
				{
				memmove(pData->cSendBuffer,pData->cSendBuffer+iNum,iLen-iNum);
				}

			iOff -= iNum;
			iLen -= iNum;
			iNum  = (iRest<=0)? iLen-iOff:0;
			continue;
			}

													// check buffer size
		if(sizeof(pData->cSendBuffer)-pData->uSendAlloc-iLen<=12)		
			{
			pData->cSendBuffer[iNum] = 0;
			pData->uRequContentLen   = iPos;
			pData->uRequKeepAlive    = 0;
			return 413;								// request entity too large:
			}

		iNum = (iRest<=0)? iLen-iOff:0;
		}



return iErr;
}

//*****************************************************************************
//*
//*		SendResponse
//*
//*****************************************************************************
//	Sends a http response via callback
//	pData		: pointer to the connection data
//	iCommand	: is the command of the request (see HTTP_CMD_???)
//	Returns an error code below zero for the request or the transmited bytes.
static int SendResponse(HttpData *pData)
{
int		iNum;
int		iMax;
int		iLen;
int		iRet;
int		iPos;


	iLen = CreateResponseHeader(pData->cSendBuffer,200,pData->uRequHttpVersion);

	if(pData->uResponseTime)									// set-modivied time
		{
		memcpy(pData->cSendBuffer+iLen,"Last-Modified: ",15);
		iLen+=15;
		iLen+=EncodeTime(pData->uResponseTime,pData->cSendBuffer+iLen);
		memcpy(pData->cSendBuffer+iLen,"\r\n"  ,2);
		iLen+=2;
		}


	iLen+=sprintf(pData->cSendBuffer+iLen,"Server: %s\r\nAccept-Ranges: bytes\r\n",HTTP_SERVER_STRING);


	if(pData->uResponseCache&~HTTP_CACHE_AGEMASK)				// set cache attributes
		{
		memcpy(pData->cSendBuffer+iLen,"Cache-Control: ",15);
		iLen+=15;

		if(pData->uResponseCache&HTTP_CACHE_NOCACHE)
			{
			memcpy(pData->cSendBuffer+iLen,"no-cache; ",10);
			iLen+=10;
			}

		if(pData->uResponseCache&HTTP_CACHE_PUBLIC)
			{
			memcpy(pData->cSendBuffer+iLen,"public; ",8);
			iLen+=8;
			}

		if(pData->uResponseCache&HTTP_CACHE_PRIVATE)
			{
			memcpy(pData->cSendBuffer+iLen,"public; ",8);
			iLen+=8;
			}

		if(pData->uResponseCache&HTTP_CACHE_NOSTORE)
			{
			memcpy(pData->cSendBuffer+iLen,"no-store; ",10);
			iLen+=10;
			}

		if(pData->uResponseCache&HTTP_CACHE_MAXAGE)
			{
			iLen+=sprintf(pData->cSendBuffer+iLen,"max-age=%u; ",pData->uResponseCache&HTTP_CACHE_AGEMASK);
			}
		}

	for(iPos=0;pData->cResponseCookie[iPos];)					// set new cookies
		{
		memcpy(pData->cSendBuffer+iLen,"Set-cookie: ",12);
		iLen += 12;
		iLen += HttpEncodeString(pData->cResponseCookie+iPos,pData->cSendBuffer+iLen,sizeof(pData->cSendBuffer)-pData->uSendAlloc-iLen);
		iPos += strlen(pData->cResponseCookie+iPos)+1;

		pData->cSendBuffer[iLen] = '\r';	iLen++;
		pData->cSendBuffer[iLen] = '\n';	iLen++;
		}

	if(pData->uRequKeepAlive)									// keep-alive mode for connection
		{
		memcpy(pData->cSendBuffer+iLen,"Connection: Keep-Alive\r\n",24);
		iLen+=24;
		}

	if(pData->cResponseType[0])									// set content-type
		{
		iLen+=sprintf(pData->cSendBuffer+iLen,"Content-Type: %s\r\n",pData->cResponseType);
		}



//******************** transmit plain *****************************************


	if(pData->iResponseLen>=0)
		{
		iLen += sprintf(pData->cSendBuffer+iLen,"Content-Length: %u\r\n\r\n",pData->iResponseLen);
		HTTP_LOG_RESP(("%.*s>>Header(%i)\n",iLen-2,pData->cSendBuffer,iLen))


		if(pData->uRequCommand==HTTP_CMD_HEAD)					// send only the header
			{
			pData->pResponseCb(pData,pData->cSendBuffer,0,0);

			   iNum = TcpPut(pData->hConnection,pData->cSendBuffer,iLen);
			if(iNum!=iLen)
				{
				HTTP_LOG_RESP((">>Error  TcpPut(%i)=%i\n",iLen,iNum))
				return -1000;
				}

			HTTP_LOG_RESP((">>TcpPut(%i)\n",iLen))

			return 0;
			}


		for(iMax=0;iMax<pData->iResponseLen;iLen=0)				// transmit data
			{
			for(;;)
				{
				SysThreadCheck();
				SysWatchdogTrigger(pData->hWatchdog);

				   iRet = pData->pResponseCb(pData,pData->cSendBuffer+iLen,sizeof(pData->cSendBuffer)-pData->uSendAlloc-iLen,iMax);
				if(iRet<0)
					{
					return iRet;
					}
				else if(iMax+iRet>pData->iResponseLen)
					{
					iRet = pData->iResponseLen-iMax;
					}

				if(iRet==0)										// end of data
					{
					break;
					}

				iLen += iRet;
				iMax += iRet;
																// is the buffer 1/2 filled
				if(iLen>(int)(sizeof(pData->cSendBuffer)-pData->uSendAlloc)/2)			
					{
					break;
					}
				}

			   iNum = TcpPut(pData->hConnection,pData->cSendBuffer,iLen);
			if(iNum!=iLen)
				{
				SysThreadCheck();
				SysWatchdogTrigger(pData->hWatchdog);

				pData->pResponseCb(pData,pData->cSendBuffer,0,iMax);
				HTTP_LOG_RESP((">>Error  TcpPut(%i)=%i\n",iLen,iNum))
				return -1000;
				}


			HTTP_LOG_RESP((">>TcpPut(%i)\n",iLen))

			if(iRet==0)
				{
				break;
				}
			}

		if(iMax>=pData->iResponseLen)							// was full file done
			{
			return iMax;
			}
		else{
			memset(pData->cSendBuffer,0,sizeof(pData->cSendBuffer)-pData->uSendAlloc);
			}

		while(iMax<pData->iResponseLen)							// fill data with zero
			{
			   iRet = pData->iResponseLen-iMax;
			if(iRet>(int)(sizeof(pData->cSendBuffer)-pData->uSendAlloc))
				{
				iRet=sizeof(pData->cSendBuffer)-pData->uSendAlloc;
				}

			iMax += iRet;

			   iNum = TcpPut(pData->hConnection,pData->cSendBuffer,iRet);
			if(iNum!=iRet)
				{
				HTTP_LOG_RESP((">>Error  TcpPut(%i)=%i\n",iLen,iNum))
				return -1000;
				}
			  
			HTTP_LOG_RESP((">>TcpPut(%i)\n",iLen))
			}

		return iMax;
		}



//******************** transmit chunked **************************************


	memcpy(pData->cSendBuffer+iLen,"Transfer-Encoding: chunked\r\n\r\n",30);
	iLen += 30;
	iMax  =  0;

	HTTP_LOG_RESP(("%.*s>>Header(%i)\n",iLen-2,pData->cSendBuffer,iLen))


	if(pData->uRequCommand==HTTP_CMD_HEAD)						// send only the header
		{
		memcpy(pData->cSendBuffer+iLen,"00\r\n\r\n\r\n",8);
		iLen += 8;

		   iNum = TcpPut(pData->hConnection,pData->cSendBuffer,iLen);
		if(iNum!=iLen)
			{
			SysThreadCheck();
			SysWatchdogTrigger(pData->hWatchdog);

			pData->pResponseCb(pData,pData->cSendBuffer,0,iMax);
			HTTP_LOG_RESP((">>Error  TcpPut(%i)=%i\n",iLen,iNum))
			return -1000;
			}

		HTTP_LOG_RESP((">>TcpPut(%i)\n",iLen))

		return 0;
		}


	for(;;)														// chunked response
		{
		for(iNum=0;;)
			{
			SysThreadCheck();
			SysWatchdogTrigger(pData->hWatchdog);

			   iRet = pData->pResponseCb(pData,pData->cSendBuffer+iLen+iNum+8,sizeof(pData->cSendBuffer)-pData->uSendAlloc-iLen-iNum-12,iMax+iNum);
			if(iRet<0)
				{
				return iRet;
				}

			if(iRet==0)											// end of data
				{
				break;
				}

			iNum += iRet;
																// is the buffer 1/2 filled
			if(iLen+iNum>(int)(sizeof(pData->cSendBuffer)-pData->uSendAlloc)/2)			
				{
				break;
				}
			}

		sprintf(pData->cSendBuffer+iLen,"%06x",iNum);

		pData->cSendBuffer[iLen+6     ] = '\r';
		pData->cSendBuffer[iLen+7     ] = '\n';
		pData->cSendBuffer[iLen+8+iNum] = '\r';
		pData->cSendBuffer[iLen+9+iNum] = '\n';

		iMax += iNum;
		iLen += iNum+10;

		if(iRet==0)												// last chunk
			{
			if(iNum>0)
				{
				memcpy(pData->cSendBuffer+iLen,"0000\r\n\r\n",8);
				iLen+=8;
				}
			else{
				memcpy(pData->cSendBuffer+iLen,"\r\n",2);
				iLen+=2;
				}
			}

		   iNum = TcpPut(pData->hConnection,pData->cSendBuffer,iLen);
		if(iNum!=iLen)
			{
			SysThreadCheck();
			SysWatchdogTrigger(pData->hWatchdog);

			pData->pResponseCb(pData,pData->cSendBuffer,0,iMax);
			HTTP_LOG_RESP((">>Error  TcpPut(%i)=%i\n",iLen,iNum))
			return -1000;
			}

		HTTP_LOG_RESP((">>TcpPut(%i)\n",iLen))

		if(iRet==0)												// end of data
			{
			break;
			}

		iLen = 0;
		}


return iMax;
}

//*****************************************************************************
//*
//*		GotoResponse
//*
//*****************************************************************************
//	Sends a goto response
//	pData		: pointer to the connection data
//	iErr		: ist the http error code (300...307)
//	Returns 0 if ok or the exit code HTTP_MSG_????
static int GotoResponse(HttpData *pData,int iErr)
{
char   *pPtr;
int		iMax;
int		iNum;
int		iLen;


	iMax = sizeof(pData->cSendBuffer)-pData->uSendAlloc;
	pPtr = pData->cSendBuffer+iMax/2;
	HttpEncodeString(pData->cRequFile,pPtr,iMax/4);

	iNum  = sprintf(pPtr+iMax/4,"<HTML><HEAD><TITLE>%u document moved</TITLE>\n"
								 "<meta http-equiv=\"refresh\" content=\"0; URL=%s\">\n"
								 "</HEAD><BODY><H1>Found</H1>The document has moved"
								 "<A HREF=\"%s\">here</A><P><HR></BODY></HTML>\n",iErr,pPtr,pPtr);

	iLen  = CreateResponseHeader(pData->cSendBuffer,iErr,pData->uRequHttpVersion,0);
	iLen += sprintf(pData->cSendBuffer+iLen,"Connection: %s\r\n"  ,(pData->uRequKeepAlive)? "Keep-Alive":"close");
	iLen += sprintf(pData->cSendBuffer+iLen,"Location: %s\r\n",pPtr);
	iLen += sprintf(pData->cSendBuffer+iLen,"Cache-Control: no-cache; min-fresh=0;  %s\r\n",pPtr);
	iLen += sprintf(pData->cSendBuffer+iLen,"Content-Length: %u\r\n\r\n",iNum);


	HTTP_LOG_RESP(("%.*s",iLen,pData->cSendBuffer))


	if(TcpPut(pData->hConnection,pData->cSendBuffer,iLen)!=iLen)
		{
		return HTTP_MSG_CONNLOST;
		}

	if(TcpPut(pData->hConnection,pPtr+iMax/4,iNum)!=iNum)
		{
		return HTTP_MSG_CONNLOST;
		}


return 0;
}

//*****************************************************************************
//*
//*		AuthResponse
//*
//*****************************************************************************
//	Sends a goto response
//	pData		: pointer to the connection data
//	iErr		: ist the http error code (300...307)
//	Returns 0 if ok or the exit code HTTP_MSG_????
static int AuthResponse(HttpData *pData,int iErr)
{
int			iLen;


	#if HTTP_HAVE_AUTH

	if(!pData->pServer->cAuthOpaque[0])
		{
		SysSemaphoreLock(pData->pServer->hSemaphore);

		if(!pData->pServer->cAuthOpaque[0])
			{
			CreateRand128(pData->pServer->cAuthOpaque);
			}

		SysSemaphoreUnlock(pData->pServer->hSemaphore);
		}


	CreateRand128(pData->cAuthNonce);


	iLen  = CreateResponseHeader(pData->cSendBuffer,iErr,pData->uRequHttpVersion,0);
	iLen += sprintf(pData->cSendBuffer+iLen,"WWW-Authenticate: Digest realm=\"%s\" nonce=\"",pData->pServer->cAuthRealm);

	memcpy(pData->cSendBuffer+iLen,pData->cAuthNonce				,32);	iLen+=32;
	memcpy(pData->cSendBuffer+iLen,"\" opaque=\""					,10);	iLen+=10;
	memcpy(pData->cSendBuffer+iLen,pData->pServer->cAuthOpaque		,32);	iLen+=32;
	memcpy(pData->cSendBuffer+iLen,"\"\r\nContent-Length: 0\r\n\r\n",24);	iLen+=24;

	HTTP_LOG_RESP(("%.*s",iLen,pData->cSendBuffer))

	if(TcpPut(pData->hConnection,pData->cSendBuffer,iLen)!=iLen)
		{
		return HTTP_MSG_CONNLOST;
		}

	#else
															// return 403 forbibben
	iLen  = CreateResponseHeader(pData->cSendBuffer,403,pData->uRequHttpVersion,1);

	if(TcpPut(pData->hConnection,pData->cSendBuffer,iLen)!=iLen)
		{
		return HTTP_MSG_CONNLOST;
		}

	#endif


return 0;
}

//*****************************************************************************
//*
//*		ConnectionProc
//*
//*****************************************************************************
//	Function which processes the input commands
//	pParam		: pointer to the connection data
//	Returns the cause off disconnect action
static void _cdecl ConnectionProc(void *pParam)
{
HttpData       *pData=(HttpData*)pParam;
HttpDataCB		pCallback;
HttpServerData *pServer;
char		   *pNext;
char		   *pPath;
unsigned		uDiff;
int				iFrom;
int				iLine;
int				iBack;
int				iSize;
int				iExit;
int				iPut;
int				iRun;
int				iOff;
int				iMax;
int				iLen;
int				iPos;
int				iEnd;
int				iErr;
int				iRet;





	iRun				= 1;
	iRet				= 0;
	iErr				= 0;
	iEnd				= 0;
	iPos				= 0;
	iLine				= 0;
	iFrom				= 0;
	iBack				= 0;
	pServer				= pData->pServer;
	pData->hWatchdog	= SysWatchdogRegister(HTTP_WDOG_TIME,"HttpConnection");


	HTTP_LOG_CONN(("\n-----------------------------------------------\n"))
	HTTP_LOG_CONN(("CONNECT: %s:%u at http(%i)",pData->cConnIp,pData->usConnPort,pServer->iCurConnections))
	HTTP_LOG_CONN(("\n-----------------------------------------------\n"))


	while(iRun)
		{
		SysThreadCheck();
		SysWatchdogTrigger(pData->hWatchdog);
															// server shutdown
		if(pServer->iShutdown)
			{
			iExit = HTTP_MSG_SHUTDOWN;
			break;
			}

		if(iEnd>0)											// remove full lines from buffer
			{
			iLen = pData->uRecvSize-iEnd;
				   pData->uRecvSize=iLen;

			if(iLen>0)
				{
				memmove(pData->cRecvBuffer,pData->cRecvBuffer+iEnd,iLen);
				}

			iEnd = 0;
			}

		pNext = pData->cRecvBuffer+pData->uRecvSize;
		iMax  = sizeof(pData->cRecvBuffer)-pData->uRecvSize-8;

		if(iMax<=0)											// buffer too small
			{
			iExit = HTTP_MSG_BUFFEROVERFLOW;
			iBack = 413;									// request entity too large
			break;
			}


		if(pData->uRecvSize==0 || iPos>=iEnd)
			{
			SysThreadCheck();
			SysWatchdogTrigger(pData->hWatchdog);

			   iLen = TcpGetNext(pData->hConnection,pNext,iMax,3000);
			if(iLen<0)										// connection lost
				{
				iExit = HTTP_MSG_CONNLOST;
				break;
				}

			if(iLen==0)										// no data received
				{
				   uDiff = SysTickCount()-pData->uRequLast;
				if(uDiff<pData->uRequTimeout)continue;
				if(uDiff<pData->uRecvTimeout && pData->uRecvSize==0 && iLine==0)
					{
					continue;
					}

				iExit = HTTP_MSG_TIMEOUT;
				break;
				}

			pData->uRequLast  = SysTickCount();
			pData->uRecvSize += iLen;
			pData->uTransmitActiv = 1;
			}


		iFrom = 0;
		iMax  = pData->uRecvSize;
		pNext =	pData->cRecvBuffer;

		for(iPos=0;iPos<iMax;iPos++)						// find end of line
			{
			SysThreadCheck();
			SysWatchdogTrigger(pData->hWatchdog);

			if(pNext[iPos  ]!='\r')continue;
			if(iPos+1>=iMax       )break;
			if(pNext[iPos+1]!='\n')continue;

			iEnd = iPos+2;

			if(iPos>iFrom)									// decode next line
				{
				pNext[iPos] = 0;

				if(iLine==0)
					{
					   iErr = ScanRequestCmd  (pData,pNext+iFrom,iPos-iFrom);
					}
				else{
					   iRet = ScanRequestParam(pData,pNext+iFrom,iPos-iFrom);
					if(iRet)iErr=iRet;
					}

				iFrom  = iEnd;
				iLine += 1;
				iPos  += 1;

				continue;
				}

			//**************** execute command ********************************

			iPos += 1;
			pPath = pData->cRequPath;
			if(*pPath=='/')pPath++;
			iSize = strlen(pPath);

			pData->pResponseCb		  =  0;
			pData->cResponseCookie[0] =  0;
			pData->cResponseCookie[1] =  0;
			pData->uResponseTime      =  0;
			pData->uResponseCache	  =  0;
			pData->iResponseLen       = -1;

			for(iOff=iSize;;iOff--)							// get content-type
				{
				if(iOff<=0 || pPath[iOff]=='/')
					{
					pData->cResponseType[0]=0;
					pData->pRequType="";
					break;
					}

				if(pPath[iOff]!='.')continue;

				pData->pRequType = pPath+iOff+1;

				GetContentType(pData->pRequType,pData->cResponseType);
				break;
				}


			if(iSize+pServer->iFilePathLen>=HTTP_MAX_PATH)	// url to file path
				{
				pData->cRequFile[0] = '*';
				pData->cRequFile[1] = 0;
				}
			else{
				memcpy(pData->cRequFile,pServer->cFilePath,pServer->iFilePathLen);
				memcpy(pData->cRequFile+pServer->iFilePathLen,pPath,iSize+1);

				if(iSize>0 && pPath[iSize-1]=='/')
					{
					pData->cRequFile[pServer->iFilePathLen+iSize-1]=0;
					}
				}


			if(!iErr && pServer->pCbRequest)				// use request callback
				{
				iErr = pServer->pCbRequest(pData,pData->uRequCommand,pData->cRequHost,pData->cRequPath,pData->cRequParam,pData->uRequPathLength,pData->uRequParamLength);
				}



			//**************** decode commands ********************************


			if(pData->uRequCommand==HTTP_CMD_POST || pData->uRequCommand==HTTP_CMD_PUT)
				{
				SysThreadCheck();
				SysWatchdogTrigger(pData->hWatchdog);

				pCallback = pData->pResponseCb;

				if(iErr>=0 && iErr<=3)						// simple post request
					{
					if(pData->uRequContentLen==0)			// no content length
						{
						pData->uRequKeepAlive = 0;
						iErr  = 411;						// length required
						}
					else if(iErr<=2)
						{
						pData->pResponseCb = 0;				// receive into internal buffer
						}
					}

				if(iErr>=100)								// http error
					{
					pData->pResponseCb = 0;

					if(GetPostData(pData,pData->cRecvBuffer+iEnd,pData->uRecvSize-iEnd,&iLen))
						{
						pData->uRequKeepAlive = 0;
						}
					}
				else if(iErr>=0 || iErr<=3)					// receive post/put data
					{
					iPut  = iErr;
					iErr  = GetPostData(pData,pData->cRecvBuffer+iEnd,pData->uRecvSize-iEnd,&iLen);
					iEnd += iLen;
					iPos += iLen;

					if(iRet)iErr=iRet;

					if(iErr<0)
						{
						iExit = -iErr-1;
						iRun  =  0;
						break;
						}

					if(iErr==0)								// default post response
						{
						if(iPut==1)
							{
							pData->pResponseCb = pCallback;
							iErr = 1;
							}
						else if(iPut==2)
							{
							iErr = pCallback(pData,pData->cSendBuffer,pData->uRequContentLen,0);
							}
						else{
							iErr = DefaultResponse(pData,pData->uRequCommand);
							}
						}
					}
				else{
					pData->uRequKeepAlive = 0;
					iErr  = 500;							// internal server error
					}

				}
			else if(iErr==0)								// do an default response
				{
				if(pData->uRequCommand==HTTP_CMD_GET || pData->uRequCommand==HTTP_CMD_HEAD)
					{
					iErr = DefaultResponse(pData,pData->uRequCommand);
					}
				else{
					iErr = 405;								// method not allowed
					}
				}



			//**************** generate response data *************************


			SysThreadCheck();
			SysWatchdogTrigger(pData->hWatchdog);


			HTTP_LOG_RESP(("---------------- %u ----------------\n",iErr))


			if(iErr==1)										// response via callback
				{
				   iLen = SendResponse(pData);
				if(iLen<0)
					{
					iExit = (iLen==-1000)? HTTP_MSG_CONNLOST:HTTP_MSG_USERBREAK;
					iRun  =  0;
					break;
					}

				pData->uRequLast = SysTickCount();
				}
			else if(iErr>=300 && iErr<=399)					// goto other webside
				{
				   iRet = GotoResponse(pData,iErr);
				if(iRet)
					{
					iExit = iRet;
					iRun  = 0;
					break;
					}
				}
			else if(iErr==401)								// send authorization request
				{
				   iRet = AuthResponse(pData,iErr);
				if(iRet)
					{
					iExit = iRet;
					iRun  = 0;
					break;
					}
				}
			else if(iErr>=100)								// send error response
				{
				iLen  = CreateResponseHeader(pData->cSendBuffer,(iErr>=100)? iErr:500,pData->uRequHttpVersion,1);

				HTTP_LOG_RESP(("%.*s\n",iLen,pData->cSendBuffer))

				if(TcpPut(pData->hConnection,pData->cSendBuffer,iLen)!=iLen)
					{
					iExit = HTTP_MSG_CONNLOST;
					iRun  = 0;
					break;
					}
				}


			SysThreadCheck();
			SysWatchdogTrigger(pData->hWatchdog);


			if(pData->hFile)								// close an open file
				{
				SysClose(pData->hFile);
				pData->hFile = 0;
				}

			iLine = 0;
			iErr  = 0;

			if(pData->uRequKeepAlive==0)					// disconnect after the request
				{
				iExit = HTTP_MSG_DISCONNECT;
				iRun  =  0;
				break;
				}

			pData->uTransmitActiv = 0;
			}
		}


//******************** close connection ***************************************


	SysThreadCheck();
	SysWatchdogTrigger(pData->hWatchdog);

	pData->uTransmitActiv = 0;


	if(pData->hFile)										// close an open file
		{
		SysClose(pData->hFile);
		pData->hFile = 0;
		}


	if(iBack)												// send an error message to client
		{
		iLen = CreateResponseHeader(pData->cRecvBuffer,iBack,pData->uRequHttpVersion,1);
		HTTP_LOG_RESP(("%.*s",iLen,pData->cRecvBuffer))
		TcpPut(pData->hConnection ,pData->cRecvBuffer,iLen);
		}


	if(pServer->pCbDisconnect)								// disconnect callback
		{
		pServer->pCbDisconnect(pData,pData->cConnIp,pData->usConnPort,iExit);
		}


	HTTP_LOG_CONN(("\n-----------------------------------------------\n"))
	HTTP_LOG_CONN(("\nDISCONNECT: %s:%u  (reason=%i)",pData->cConnIp,pData->usConnPort,iExit))
	HTTP_LOG_CONN(("\n-----------------------------------------------\n\n\n"))

	TcpClose(pData->hConnection);

	SysSemaphoreLock(pServer->hSemaphore);

	pData->hConnection = 0;
	pServer->iCurConnections--;

	SysSemaphoreUnlock(pServer->hSemaphore);
	SysWatchdogUnregister(pData->hWatchdog);

	FreeHttpData(pData);

	if(iCloseFlag)											// waits the listen thread for a closing
		{
		SysSemaphoreUnlock(hCloseSem);	
		}


}

//*****************************************************************************
//*
//*		CloseOldestClient
//*
//*****************************************************************************
//	Closes the oldes client connection and allocate new client data.
//	hWdog		: Is the watchdog of the current thread
//	uTime		: Is the current tick count
//	Returns 1 if a connection was closed.
static HttpData *CloseOldestClient(SysWdog hWdog)
{
HttpData   *pData;
HttpData   *pConn;
unsigned	uTime;
unsigned 	uDelta;
unsigned	uDiff;
unsigned	uTurn;
unsigned	uPos;



	SysThreadCheck();

	pData = 0;


	for(uTurn=0;uTurn<4;uTurn++)
		{
		uTime	= SysTickCount();
		uDelta	= 0;
		pConn	= 0;


		#if HTTP_STATIC_MEMORY

		for(uPos=0;uPos<HTTP_STATIC_COUNT;uPos++)		// find oldest connection
			{
			if(aConnAlloc  [uPos]==0            )continue;
			if(aConnections[uPos].uTransmitActiv)continue;

			   uDiff = uTime-aConnections[uPos].uRequLast;
			if(uDiff>uDelta)pConn=aConnections+uPos;
			}

		#else
		
		SysSemaphoreLock(hGlobalSem);

		for(HttpEntry *pClient=pFirstClient;pClient;pClient=pClient->pNextClient)
			{
			if(pClient->sHttpData.uTransmitActiv)continue;

			   uDiff = uTime-pClient->sHttpData.uRequLast;
			if(uDiff>uDelta)pConn=&pClient->sHttpData;
			}

		SysSemaphoreUnlock(hGlobalSem);
		
		#endif 	


		if(!pConn)										// no connection found
			{
			iCloseFlag = 1;
			SysSemaphoreLockWait(hCloseSem,100);
			iCloseFlag = 0;
			continue;
			}


		if(uDiff>4000 && uTurn>=2)
			{
			TcpShutdown(pConn->hConnection);
			}

		pConn->uRequLast	= uTime-0x40000000;
		pConn->uRequTimeout	= 0;
		iCloseFlag			= 1;						// the listen thread waits for closing
		pData				= 0;
				

		for(uPos=0;uPos<8;uPos++)						// wait until a entry is free
			{
			SysSemaphoreLockWait(hCloseSem,500);	
			SysThreadCheck();
			SysWatchdogTrigger(hWdog);

			   pData = AllocHttpData();
			if(pData)break;
			}

		iCloseFlag = 0;

		break;
		}



return pData;
}


//*****************************************************************************
//*
//*		ListenProc
//*
//*****************************************************************************
//	Function which install the server socket and waits for new connections
//	pParam		: pointer to the config data of the server
static void _cdecl ListenProc(void *pParam)
{
HttpServerData *pServer=(HttpServerData*)pParam;
HttpData	   *pData;
SysWdog			hWdog;
unsigned short	usPort;
unsigned		uTime;
TcpHandle		hClient;
IpAddr			iAddr;


	
	
	hWdog = SysWatchdogRegister(HTTP_WDOG_TIME,"HttpListen");


	while(!pServer->iShutdown)
		{
		SysThreadCheck();
		SysWatchdogTrigger(hWdog);

			hClient = TcpAcceptWait(pServer->hServer,&iAddr,&usPort,3000);
		if(!hClient)										// wait for client connection
			{
			continue;
			}


		SysThreadCheck();

		uTime = SysTickCount();

															// check connection count
		if(pServer->iCurConnections>=pServer->iMaxConnections)
			{
				pData = CloseOldestClient(hWdog);
			if(!pData)
				{
				TcpPut  (hClient,"HTTP/1.1 503 Service Unavailable\r\n\r\n",36);
				TcpClose(hClient);
				continue;
				}
			}
		else{
				pData = AllocHttpData();
			if(!pData)										// no free enty
				{
				SysThreadCheck();

					pData = CloseOldestClient(hWdog);
				if(!pData)
					{
					TcpPut  (hClient,"HTTP/1.1 501 Internal Server Error\r\n\r\n",38);
					TcpClose(hClient);
					continue;
					}
				}
			}

		
		SysThreadCheck();
		
		memset(pData,0,sizeof(HttpData));
		pData->hFile			= 0;
		pData->pServer			= pServer;
		pData->usConnPort		= usPort;
		pData->hConnection		= hClient;
		pData->uRequLast		= uTime;
		pData->uRequHttpVersion	= 0x0101;
		pData->uRequTimeout		= HTTP_REQU_TIMEOUT;
		pData->uRecvTimeout		= HTTP_RECV_TIMEOUT;
		pData->uUserOffset		= pServer->uUserOffset;

		IpToString(&iAddr,pData->cConnIp);

		
		if(pServer->pCbConnect)								// connect callback
			{
			SysThreadCheck();
			SysWatchdogTrigger(hWdog);

			if(!pServer->pCbConnect(pData,pData->cConnIp,pData->usConnPort))
				{
				TcpPut(pData->hConnection,"HTTP/1.1 403 Forbidden\r\n\r\n",26);
				TcpClose(hClient);
				FreeHttpData(pData);
				continue;
				}
			}


		SysSemaphoreLock(pServer->hSemaphore);				// start connection thread

		if(SysThreadStart(ConnectionProc,HTTP_STACK_SIZE,pData,pServer->iPriority|SYS_GLOBAL,"HttpConnection"))
			{
			pServer->iCurConnections++;
			}
		else{
			TcpPut  (hClient,"HTTP/1.1 501 Internal Server Error\r\n\r\n",38);
			TcpClose(hClient);
			FreeHttpData(pData);
			}

		SysSemaphoreUnlock(pServer->hSemaphore);
		}



	while(pServer->iCurConnections>0)						// wait until all connections are closed
		{
		SysWatchdogTrigger(hWdog);
		SysThreadCheck();
		SysSleep(100);
		}

	SysSemaphoreLock(pServer->hSemaphore);					// close server socket
	TcpClose(pServer->hServer);
	pServer->hServer = NULL;
	SysSemaphoreUnlock(pServer->hSemaphore);

	SysSemaphoreDelete(pServer->hSemaphore);
	pServer->hSemaphore = 0;
	pServer->iRunning   = 0;

	SysWatchdogUnregister(hWdog);

}


//*****************************************************************************
//*
//*		HttpDaemon
//*
//*****************************************************************************
//	Starts the http server
//	pConfig		: pointer to the config data for the server
//	Returns 0 it daemon was started, or an error code greather zero
//					HTTP_ERR_NONE         0   no error
//					HTTP_ERR_SOCKET       1   socket error
//					HTTP_ERR_ISRUNNING    2   nmb server is allready running
//					HTTP_ERR_WRONGPATH    4   the file path is wrong
//					HTTP_ERR_WRONGIP      5   the ip address is wrong
//					HTTP_ERR_INIT         6   initialization error
HTTP_API int HttpDaemon(HttpConfigData *pConfig)
{
char	   *pPath;
IpAddr		iAddr;
int			iAttr;
int			iPos;



	if(iServerRunning)										// is the server allready running
		{
		return HTTP_ERR_ISRUNNING;
		}

	if(!hHttpHeap)											// create heap for server
		{
			hHttpHeap = SysHeapCreate(HTTP_MEMORY_MAX);
		if(!hHttpHeap)return HTTP_ERR_INIT;
		}

	if(!hGlobalSem)											// global semaphore
		{
		hGlobalSem = SysSemaphoreCreate();
		}

	if(!hCloseSem)											// signal for closing
		{
		hCloseSem = SysSemaphoreCreate();
		}

	memset(&sHttpServer,0,sizeof(sHttpServer));


	if(!IpAddress(&iAddr,pConfig->cIpAddress,NULL))			// convert ip-text to address
		{
		return HTTP_ERR_WRONGIP;
		}

	   pPath = pConfig->cFilePath;
	   iAttr = SysGetAttributes(pPath);
	if(iAttr<0 || !(iAttr&SYS_DIR) || !*pPath)				// check server path
		{
		return HTTP_ERR_WRONGPATH;
		}

		sHttpServer.hSemaphore = SysSemaphoreCreate();
	if(!sHttpServer.hSemaphore)								// create server semaphore
		{
		return HTTP_ERR_INIT;
		}

		sHttpServer.hServer = TcpOpen(&iAddr,pConfig->usTcpPort,TCP_SERVER,0);
	if(!sHttpServer.hServer)								// check server socket
		{
		SysSemaphoreDelete(sHttpServer.hSemaphore);
		return HTTP_ERR_SOCKET;
		}

	sHttpServer.uUserOffset		= pConfig->uUserOffset;		// start offest for user parameters (set to zero)
	sHttpServer.usTcpPort		= pConfig->usTcpPort;		// is the tcp port for the server
	sHttpServer.iPriority		= pConfig->iPriority;		// taks priority for client connections
	sHttpServer.iMaxConnections	= pConfig->iMaxConnections;	// maximum count of  client connections
	sHttpServer.iCurConnections	= 0;						// current count of  client connections

	if(!pConfig->iMaxConnections)
		{
		pConfig->iMaxConnections = 16;
		}

	#if HTTP_STATIC_MEMORY

	if(pConfig->iMaxConnections>HTTP_STATIC_COUNT)			// is the connection count higher than mem count
		{
		pConfig->iMaxConnections = HTTP_STATIC_COUNT;
		}

	#endif

	if(!pConfig->usTcpPort)
		{
		pConfig->usTcpPort = 80;
		}

	if(!pConfig->cDefaultExt[0])							// default file name for access to paths
		{
		strcpy(sHttpServer.cDefaultExt,"/index.html");
		}
	else{
		sHttpServer.cDefaultExt[0] = '/';
		strcpy(sHttpServer.cDefaultExt+1,pConfig->cDefaultExt);
		}

	sHttpServer.iDefaultExtLen = strlen(sHttpServer.cDefaultExt);


	#if HTTP_HAVE_AUTH

	if(!pConfig->cAuthRealm[0])								// is the realm value for the authorization
		{
		strcpy(sHttpServer.cAuthRealm,"az_http@host.com");
		}
	else{
		strcpy(sHttpServer.cAuthRealm,pConfig->cDefaultExt);
		}

	#endif


	for(iPos=0;iPos<pPath[iPos];iPos++)
		{
		if(pPath[iPos]!='\\')
				sHttpServer.cFilePath[iPos] = pPath[iPos];
		else	sHttpServer.cFilePath[iPos] = '/';
		}

	if(sHttpServer.cFilePath[iPos-1]!='/')					// add a slash at the end
		{
		sHttpServer.cFilePath[iPos] = '/';
		iPos++;
		}

	sHttpServer.iRunning		= 1;
	sHttpServer.iFilePathLen	= iPos;
	sHttpServer.pCbRequest		= pConfig->pCbRequest;
	sHttpServer.pCbConnect		= pConfig->pCbConnect;
	sHttpServer.pCbPassword		= pConfig->pCbPassword;
	sHttpServer.pCbDisconnect	= pConfig->pCbDisconnect;



															// start listen thread
	if(!SysThreadStart(ListenProc,2048,&sHttpServer,sHttpServer.iPriority|SYS_GLOBAL,"HttpListen"))
		{
		SysSemaphoreLock(sHttpServer.hSemaphore);
		TcpClose(sHttpServer.hServer);
		sHttpServer.hServer = NULL;
		SysSemaphoreUnlock(sHttpServer.hSemaphore);

		SysSemaphoreDelete(sHttpServer.hSemaphore);
		sHttpServer.hSemaphore = NULL;


		return HTTP_ERR_INIT;
		}


	iServerRunning = 1;



return HTTP_ERR_NONE;
}

//*****************************************************************************
//*
//*		HttpStopDaemon
//*
//*****************************************************************************
//	Stop the http server
//	pConfig		: pointer to the config data for the server
//	Returns 0 it daemon was started, or an error code greather zero
//					HTTP_ERR_NONE         0   no error
//					HTTP_ERR_INIT         6   initialization error
//					HTTP_ERR_NOTSTOPED    7   the server could not stopped
HTTP_API int HttpStopDaemon()
{
int		iTurn;

	if(!iServerRunning)										// is the server allready running
		{
		return HTTP_ERR_ISNOTRUNNING;
		}

	SysSemaphoreLock(sHttpServer.hSemaphore);
	sHttpServer.iShutdown = 1;
	if(sHttpServer.hServer)TcpShutdown(sHttpServer.hServer);
	SysSemaphoreUnlock(sHttpServer.hSemaphore);

	for(iTurn=0;;iTurn++)									// wait until server was stopped
		{
		if(iTurn>10*20)return HTTP_ERR_NOTSTOPED;
		if(!sHttpServer.iRunning)
		if(!sHttpServer.iCurConnections)break;
		SysSleep(100);
		}


	iServerRunning = 0;


return HTTP_ERR_NONE;
}

//*****************************************************************************
//*
//*		HttpTransmitGoto
//*
//*****************************************************************************
//	Used in request callback for moving to an other url
//	hConn		: pointer to the connection data
//	pUrl		: is the new url  (i.e: http://www.google.at/xyz.html")
//	pParam		: will be added behind the url (i.e.: "url:param")
//	iDissconect	: should the connection be closed after that request
//	Returns 307 "Moved temporarily" or 500 if buffer too small
HTTP_API int HttpTransmitGoto(HttpConnection hConn,const char *pUrl,const char *pParam,int iDissconect)
{
HttpData   *pData=(HttpData*)hConn;
int			iLen;
int			iAdd;



	if(iDissconect)pData->uRequKeepAlive=0;

	iLen = strlen(pUrl);

	if(iLen>=sizeof(pData->cRequFile))
		{
		return 500;											// internal server error
		}

	memcpy(pData->cRequFile,pUrl,iLen+1);



	if(pParam && *pParam)									// add the parameter
		{
		iAdd = strlen(pParam);

		if(iLen+1+iAdd>=sizeof(pData->cRequFile))
			{
			return 500;										// internal server error
			}

		memcpy(pData->cRequFile+1,pParam,iAdd+1);
		pData->cRequFile[iLen] = ':';
		}


return 307;
}


//*****************************************************************************
//*
//*		HttpTransmitFile
//*
//*****************************************************************************
//	Used in request callback for returning a other file than the request url
//	hConn		: pointer to the connection data
//	pFilePath	: is the new file path
//	Returns 0 for the default response
HTTP_API int HttpTransmitFile(HttpConnection hConn,const char *pFilePath)
{
HttpData   *pData=(HttpData*)hConn;
int			iLen;


	iLen = strlen(pFilePath);

	if(iLen>=sizeof(pData->cRequFile))
		{
		return 500;											// internal server error
		}

	memcpy(pData->cRequFile,pFilePath,iLen+1);


return 0;
}

//*****************************************************************************
//*
//*		HttpTransmitMemory
//*
//*****************************************************************************
//	Used in request callback for returning a memory block
//	hConn		: pointer to the connection data
//	pMemory		: is the pointer to the memory
//	iSize		: is the size of the memory
//	Returns 1 for a callback response
HTTP_API int HttpTransmitMemory(HttpConnection hConn,const void *pMemory,int iSize)
{
HttpData   *pData=(HttpData*)hConn;


	if(!pMemory)iSize=0;

	pData->iMemoryPos	= 0;
	pData->pMemoryPtr	= (char*)pMemory ;
	pData->pResponseCb  = CallbackReadMem;
	pData->iResponseLen = iSize;


return 1;
}

//*****************************************************************************
//*
//*		HttpTransmitCallback
//*
//*****************************************************************************
//	Used in request callback for returning a data via a callback
//	hConn		: pointer to the connection data
//	pCallback	: is the callback function for transmiting data
//	Returns 1 for a callback response
HTTP_API int _cdecl HttpTransmitCallback(HttpConnection hConn,HttpDataCB pCallback)
{
HttpData   *pData=(HttpData*)hConn;


	if(!pCallback)return 0;

	pData->pResponseCb  =  pCallback;
	pData->iResponseLen = -1;


return 1;
}

//*****************************************************************************
//*
//*		HttpPostMemory
//*
//*****************************************************************************
//	Used in request callback for geting a memory block in POST/PUT commands
//	hConn		: pointer to the connection data
//	pMemory		: is the pointer to the memory
//	iSize		: is the size of the memory
//	Returns 1 for a callback response
HTTP_API int HttpPostMemory(HttpConnection hConn,void *pMemory,int iSize)
{
HttpData   *pData=(HttpData*)hConn;


	if(!pMemory)iSize=0;

	pData->iMemoryPos	= 0;
	pData->iMemoryMax	= iSize;
	pData->pMemoryPtr	= (char*)pMemory ;
	pData->pResponseCb  = CallbackWriteMem;


return 1;
}

//*****************************************************************************
//*
//*		HttpPostCallback
//*
//*****************************************************************************
//	Used in request callback for geting data via callback in POST/PUT commands
//	hConn		: pointer to the connection data
//	pCallback	: is the callback function for getting data
//	iRecall		: if this parameter isn't zero the call back will be called
//				  until all data is tranfered, in the oher case the data
//				  is stored in a internal buffer and the function is called
//				  only one times.
//	Returns 3 for a recall callback or 2 for a single callback.
HTTP_API int HttpPostCallback(HttpConnection hConn,HttpDataCB pCallback,int iRecall)
{
HttpData   *pData=(HttpData*)hConn;


	if(!pCallback)return 0;

	pData->pResponseCb  = pCallback;


return (iRecall)? 2:3;
}

//*****************************************************************************
//*
//*		HttpReserveUserParams
//*
//*****************************************************************************
//	Reserves some user parameter for a plug in
//	pConfig		: is the config data for the server
//	uCount		: is the count of parameters which the plugin need
//	Returns the access index or 0 if an error occurs
HTTP_API int HttpReserveUserParams(HttpConfigData *pConfig,unsigned uCount)
{

	if(pConfig->uUserOffset+uCount>HTTP_USER_PARAMS)
		{
		return 0;
		}

	pConfig->uUserOffset += uCount;


return -(int)pConfig->uUserOffset;
}

//*****************************************************************************
//*
//*		HttpGetUserValue
//*
//*****************************************************************************
//	Gets an user value
//	hConn		: pointer to the connection data
//	uNumber		: number of the uservalue
//	Returns the user value
HTTP_API int HttpGetUserValue(HttpConnection hConn,unsigned uNumber)
{
HttpData   *pData=(HttpData*)hConn;


	   uNumber+=pData->uUserOffset;
	if(uNumber>=HTTP_USER_PARAMS)return 0;

return (int)pData->pUserParams[uNumber];
}

//*****************************************************************************
//*
//*		HttpGetUserPointer
//*
//*****************************************************************************
//	Gets an user pointer
//	hConn		: pointer to the connection data
//	uNumber		: number of the uservalue
//	Returns the user pointer
HTTP_API void *HttpGetUserPointer(HttpConnection hConn,unsigned uNumber)
{
HttpData   *pData=(HttpData*)hConn;

	   uNumber+=pData->uUserOffset;
	if(uNumber>=HTTP_USER_PARAMS)return 0;

return pData->pUserParams[uNumber];
}

//*****************************************************************************
//*
//*		HttpSetUserValue
//*
//*****************************************************************************
//	Sets an user value
//	hConn		: pointer to the connection data
//	uNumber		: number of the uservalue
//	iValue		: is the new value
HTTP_API void HttpSetUserValue(HttpConnection hConn,unsigned uNumber,int iValue)
{
HttpData   *pData=(HttpData*)hConn;

	   uNumber+=pData->uUserOffset;
	if(uNumber>=HTTP_USER_PARAMS)return;

	pData->pUserParams[uNumber] = (void*)iValue;
}

//*****************************************************************************
//*
//*		HttpSetUserValue
//*
//*****************************************************************************
//	Sets an user pointer
//	hConn		: pointer to the connection data
//	uNumber		: number of the uservalue
//	pPointer	: is the new pointer
HTTP_API void HttpSetUserPointer(HttpConnection hConn,unsigned uNumber,void *pPointer)
{
HttpData   *pData=(HttpData*)hConn;

	   uNumber+=pData->uUserOffset;
	if(uNumber>=HTTP_USER_PARAMS)return;

	pData->pUserParams[uNumber] = pPointer;

}

//*****************************************************************************
//*
//*		HttpGetCookies
//*
//*****************************************************************************
//	Gets the cookies of the current request
//	hConn		: pointer to the connection data
//	Returns a pointer to all cookies
HTTP_API const char *HttpGetCookies(HttpConnection hConn)
{
HttpData   *pData=(HttpData*)hConn;


return pData->cRequCookie;
}

//*****************************************************************************
//*
//*		HttpGetCookie
//*
//*****************************************************************************
//	Gets a single cookie of the current request
//	hConn		: pointer to the connection data
//	pCookie		: is the name of the cookie
//	Returns a pointer to the cookie or NULL if not found
//	The cookie is only valid until the ';' sign.
HTTP_API const char *HttpGetCookie(HttpConnection hConn,const char *pCookie)
{
HttpData   *pData=(HttpData*)hConn;
char	   *pText;
int			iLen;


	pText = pData->cRequCookie;
	iLen  = strlen(pCookie);

	while(*pText)
		{
		if(pText[iLen]=='=' && !memcmp(pText,pCookie,iLen))
			{
			return pText+iLen+1;
			}

		for(;*pText;pText++)
			{
			if(*pText==';')continue;
			pText++;
			break;
			}

		for(;*pText;pText++)
			{
			if(*pText!=' ')break;
			}
		}


return 0;
}

//*****************************************************************************
//*
//*		HttpGetCookieValue
//*
//*****************************************************************************
//	Gets a single cookie of the current request
//	hConn		: pointer to the connection data
//	pCookie		: is the name of the cookie
//	Returns a pointer to the cookie or NULL if not found
//	The cookie is only valid until the ';' sign.
HTTP_API int HttpGetCookieValue(HttpConnection hConn,const char *pCookie)
{
const char	   *pText;


		pText = HttpGetCookie(hConn,pCookie);
	if(!pText)return 0;


	if(*pText=='0')
		{
		if(pText[1]=='x' || pText[1]=='X')
			{
			return strtol(pText+2,NULL,16);
			}

		if(pText[1]=='b' || pText[1]=='B')
			{
			return strtol(pText+2,NULL,2);
			}

		if(pText[1]=='o' || pText[1]=='O')
			{
			return strtol(pText+2,NULL,8);
			}
		}


return strtol(pText,NULL,10);
}

//*****************************************************************************
//*
//*		HttpSetModifyTime
//*
//*****************************************************************************
//	Sets the modify time for the request
//	hConn		: pointer to the connection data
//	uTime		: is the modify time. zero mean no transmit
HTTP_API void HttpSetModifyTime(HttpConnection hConn,SysDW uTime)
{
HttpData   *pData=(HttpData*)hConn;


	pData->uResponseTime = uTime;


}

//*****************************************************************************
//*
//*		HttpSetCacheMode
//*
//*****************************************************************************
//	Sets the cache options for the current request
//	hConn		: pointer to the connection data
//	uMode		: are the cache flags
//					HTTP_CACHE_NOCACHE		don't cache the side
//					HTTP_CACHE_NOSTORE		don't store the side
//					HTTP_CACHE_MAXAGE		set max age of this side
//					HTTP_CACHE_PUBLIC		set public cache mode
//					HTTP_CACHE_PRIVATE		set private cache mode
HTTP_API void HttpSetCacheMode(HttpConnection hConn,unsigned uMode)
{
HttpData   *pData=(HttpData*)hConn;


	pData->uResponseCache = uMode;


}

//*****************************************************************************
//*
//*		HttpAddCookie
//*
//*****************************************************************************
//	Adds a cookie at the next response
//	hConn		: pointer to the connection data
//	pCookie		: is the name of the cookie
//	pValue		: is the value of the cookie (don't use ";=%\r\n" in the value)
//	pPath		: is the path to which the cookie belongs (0 means the current path)
//	pDomain		: is the domain to which the cookie belongs (0 means the current domain)
//	uMaxAge		: is the maximum age of the cookie in seconds (0 means infinite)
//	Returns the cookie length if the cookie was added or 0 if the buffer was too small
HTTP_API int HttpAddCookie(HttpConnection hConn,const char *pCookie,const char *pValue,const char *pPath,const char *pDomain,unsigned uMaxAge)
{
HttpData   *pData=(HttpData*)hConn;
int			iLenC;
int			iLenV;
int			iLenP;
int			iLenD;
int			iLen;
int			iAdd;


	if(!pDomain)pDomain="";
	if(!pValue )pValue ="";
	if(!pPath  )pPath  ="";

	for(iLen=0;pData->cResponseCookie[iLen];)
		{
		iLen += strlen(pData->cResponseCookie)+1;
		}

	iLenV	= strlen(pValue );
	iLenC	= strlen(pCookie);
	iLenD	= strlen(pDomain);
	iLenP	= strlen(pDomain);
	iAdd	= iLen+iLenC+iLenD+iLenV+iLenP+14;

	if(iLenP>0)iAdd+= 6;
	if(iLenD>0)iAdd+= 8;
	if(uMaxAge)iAdd+=20;

	if(iAdd>=sizeof(pData->cResponseCookie))
		{
		return 0;
		}

	memcpy(pData->cResponseCookie+iLen,pCookie,iLenC);		iLen+=iLenC;
	pData->cResponseCookie[iLen]='=';						iLen++;
	memcpy(pData->cResponseCookie+iLen,pValue ,iLenV);		iLen+=iLenV;
	memcpy(pData->cResponseCookie+iLen,";Version=1;",11);	iLen+=11;

	if(iLenP>0)
		{
		memcpy(pData->cResponseCookie+iLen,"Path=",5);		iLen+=5;
		memcpy(pData->cResponseCookie+iLen,pPath,iLenP);	iLen+=iLenP;
		pData->cResponseCookie[iLen]=';';					iLen++;
		}

	if(iLenD>0)
		{
		memcpy(pData->cResponseCookie+iLen,"Domain=",7);	iLen+=7;
		memcpy(pData->cResponseCookie+iLen,pDomain,iLenD);	iLen+=iLenD;
		pData->cResponseCookie[iLen]=';';					iLen++;
		}

	if(uMaxAge)
		{
		iLen+=sprintf(pData->cResponseCookie+iLen,"Max-Age=%u;",uMaxAge);
		}


return iLen;
}

//*****************************************************************************
//*
//*		HttpGetHost
//*
//*****************************************************************************
//	Gets the ip address of the client (i.e. "10.17.10.234")
//	hConn		: pointer to the connection data
//	Returns a pointer to the string
HTTP_API const char *HttpGetIpAddr(HttpConnection hConn)
{
HttpData   *pData=(HttpData*)hConn;

return pData->cConnIp;
}

//*****************************************************************************
//*
//*		HttpGetHost
//*
//*****************************************************************************
//	Gets the hostname of the current request (i.e. "www.myserver.com")
//	hConn		: pointer to the connection data
//	Returns a pointer to the string
HTTP_API const char *HttpGetHost(HttpConnection hConn)
{
HttpData   *pData=(HttpData*)hConn;


return pData->cRequHost;
}

//*****************************************************************************
//*
//*		HttpGetPath
//*
//*****************************************************************************
//	Gets the url of the current request without the hostname (i.e. "/index.html")
//	hConn		: pointer to the connection data
//	pLength		: if not zero the length of the string will be storted there.
//	Returns a pointer to the string
HTTP_API const char *HttpGetPath(HttpConnection hConn,void *pLength)
{
HttpData   *pData=(HttpData*)hConn;

	if(pLength)*(unsigned*)pLength=pData->uRequPathLength;

return pData->cRequPath;
}

//*****************************************************************************
//*
//*		HttpGetParam
//*
//*****************************************************************************
//	Gets the paramter of the current request (i.e. "&all+find=hallo")
//	hConn		: pointer to the connection data
//	pLength		: if not zero the length of the string will be storted there.
//	Returns a pointer to the string
HTTP_API const char *HttpGetParam(HttpConnection hConn,void *pLength)
{
HttpData   *pData=(HttpData*)hConn;

	if(pLength)*(unsigned*)pLength=pData->uRequParamLength;

return pData->cRequParam;
}

//*****************************************************************************
//*
//*		HttpGetFile
//*
//*****************************************************************************
//	Gets the filename for the current request (i.e. "c:/temp/index.html")
//	hConn		: pointer to the connection data
//	Returns a pointer to the string
HTTP_API const char *HttpGetFile(HttpConnection hConn)
{
HttpData   *pData=(HttpData*)hConn;


return pData->cRequFile;
}

//*****************************************************************************
//*
//*		HttpGetUser
//*
//*****************************************************************************
//	Gets the username of the current request
//	hConn		: pointer to the connection data
//	Returns a pointer to the string
HTTP_API const char *HttpGetUser(HttpConnection hConn)
{
HttpData   *pData=(HttpData*)hConn;


return pData->cAuthUser;
}

//*****************************************************************************
//*
//*		HttpLogoutUser
//*
//*****************************************************************************
//	Logout an user
//	hConn		: pointer to the connection data
//	pUser		: is the username
//	Returns 1 if the user was found or 0 if not
HTTP_API int HttpLogoutUser(HttpConnection hConn,const char *pUser)
{

	#if HTTP_HAVE_AUTH

	HttpData   *pData=(HttpData*)hConn;

	pData->cAuthUser [0] = 0;
	pData->cAuthNonce[0] = 0;

	return DelUserCacheEntry(pData->pServer,pUser);

	#else

	return 0;

	#endif

}

//*****************************************************************************
//*
//*		HttpSetAuthMode
//*
//*****************************************************************************
//	Set the authorizaion mode
//	hConn		: pointer to the connection data
//	uMode		: is the new authorizaion mode
//					0 = check login
//					1 = ignore login
//	Returns the previous authorizaion mode
HTTP_API int HttpSetAuthMode(HttpConnection hConn,unsigned uMode)
{
unsigned		uOld;


	#if HTTP_HAVE_AUTH

	HttpData   *pData=(HttpData*)hConn;

	uOld = pData->uIgnoreLogin;
	pData->uIgnoreLogin = uMode;

	#else

	uOld = 1;

	#endif


return uOld;
}

//*****************************************************************************
//*
//*		HttpGetContent
//*
//*****************************************************************************
//	Gets the content type of the current request
//	hConn		: pointer to the connection data
//	Returns a pointer to the string
HTTP_API const char *HttpGetContent(HttpConnection hConn)
{
HttpData	   *pData=(HttpData*)hConn;


return pData->cResponseType;
}

//*****************************************************************************
//*
//*		HttpSetContent
//*
//*****************************************************************************
//	Sets the content type of the current request
//	hConn		: pointer to the connection data
//	pContent	: is the new content type
//	Returns 1 if set, 0 if an error occurs.
HTTP_API int HttpSetContent(HttpConnection hConn,const char *pContent)
{
HttpData   *pData=(HttpData*)hConn;
int			iLen;


	iLen = strlen(pContent);

	if(iLen>=sizeof(pData->cResponseType))
		{
		return 0;
		}

	memcpy(pData->cResponseType,pContent,iLen+1);


return 0;
}
