//*****************************************************************************
//*
//*
//*      HttpServer.h
//*
//*
//*****************************************************************************
//
//  Copyright  2009    Anton Zechner
//
//  AzHttp is distributed under the GNU GENERAL PUBLIC LICENSE (GPL)
//  Sourcecode which use AzHttp must be published. Commercial users
//  must published their code too, or make an licence agreement with me.
//
//
//  AzHttp wird unter GNU GENERAL PUBLIC LICENSE (GPL) vertreiben.
//  Sourcecode welcher AzHttp verwendet muss verffentlicht werden.
//  Kommerzielle Nutzer mssen ihren Code ebenfalls verffentlichen, oder
//  eine Nutzungsvereinbarung mit mir treffen.
//
//  az_software@inode.at
//
#ifndef     __HTTPSERVER_H__
#define     __HTTPSERVER_H__

#include    "System/System.h"
#include    "System/SocketUtility.h"

#define 	HTTP_API				extern "C"


// http server errors

#define     HTTP_ERR_NONE			0       			// no error
#define     HTTP_ERR_SOCKET      	1       			// socket error
#define     HTTP_ERR_ISRUNNING   	2       			// http server is allready running
#define     HTTP_ERR_ISNOTRUNNING	3       			// http server is not unning
#define     HTTP_ERR_WRONGPATH   	4       			// the file path is wrong
#define     HTTP_ERR_WRONGIP     	5       			// the ip address is wrong
#define 	HTTP_ERR_INIT		 	6					// initialization error
#define 	HTTP_ERR_NOTSTOPED		7					// the server could not stopped

// http reason macros for disconnect callback

#define 	HTTP_MSG_CONNLOST		0					// connection closed by clinent or an error
#define 	HTTP_MSG_DISCONNECT		1					// the server disconnects the connection
#define 	HTTP_MSG_TIMEOUT		2					// no request from client
#define 	HTTP_MSG_SHUTDOWN		3					// the server willbe shutdown
#define 	HTTP_MSG_USERBREAK		4					// a callback initiate a abort
#define 	HTTP_MSG_BUFFEROVERFLOW	5					// the server buffer are too small for a request

// http commands for request callback

#define 	HTTP_CMD_GET			0					// GET command
#define 	HTTP_CMD_HEAD			1					// HEAD command
#define 	HTTP_CMD_POST			2					// POST command
#define 	HTTP_CMD_PUT			3					// PUT command
#define 	HTTP_CMD_DELETE			4					// DELETE command
#define 	HTTP_CMD_TRACE			5					// TRACE command
#define 	HTTP_CMD_OPTIONS		6					// OPTIONS command
#define 	HTTP_CMD_CONNECT 		7					// CONNECT command
#define 	HTTP_CMD_LINK	 		8					// LINK command
#define 	HTTP_CMD_UNLINK	 		9					// UNLINK command
#define 	HTTP_CMD_NOTSOPPORTED	10					// unknown command

// http request callback response

#define 	HTTP_RESP_DEFAULT		0					// use the default
#define 	HTTP_RESP_CALLBACK		1					// use a callback for the response
#define 	HTTP_POST_DEFAULT		2					// use the default post via internal memory
#define 	HTTP_POST_CALLBACK		3					// use a post via callback

// constand for HttpSetCacheMode

#define 	HTTP_CACHE_NOCACHE		0x80000000			// don't cache the side
#define 	HTTP_CACHE_NOSTORE		0x40000000			// don't store the side
#define 	HTTP_CACHE_MAXAGE		0x20000000			// set max age of this side
#define 	HTTP_CACHE_PUBLIC		0x10000000			// set public cache mode
#define 	HTTP_CACHE_PRIVATE		0x08000000			// set private cache mode
#define 	HTTP_CACHE_AGEMASK		0x07FFFFFF			// bit mask for the age in seconds



#ifndef	   _MSC_VER
#ifndef	   _cdecl
#define    _cdecl
#endif
#endif


#ifdef  __cplusplus
#define D(a)	a
extern "C" {
#else
#define D(a)
#endif

typedef void *HttpConnection;

typedef int  (_cdecl *HttpDataCB      )(HttpConnection hConn,char *pBuffer,int iSize,int iOffset);
typedef int  (_cdecl *HttpRequestCB   )(HttpConnection hConn,int iCommand,const char *pHost,const char *pUrl,const char *pParams,int iUrlLen,int iParamLen);
typedef int  (_cdecl *HttpConnectCB   )(HttpConnection hConn,const char *pIpAddr,unsigned short usPort);
typedef int  (_cdecl *HttpPasswordCB  )(HttpConnection hConn,const char *pUser,char *pBuffer,int iSize);
typedef void (_cdecl *HttpDisconnectCB)(HttpConnection hConn,const char *pIpAddr,unsigned short usPort,int iReason);

typedef struct
    {
    char        		cIpAddress [ 64];					// ip addres off the http-server
	char				cDefaultExt[ 32];					// default file name for access to paths
	char				cAuthRealm [ 64];					// is the realm value for the authorization
	char				cFilePath  [256];					// is the path for server files
    unsigned short 		usTcpPort;							// is the tcp port for the server
	unsigned			uUserOffset;						// start offest for user parameters (set to zero)
	int					iMaxConnections;					// maximum count of client connections
	int					iPriority;							// taks priority 0-256
	HttpRequestCB		pCbRequest;							// callback for a client request
	HttpConnectCB   	pCbConnect;							// callback for client connection
	HttpPasswordCB		pCbPassword;						// callback for password returning
	HttpDisconnectCB	pCbDisconnect;						// callback for client disconnection
    }HttpConfigData;


//*****************************************************************************
//*
//*		Functions
//*
//*****************************************************************************

int			HttpStopDaemon		 ();
int			HttpDaemon			 (HttpConfigData *pConfig);
int			HttpReserveUserParams(HttpConfigData *pConfig,unsigned uCount);
int			HttpTransmitGoto	 (HttpConnection hConn,const char *pUrl,const char *pParam D(=0),int iDissconect D(=0));
int			HttpTransmitFile	 (HttpConnection hConn,const char *pFilePath);
int			HttpTransmitMemory	 (HttpConnection hConn,const void *pMemory,int iSize);
int _cdecl	HttpTransmitCallback (HttpConnection hConn,HttpDataCB  pCallback);
void		HttpSetModifyTime    (HttpConnection hConn,SysDW       uTime);
void		HttpSetCacheMode	 (HttpConnection hConn,unsigned    uMode);
int			HttpGetUserValue	 (HttpConnection hConn,unsigned    uNumber);
void	 	HttpSetUserValue	 (HttpConnection hConn,unsigned    uNumber,int iValue);
void	   *HttpGetUserPointer	 (HttpConnection hConn,unsigned    uNumber);
void	 	HttpSetUserPointer	 (HttpConnection hConn,unsigned    uNumber,void *pPointer);
int			HttpSetAuthMode		 (HttpConnection hConn,unsigned	   uMode);
int			HttpLogoutUser		 (HttpConnection hConn,const char *pUser);
int			HttpAddCookie		 (HttpConnection hConn,const char *pCookie,const char *pValue,const char *pPath D(=0),const char *pDomain D(=0),unsigned uMaxAge D(=0));
int			HttpGetCookieValue	 (HttpConnection hConn,const char *pCookie);
const char *HttpGetCookie		 (HttpConnection hConn,const char *pCookie);
const char *HttpGetCookies		 (HttpConnection hConn);
const char *HttpGetContent		 (HttpConnection hConn);
const char *HttpGetIpAddr		 (HttpConnection hConn);
const char *HttpGetHost			 (HttpConnection hConn);
const char *HttpGetFile			 (HttpConnection hConn);
const char *HttpGetUser			 (HttpConnection hConn);
const char *HttpGetPath			 (HttpConnection hConn ,void *pLength D(=0));
const char *HttpGetParam		 (HttpConnection hConn ,void *pLength D(=0));
int			HttpSetContent       (HttpConnection hConn ,const char *pContent);
const char *HttpGetContent       (HttpConnection hConn);
int			HttpEncodeHtml		 (const char     *pText,char *pTarget,int iSize,int iLen D(=0x7FFFFFFF));
int			HttpEncodeString	 (const char     *pText,char *pTarget,int iSize,int iLen D(=0x7FFFFFFF));

#ifdef  __cplusplus
}
#endif

#undef 	D


#endif

