//*****************************************************************************
//*
//*
//*		HttpTreePlugin.cpp
//*
//*
//*****************************************************************************
//
//	This plugin allows to show trees in a html side.
//
#include	<string.h>
#include	<stdlib.h>
#include	<stdio.h>

#include	"HttpInternal.h"
#include	"HttpTreePlugin.h"
#include	"HttpIconPlugin.h"
#include	"System/System.h"


#define 	ENTRIES(a)		(sizeof(a)/sizeof((a)[0]))
#define 	HTTP_API		extern "C"
#define 	TP_MAX_TARGET	16								// largest target string
#define 	TP_MAX_PREFIX	16								// maximum size of the url prefix
#define 	TP_MAX_SUFFIX	16								// maximum size of the url suffux
#define 	TP_MAX_SCAN		600								// maximum count of files in scan
#define 	TP_MAX_LIST		512								// maximum count of files in list
#define 	TP_MAX_PATH		256								// langest path name for the list
#define 	TP_MAX_BODY		54								// langest body exension
#define 	TP_MAX_NAME		54								// langest file name for the list
#define 	TP_MAX_URL		64								// langest file name for the list
#define 	TP_COUNT		2								// count of lists in the plugin
#define 	TP_SCANS		4								// count of scans in the plugin
#define 	TP_STATIC_MEM	1								// use only staic memory

#if			TP_COUNT>256
#error		TP_COUNT is to high (max=256)
#endif

#if			TP_MAX_LIST>4090
#error		TP_MAX_LIST is to high (max=4090)
#endif



typedef	struct
	{
	unsigned				uParent;						// is the number of the parent entry plus 1
	unsigned char 			ucType;							// type of entry (0=file 1=folder 2=file-folder 3=other-folder)
	unsigned char 			ucLevel;						// the level ov the entry (0=root, bit7=last at level)
	unsigned char 			ucNameLen;						// length of the filename
	unsigned char 			ucSuffixPos;					// position of the suffix
	char					cName[TP_MAX_NAME];				// file name
	}TreeEntry;

typedef	struct
	{
	char					cUrl   [TP_MAX_URL   ];			// http url for the list
	char					cPath  [TP_MAX_PATH  ];			// http url for the list
	char					cBody  [TP_MAX_BODY  ];			// html body exension
	char					cTarget[TP_MAX_TARGET];			// html target exension
	char					cPrefix[TP_MAX_PREFIX];			// prefix for the url
	char					cSuffix[TP_MAX_SUFFIX];			// suffix for the url
	unsigned				uLocked;						// is the data block in use
	unsigned				uUrlLength;						// size of url
	unsigned				uPathLength;					// size of the path
	unsigned				uBodyLength;					// size of body exension
	unsigned				uTargetLength;					// size of target exension
	unsigned				uPrefixLength;					// size of url prefix
	unsigned				uSuffixLength;					// size of url suffix

	unsigned char			ucNoLines;						// no lines in the tree
	unsigned char			ucNoSuffux;						// don't show file suffixes
	unsigned char			ucScanRoot;						// add files from root path
	unsigned char			ucNoRootUrl;					// don't insert href urls from the root
	unsigned char			ucNoTreeUrl;					// don't insert the url of the tree
	unsigned char			ucExpandAll;					// expand all entries
	unsigned char			ucSortByType;					// sort files by type
	unsigned char			ucNoFolderUrl;					// don't insert href urls for folders

	const char			   *pSuffixFilter;					// suffix of all allowed file types
	unsigned				uEntryCount;					// count of entries
	TreeEntry				aEntries[TP_MAX_LIST+1];		// list off entries
	}TreeList;

typedef	struct
	{
	unsigned				uLocked;						// is the data block in use
	unsigned				uEntryPos;						// current entry
	unsigned				uEntryCount;					// count of entries
	TreeEntry				aEntries[TP_MAX_SCAN];			// list off entries
	}TreeScan;


static	const char			cHeader1 [  ] = "<html><head><style>TD{font-size: 8pt;font-family: verdana,helvetica;text-decoration: none;white-space:nowrap;}A{text-decoration: none;color: black}\ndiv#f1\n\t{\n\tmargin:0px 0px 0px 0px;\n\tborder:0px;\n\tpadding: 0px;\n\theight:%ipx;\n\t}\n</style></head><body ";
static	const char			cHeader2 [  ] = " topmargin=16 marginheight=16>";
static	const char			cEntry0A [  ] = "<td valign=top><a href=";
static	const char			cEntry0F [  ] = "<td valign=top><!--    ";
static	const char			cEntry0B [  ] = "><img src=";
static	const char			cEntry0C [  ] = " border=0></a></td><td nowrap=nowrap valign=middle width=100%>&#160;<a href=";
static	const char			cEntry0G [  ] = " border=0></a></td><td nowrap=nowrap valign=middle width=100%>&#160;<!--    ";
static	const char			cEntry0D [  ] = ">";
static	const char			cEntry0E [  ] = "</a></td>";
static	const char			cEntry1A [  ] = "<td valign=top><a href=";
static	const char			cEntry1B [  ] = "<td valign=top";
static	const char			cEntry1C [  ] = "><img src=";
static	const char			cEntry1D [  ] = " border=0 height=18 width=16></a></td>";
static	const char			cEntry2A [  ] = "<td valign=top><img src=";
static	const char			cEntry2B [  ] = " border=0 height=18 width=16></a></td>";
static	const char			cEntry3A [  ] = "<div id=f1 style=\"display: block;\"><table border=0 cellpadding=0 cellspacing=0 width=100%><tbody><tr>";
static	const char			cEntry3B [  ] = "</tr></tbody></table></div>";
static	const char			cTail1   [  ] = "</body><html>";


#if TP_STATIC_MEM
static  TreeList			aTrees[TP_COUNT];				// memory for static trees
static  TreeScan			aScans[TP_SCANS];				// memory for static scans
static  unsigned			uScanNext;						// next search position for scanns
#endif

static	SysSemaphore		hTreeSem     = 0;				// semaphor for tree access
static	SysSemaphore		hTreeSemNum  = 0;				// semaphor for tree count
static	SysSemaphore		hTreeSemScan = 0;				// semaphor for scan count
static  unsigned			uTreeNext;						// next search position for trees
static	TreeList		   *aTreeArray[TP_COUNT];			// pointer of all lists
static	int					iUserParamId=0;					// id for the user parameter
static	HttpRequestCB		pOldRequest;					// next callback in chain
static	HttpDisconnectCB	pOldDisconnect;					// next disconnect in chain


//*****************************************************************************
//*
//*		CmpByName
//*
//*****************************************************************************
//	Compare function for name sorting
static int _cdecl CmpByName(const void *pPtr1,const void *pPtr2)
{
const TreeEntry    *pEntry1=(TreeEntry*)pPtr1;
const TreeEntry    *pEntry2=(TreeEntry*)pPtr2;



return stricmp(pEntry1->cName,pEntry2->cName);
}


//*****************************************************************************
//*
//*		CmpByType
//*
//*****************************************************************************
//	Compare function for type sorting
static int _cdecl CmpByType(const void *pPtr1,const void *pPtr2)
{
const TreeEntry    *pEntry1=(TreeEntry*)pPtr1;
const TreeEntry    *pEntry2=(TreeEntry*)pPtr2;
const char		   *pName1;
const char		   *pName2;
int					iCmp;



	pName1 = pEntry1->cName+pEntry1->ucSuffixPos;
	pName2 = pEntry2->cName+pEntry2->ucSuffixPos;


	   iCmp = stricmp(pName1,pName2);
	if(iCmp)return iCmp;


return stricmp(pEntry1->cName,pEntry2->cName);
}


//*****************************************************************************
//*
//*		CmpFileSuffix
//*
//*****************************************************************************
//	Compares a file suffix with a type list
//	pSufix		: pointer to the suffix
//	pType		: pointer to the type list (i.e. "exe;txt;cpp" )
//	iNum		: size of the suffix
//	Returns 1 if a type is equal
static int CmpFileSuffix(const char *pSufix ,const char *pType,int iNum)
{
int		iLen;



	iNum = strlen(pSufix);

	for(;;)
		{
		for(iLen=0;;iLen++)								// get length
			{
			if(pType[iLen]== 0 )break;
			if(pType[iLen]==';')break;
			}

		if(iLen==iNum && !strnicmp(pSufix,pType,iNum))
			{
			return 1;
			}

			pType+=iLen;								// next type
		if(*pType==0)break;

		pType++;
		}


return 0;
}

//*****************************************************************************
//*
//*		SetParameter
//*
//*****************************************************************************
//	Sets a parameter after a url
//	pParam		: is the current parameter
//	iEntry		: is the entry which should be changed
//	iExpand		: is the entry expanted
//	pBuffer		: is the target buffer
//	iSize		: is the size of the buffer
//	iDefault	: is the default value if the bit wan't found
//	Returns the length of the stored parameter
static int SetParameter(const char *pParam,int iEntry,int iExpand,char *pBuffer,int iSize,int iDefault)
{
char	cChar;
int		iPos;




	if(iSize<2)return 0;

	pBuffer[0] = '?';

	for(iPos=1;iPos<iSize;iPos++)
		{
			cChar = *pParam;
		if(!cChar)
			{
			if(iEntry<0)break;
			cChar = (iDefault)? 'a'+15:'a';
			}
		else{
			pParam++;
			}

		if((iEntry&~3)==0)
			{
			cChar -= 'a';

			if(iExpand)
					cChar |=  (1<<(iEntry&3));
			else	cChar &= ~(1<<(iEntry&3));

			cChar += 'a';
			}

		pBuffer[iPos] = cChar;
		iEntry       -= 4;
		}



return iPos;
}

//*****************************************************************************
//*
//*		IsExpaned
//*
//*****************************************************************************
//	Gets the expanted bit for a entry
//	pParam		: is the parameter string
//	iEntry		: is the entry number
//	iDefault	: is the default return value
//	Returns 1 if the entry is expanted
static int IsExpaned(const char *pParam,int iEntry,int iDefault=0)
{
int		iVal;


	while(iEntry>=4)
		{
		if(!*pParam)return iDefault;
		pParam+=1;
		iEntry-=4;
		}


		iVal = *pParam;
	if(!iVal)return iDefault;


return ((iVal-'a')>>iEntry)&1;
}



//*****************************************************************************
//*
//*		AllocTreeList
//*
//*****************************************************************************
//	Allocates a new folder list
//	Returns the pointer to the list, or 0 if out of memory
static TreeList *AllocTreeList()
{
TreeList   *pTree;
int			iCount;



	if(!SysSemaphoreLockWait(hTreeSemNum,3000))
		{
		return 0;
		}


	SysSemaphoreLock(hTreeSem);

	pTree = 0;

	#if TP_STATIC_MEM


	for(iCount=0;iCount<TP_COUNT;iCount++,uTreeNext++)
		{
		if(uTreeNext>=TP_COUNT)uTreeNext=0;
		if(aTrees[uTreeNext].uLocked)continue;

		pTree= aTrees+uTreeNext;

		break;
		}

	#else


	pTree = new TreeList;

	#endif


	if(pTree)
		{
		for(iCount=0;iCount<TP_COUNT;iCount++,uTreeNext++)
			{
			if(uTreeNext>=TP_COUNT)uTreeNext=0;
			if(aTreeArray[uTreeNext])continue;

			aTreeArray[uTreeNext] = pTree;
			pTree->uLocked        = 1;

			break;
			}
		}


	SysSemaphoreUnlock(hTreeSem);


return pTree;
}

//*****************************************************************************
//*
//*		DeallocTreeList
//*
//*****************************************************************************
//	Deallocates a folder list
//	pTree	: is the pointer to the list
static void DeallocTreeList(TreeList *pTree)
{
int		iPos;


	if(!pTree)return;


	pTree->uLocked     = 0;

	for(iPos=0;iPos<TP_COUNT;iPos++)
		{
		if(aTreeArray[iPos]!=pTree)continue;
		aTreeArray[iPos] = 0;
		break;
		}


	#if !TP_STATIC_MEM
	delete pTree;
	#endif

	SysSemaphoreUnlock(hTreeSemNum);


}

//*****************************************************************************
//*
//*		AllocScanList
//*
//*****************************************************************************
//	Allocates a new folder scan list
//	Returns the pointer to the list, or 0 if out of memory
static TreeScan *AllocScanList()
{
TreeScan   *pScan;
int			iCount;





	#if TP_STATIC_MEM

	if(!SysSemaphoreLockWait(hTreeSemScan,3000))
		{
		return 0;
		}

	SysSemaphoreLock(hTreeSem);

	pScan = 0;

	for(iCount=0;iCount<TP_COUNT;iCount++,uScanNext++)
		{
		if(uScanNext>=TP_COUNT)uScanNext=0;
		if(aScans[uScanNext].uLocked)continue;

		pScan= aScans+uScanNext;

		break;
		}

	SysSemaphoreUnlock(hTreeSem);

	#else


	pScan = new TreeScan;

	#endif




return pScan;
}

//*****************************************************************************
//*
//*		DeallocScanList
//*
//*****************************************************************************
//	Deallocates a folder scan list
//	pScan	: is the pointer to the list
static void DeallocScanList(TreeScan *pScan)
{


	if(!pScan)return;


	#if TP_STATIC_MEM

	pScan->uLocked     = 0;
	SysSemaphoreUnlock(hTreeSemScan);

	#else

	delete pScan;

	#endif


}


//*****************************************************************************
//*
//*		GetEntryPath
//*
//*****************************************************************************
//	Gets the file path of a entry
//	pTree		: pointer to the tree
//	uEntry		: number of the entry in the list
//	pBuffer		: pointer to the buffer for the file path
//	iSize		: size of the buffer
//	Returns the size of the path or 0 an error occurs
static int GetEntryPath(TreeList *pTree,unsigned uEntry,char *pBuffer,int iSize)
{
TreeEntry	   *pEntry;
unsigned		aLevels[8];
int				iCount;
int				iNum;
int				iLen;


	if(uEntry>pTree->uEntryCount)							// check entry
		{
		return 0;
		}

	iLen = pTree->uPathLength;
	if(iLen>=iSize)return 0;
	memcpy(pBuffer,pTree->cPath,iLen);


	if(uEntry>pTree->uEntryCount)							// root path
		{
		if(iLen>0)iLen--;
		pBuffer[iLen]=0;
		return  iLen;
		}


	pEntry		= pTree->aEntries+uEntry;
	aLevels[0]	= uEntry;
	iCount		= 1;


	while(pEntry->uParent)									// goto root entry
		{
		aLevels[iCount]	= pEntry->uParent-1;
		pEntry			= pTree->aEntries+aLevels[iCount];

		if(pEntry->ucType>2)break;

		iCount ++;

		if(iCount>=8)return 0;
		}


	for(iCount--;;iCount--)									// copy path names
		{
		pEntry	= pTree->aEntries+aLevels[iCount];
		iNum	= pEntry->ucNameLen;

		if(iLen+iNum+1>=iSize)return 0;

		memcpy(pBuffer+iLen,pEntry->cName,iNum);
		iLen += iNum;

		if(iCount<=0)break;

		pBuffer[iLen] = '/';
		iLen++;
		}

	pBuffer[iLen] = 0;



return iLen;
}

//*****************************************************************************
//*
//*		GetEntryUrl
//*
//*****************************************************************************
//	Gets the url of a entry
//	pTree		: pointer to the tree
//	uEntry		: number of the entry in the list
//	pBuffer		: pointer to the buffer for the file path
//	iSize		: size of the buffer
//	Returns the size of the path or 0 an error occurs
static int GetEntryUrl(TreeList *pTree,unsigned uEntry,char *pBuffer,int iSize)
{
TreeEntry	   *pEntry;
unsigned		aLevels[8];
int				iCount;
int				iNum;
int				iLen;
int				iMax;



	if(pTree->uPrefixLength)								// insert prefix
		{
		iLen = pTree->uPrefixLength;
		if(iLen+1>=iSize)return 0;
		memcpy(pBuffer,pTree->cPrefix,iLen);
		}
	else{
		iLen = 0;
		}


	if(!pTree->ucNoTreeUrl)									// insert url of the tree
		{
		iMax = iLen+pTree->uUrlLength;
		if(iMax+1>=iSize)return 0;
		memcpy(pBuffer+iLen,pTree->cUrl,pTree->uUrlLength);
		iMax = iLen;
		}


	if(uEntry>=pTree->uEntryCount)							// root path
		{
		if(iLen>0)iLen--;
		pBuffer[iLen]=0;
		goto Suffix;
		}

	pBuffer[iLen] = '/';iLen++;
	pEntry		  = pTree->aEntries+uEntry;
	aLevels[0]	  = uEntry;
	iCount		  = 1;


	while(pEntry->uParent)									// goto root entry
		{
		aLevels[iCount]	= pEntry->uParent-1;
		pEntry			= pTree->aEntries+aLevels[iCount];
		iCount		   += 1;

		if(iCount>=8)
			{
			goto Suffix;
			}
		}


	if(pTree->ucNoRootUrl)									// skip the root entry
		{
		   iCount--;
		if(iCount<=0)
			{
			pBuffer[iLen] = 0;
			goto Suffix;
			}
		}


	for(iCount--;;iCount--)									// copy path names
		{
		pEntry	= pTree->aEntries+aLevels[iCount];
		iNum	= pEntry->ucNameLen;

		if(iLen+iNum+1>=iSize)
			{
			goto Suffix;
			}

		memcpy(pBuffer+iLen,pEntry->cName,iNum);
		iLen += iNum;

		if(iCount<=0)break;

		pBuffer[iLen] = '/';
		iLen++;
		}


Suffix:


	if(pTree->uSuffixLength)								// insert suffix
		{
		   iMax = iLen+pTree->uSuffixLength;
		if(iMax+1<iSize)
			{
			memcpy(pBuffer+iLen,pTree->cSuffix,pTree->uSuffixLength);
			iLen=iMax;
			}
		}


	pBuffer[iLen] = 0;



return iLen;
}

//*****************************************************************************
//*
//*		ScanFolder
//*
//*****************************************************************************
//	Read all files of a folder
//	pTree		: pointer to the tree
//	pScan		: is the pointer for the scan list
//	uEntry		: number of the entry with the folder
//	iOnlyFirst	: only check if the path contains files
//	hWdog		: watchdog handel of thread
//	Returns the pointer to the scan list
static TreeScan *ScanFolder(TreeList *pTree,unsigned uEntry,int iOnlyFirst,SysWdog hWdog)
{
char			cTemp[sizeof(TreeScan)-sizeof(TreeEntry)*(TP_MAX_SCAN-1)];
char			cPath[256];
TreeEntry	   *pEntry;
TreeScan	   *pScan;
SysFindStruct	sData;
SysFind			hFind;
unsigned		uParent;
unsigned		uCount;
unsigned		uLevel;
int				iLen;
int				iOff;



		iOff = GetEntryPath(pTree,uEntry,cPath,sizeof(cPath));
	if(!iOff)return 0;


	SysThreadCheck();
	SysWatchdogTrigger(hWdog);

	strcpy(cPath+iOff,"/*");
	hFind = SysFindFirst(cPath,&sData);

	if(!hFind)return 0;


	if(iOnlyFirst)
		{
		pScan = (TreeScan*)cTemp;
		}
	else{
			pScan = AllocScanList();
		if(!pScan)return 0;
		}


	pScan->uEntryCount	= 0;
	pScan->uEntryPos	= 0;


	if(uEntry>=pTree->uEntryCount)
		{
		uLevel  =  0;
		uParent	=  0;
		}
	else{
		uLevel	=  pTree->aEntries[uEntry].ucLevel;
		uLevel  = (uLevel&0x7F)+1;
		uParent	=  uEntry+1;
		}

	pEntry	=  pScan->aEntries;
	uCount	=  0;


	for(;;)													// get all files
		{
		SysThreadCheck();
		SysWatchdogTrigger(hWdog);

		if(!(sData.dwAttr&SYS_DIR))
			{
			iLen = strlen(sData.cName);

			for(iOff=iLen;;iOff--)							// get suffix position
				{
				if(iOff<=0){iOff=iLen;break;}
				if(sData.cName[iOff-1]=='.')break;
				}

			if(!pTree->pSuffixFilter || CmpFileSuffix(sData.cName+iOff,pTree->pSuffixFilter,iLen-iOff))
				{
				if(uCount>=TP_MAX_SCAN)						// too much entries
					{
					pEntry--;

					memcpy(pEntry->cName,"...",4);
					pEntry->ucSuffixPos	= 3;
					pEntry->ucNameLen	= 3;
					pEntry->uParent		= uParent;
					pEntry->ucLevel		= uLevel;
					pEntry->ucType		= 0;

					break;
					}

				memcpy(pEntry->cName,sData.cName,iLen+1);	// insert entry
				pEntry->ucSuffixPos	= iOff;
				pEntry->ucNameLen	= iLen;
				pEntry->uParent		= uParent;
				pEntry->ucLevel		= uLevel;
				pEntry->ucType		= 0;

				uCount++;
				pEntry++;

				if(iOnlyFirst)break;
				}
			}

		if(SysFindNext(hFind,&sData)<0)						// get next file
			{
			break;
			}
		}


	SysFindClose(hFind);


	if(uCount==0)											// no files found
		{
		if(!iOnlyFirst)DeallocScanList(pScan);
		return 0;
		}

	SysThreadCheck();
	SysWatchdogTrigger(hWdog);


	if(uCount>0)											// sort the entries
		{
		qsort(pScan->aEntries,uCount,sizeof(TreeEntry),(pTree->ucSortByType)? CmpByType:CmpByName);
		}

	pScan->aEntries[uCount-1].ucLevel |= 0x80;
	pScan->uEntryCount = uCount;



return pScan;
}

//*****************************************************************************
//*
//*		InsertHref
//*
//*****************************************************************************
//	Inserts a HERF entry into an output buffer
//	pTree		: pointer to the tree
//	pBuffer		: pointer to the output buffer
//	iSize		: size of the output buffer
//	pPrefix		: is the prefix wich will be set befor the HREF path
//	iPreLen		: is the length of the prefix
//	pUrl		: is the pointer to the HREF path
//	iUrlLen		: is the size of the HREF path
//	Returns the count of bytes which will be copied to the output buffer
static int InsertHref(TreeList *pTree,char *pBuffer,int iSize,const char *pPrefix,int iPreLen,const char *pUrl,int iUrlLen)
{
int		iLen;



	if(iPreLen>=iSize)return 0;

	memcpy(pBuffer,pPrefix,iPreLen);

	iLen  = iPreLen;
	iLen += HttpEncodeString(pUrl,pBuffer+iLen,iSize-iLen,iUrlLen);

	if(pTree->uTargetLength)
		{
		memcpy(pBuffer+iLen  ," target=",8);
		memcpy(pBuffer+iLen+8,pTree->cTarget,pTree->uTargetLength);
		iLen += 8+pTree->uTargetLength;
		}



return iLen;
}


//*****************************************************************************
//*
//*		HttpTreeData
//*
//*****************************************************************************
//	Returns the html data of the folder
//	pBuffer		: pointer to the tranmit buffer
//	iSize		: is the size of the buffer (0=means dissconnect)
//	iOffset		: is the offest in the response
//	Returns the size of the stored data or an error number below zero
HTTP_API int _cdecl HttpTreeData(HttpConnection hConn,char *pBuffer,int iSize,int iOffset)
{
HttpData	   *pData=(HttpData*)hConn;
char			cPath[256];
unsigned char	aLevels[8];
const char     *pUrlPtr;
TreeEntry	   *pEntry;
TreeEntry	   *pNext;
TreeList	   *pTree;
TreeScan	   *pScan;
const char	   *pParam;
unsigned		uValue;
unsigned		uEntry;
unsigned		uNext;
unsigned		uList;
unsigned		uBits;
unsigned		uIcon;
unsigned		uTemp;
unsigned		uSub;
unsigned		uNow;
int				iLen;
int				iOff;
int				iCnt;
int				iNum;
int				iPos;
int				iLev;
int				iExp;



	SysThreadCheck();
	SysWatchdogTrigger(pData->hWatchdog);



	pScan	= (TreeScan*)HttpGetUserPointer(hConn,iUserParamId+1);
	uValue	=  HttpGetUserValue(hConn,iUserParamId);
	pParam	=  HttpGetParam(hConn);
	uNow	=  0;
	iExp	=  0;

	uSub	= (uValue>>31);
	uBits	= (uValue>>24)&0x007F;
	uEntry	= (uValue>> 8)&0xFFFF;
	uList	= (uValue    )&0x00FF;
	pTree	= (uList<TP_COUNT)? aTreeArray[uList]:0;



	if(!pTree)												// check tree pointer
		{
		if(pScan)
			{
			HttpSetUserPointer(hConn,iUserParamId+1,0);
			DeallocScanList(pScan);
			}

		return -1000;
		}

	if(uEntry>=0xFFFF)										// all data transmitted
		{
		return 0;
		}

	if(iOffset==0)											// beginn of the file
		{
		if(iSize<(int)(2048+sizeof(cHeader1)+sizeof(cHeader2)+pTree->uBodyLength))
			{
			return -1000;
			}

		memcpy(pBuffer,cHeader1,sizeof(cHeader1)-1);
		iLen  = sizeof(cHeader1)-1;

		memcpy(pBuffer+iLen,pTree->cBody,pTree->uBodyLength);
		iLen += pTree->uBodyLength;

		memcpy(pBuffer+iLen,cHeader2,sizeof(cHeader2)-1);
		iLen += sizeof(cHeader2)-1;

		if(pTree->ucScanRoot && !pTree->ucNoLines)
		if(ScanFolder(pTree,pTree->uEntryCount,1,pData->hWatchdog))
			{
			uBits = 1;
			}
		}
	else{
		iLen = 0;
		}


	while(iLen+2048<iSize)									// output files
		{
		if(!uSub && uEntry>=pTree->uEntryCount)
			{
			if(pScan)										// delalloc scan
				{
				HttpSetUserPointer(hConn,iUserParamId+1,0);
				DeallocScanList(pScan);
				pScan = 0;
				}

			if(pTree->ucScanRoot)							// scan root folder
			if(uEntry==pTree->uEntryCount)
				{
				uEntry++;

				   pScan = ScanFolder(pTree,pTree->uEntryCount,0,pData->hWatchdog);
				if(pScan)
					{
					pScan->uEntryPos = 1;
					pEntry = pScan->aEntries;
					uSub   = 1;
					continue;
					}
				}

			memcpy(pBuffer+iLen,cTail1,sizeof(cTail1)-1);
			iLen  += sizeof(cTail1)-1;
			uEntry = 0xFFFF;
			break;
			}


		pEntry = pTree->aEntries+uEntry;


		if(uSub && pScan)									// get entry of scan
			{
			if(pScan->uEntryPos<pScan->uEntryCount)
				{
				pEntry  = pScan->aEntries+pScan->uEntryPos;
				uNow	= pEntry->uParent-1;
				pScan->uEntryPos++;
				}
			else{
				DeallocScanList(pScan);

				pNext	= pScan->aEntries+pScan->uEntryPos-1;
				pNext	= pTree->aEntries+pNext->uParent  -1;
				iLev	= pEntry->ucLevel&0x7F;
				pScan   = 0;
				uSub    = 0;

				for(;;)										// find next scan level
					{
					if(!pNext->uParent)
						{
						break;
						}

					uNext = pNext->uParent-1;
					pNext = pTree->aEntries+uNext;

					if(iLev>(pNext->ucLevel&0x7F))
						{
						break;
						}

					if(pNext->ucType==2)
						{
						   pScan = ScanFolder(pTree,uNext,0,pData->hWatchdog);
						if(pScan)
							{
							pScan->uEntryPos = 1;
							pEntry  = pScan->aEntries;
							uNow	= uNext;
							uSub    = 1;
							break;
							}
						}
					}

				if(!uSub && uEntry>=pTree->uEntryCount)		// scan root folder
					{
					continue;
					}
				}
			}


		if(!uSub || !pScan)									// get entry of tree
			{
			if(pScan)
				{
				DeallocScanList(pScan);
				pScan = 0;
				}

			uNow	= uEntry;
			uNext	= uEntry+1;
			pNext	= pTree->aEntries+uNext;
			iLev	= pEntry->ucLevel&0x7F;
			iExp	= IsExpaned(pParam,uEntry,pTree->ucExpandAll);
			uBits  &= ~(1<<(iLev));

			if(!iExp)										// ignore if not expandet
				{
				while(iLev<(pNext->ucLevel&0x7F))
					{
					uNext += 1;
					pNext  = pTree->aEntries+uNext;
					}

				if(iLev>(pNext->ucLevel&0x7F))
					{
					iLev	= pNext->ucLevel&0x7F;
					pNext	= pEntry;

					while(pNext->uParent)
						{
						uTemp	= pNext->uParent-1;
						pNext	= pTree->aEntries+uTemp;

						if(iLev>(pNext->ucLevel&0x7F))
							{
							break;
							}

						if(pNext->ucType!=2)continue;

						   pScan = ScanFolder(pTree,uTemp,0,pData->hWatchdog);
						if(pScan)
							{
							uSub   = 1;
							break;
							}
						}
					}
				}
			else if(pEntry->ucType==2)						// scan for files
				{
				if(iLev>=(pNext->ucLevel&0x7F))
					{
					   pScan = ScanFolder(pTree,uEntry,0,pData->hWatchdog);
					if(pScan)
						{
						uSub = 1;
						if(pEntry->uParent)
							{
							uBits |= 1<<(iLev);
							}
						}
					}
				else if(!pTree->ucNoLines && ScanFolder(pTree,uEntry,1,pData->hWatchdog))
					{
					uBits |= 1<<(iLev);
					}
				}

			uEntry = uNext;
			}

		for(pNext=pEntry,iCnt=0;iCnt<8;)					// get levels
			{
			aLevels[iCnt] = pNext->ucLevel;
			iCnt++;

			if(!pNext->uParent)break;

			pNext = pTree->aEntries+pNext->uParent-1;
			}


		memcpy(pBuffer+iLen,cEntry3A,sizeof(cEntry3A)-1);
		iLen += sizeof(cEntry3A)-1;
		iLev  = pEntry->ucLevel&0x7F;

		if(pEntry->ucType==0)
			{
			iLev--;
			}


		uBits &= ~(0xFFFFFFFF<<(iLev+1));


		for(iPos=iCnt;iPos>=0;iPos--)						// create single entry
			{
			if(iPos==0)										// set text and file icon
				{
				if(pEntry->cName[pEntry->ucNameLen]=='#')
					{
					pUrlPtr = pEntry->cName+pEntry->ucNameLen+1;
					iNum    = strlen(pUrlPtr);
					}
				else if(pEntry->ucType && pTree->ucNoFolderUrl)
					{
					pUrlPtr = 0;
					iNum    = 0;
					}
				else{
					pUrlPtr = cPath;
					iNum    = GetEntryUrl(pTree,uNow,cPath,sizeof(cPath)-1);

					if(!pEntry->ucType && pScan && pScan->uEntryPos>0 && iNum+pEntry->ucNameLen<sizeof(cPath)-2)
						{
						cPath[iNum]='/';iNum++;
						memcpy(cPath+iNum,pEntry->cName,pEntry->ucNameLen);
						iNum+=pEntry->ucNameLen;
						}
					}


				if(!pUrlPtr)
					{
					memcpy(pBuffer+iLen,cEntry0F,sizeof(cEntry0F)-1);
					iLen += sizeof(cEntry0F)-1;
					}
				else{
					iLen += InsertHref(pTree,pBuffer+iLen,iSize-iLen,cEntry0A,sizeof(cEntry0A)-1,pUrlPtr,iNum);
					}

				memcpy(pBuffer+iLen,cEntry0B,sizeof(cEntry0B)-1);
				iLen += sizeof(cEntry0B)-1;
				iLen += HttpGetIconPath(hConn,pEntry->cName+pEntry->ucSuffixPos,pBuffer+iLen,iSize-iLen,(pEntry->ucType)? 1+iExp:0);

				if(!pUrlPtr)
					{
					memcpy(pBuffer+iLen,cEntry0G,sizeof(cEntry0G)-1);
					iLen += sizeof(cEntry0G)-1;
					}
				else{
					iLen += InsertHref(pTree,pBuffer+iLen,iSize-iLen,cEntry0C,sizeof(cEntry0C)-1,pUrlPtr,iNum);
					}

				memcpy(pBuffer+iLen,cEntry0D,sizeof(cEntry0D)-1);
				iLen += sizeof(cEntry0D)-1;


				if(pTree->ucNoSuffux)						// don't show suffix of filename
					{
					   iOff = pEntry->ucSuffixPos;
					if(iOff>0 && pEntry->cName[iOff-1]=='.')
						{
						iOff--;
						}
					}
				else{
					iOff = pEntry->ucNameLen;
					}

				iLen += HttpEncodeHtml(pEntry->cName,pBuffer+iLen,iSize-iLen,iOff);

				// REM AZ
				//itoa(uBits,cPath,2);
				//iLen += sprintf(pBuffer+iLen," &#160; &#160; &#160; %i%s (%u) bit=%sb %08Xh",pEntry->ucLevel&0x7F,(pEntry->ucLevel&0x80)? "->end":"",pEntry->uParent,cPath,pEntry);


				memcpy(pBuffer+iLen,cEntry0E,sizeof(cEntry0E)-1);
				iLen += sizeof(cEntry0E)-1;
				}
			else if(iPos==1)								// set plus minus icon
				{
				if(pEntry->ucType)
					{
					memcpy(pBuffer+iLen,cEntry1A,sizeof(cEntry1A)-1);
					iLen += sizeof(cEntry1A)-1;
					iLen += HttpEncodeString(pTree->cUrl  ,pBuffer+iLen,iSize-iLen,pTree->uUrlLength);
					iLen += SetParameter(pParam,uNow,!iExp,pBuffer+iLen,iSize-iLen,pTree->ucExpandAll);

					uIcon = iExp;

					if(pTree->ucNoLines)
						{
						uIcon+=8;
						}
					else if((aLevels[iPos-1]&0x80) && !((uBits>>(iCnt-iPos))&1))
						{
						uIcon+=4;
						}
					}
				else{
					memcpy(pBuffer+iLen,cEntry1B,sizeof(cEntry1B)-1);
					iLen += sizeof(cEntry1B)-1;

					if(pTree->ucNoLines)
						{
						uIcon=7;
						}
					else{
						uIcon = (aLevels[0]&0x80)? 6:2;
						}
					}

				memcpy(pBuffer+iLen,cEntry1C,sizeof(cEntry1C)-1);
				iLen += sizeof(cEntry1C)-1;
				iLen += HttpGetIconTree(hConn,uIcon,pBuffer+iLen,iSize-iLen);
				memcpy(pBuffer+iLen,cEntry1D,sizeof(cEntry1D)-1);
				iLen += sizeof(cEntry1D)-1;
				}
			else{											// set spacer icons ' ' or '|'
				if(pTree->ucNoLines)
					{
					uIcon = 7;
					}
				else if((aLevels[iPos-1]&0x80) && !((uBits>>(iCnt-iPos))&1))
					{
					uIcon = 7;
					}
				else{
					uIcon = 3;
					}

				memcpy(pBuffer+iLen,cEntry2A,sizeof(cEntry2A)-1);
				iLen += sizeof(cEntry2A)-1;
				iLen += HttpGetIconTree(hConn,uIcon,pBuffer+iLen,iSize-iLen);
				memcpy(pBuffer+iLen,cEntry2B,sizeof(cEntry2B)-1);
				iLen += sizeof(cEntry2B)-1;
				}
			}

		memcpy(pBuffer+iLen,cEntry3B,sizeof(cEntry3B)-1);
		iLen+=sizeof(cEntry3B)-1;
		}


	HttpSetUserPointer(hConn,iUserParamId+1,pScan);
	HttpSetUserValue  (hConn,iUserParamId  ,uList|(uEntry<<8)|(uBits<<24)|(uSub<<31));



return iLen;
}


//*****************************************************************************
//*
//*		HttpTreeRequest
//*
//*****************************************************************************
//	Callback for a http request
//	hConn		: handle of the http connection
//	iCommand	: is the type of the request command (HTTP_CMD_???)
//	pHost		: is the domain in the request (i.e. "www.myaddress.com")
//	pUrl		: is the url of the request (i.e. "/folder/index.html")
//	pParams		: is the parameter of the request
//	iUrlLen		: is the length of the domain
//	iParamLen	: is the length of the parameter
//	Retruns a http error (>=100) or HTTP_RESP_???/HTTP_POST_???
HTTP_API int _cdecl HttpTreeRequest(HttpConnection hConn,int iCommand,const char *pHost,const char *pUrl,const char *pParams,int iUrlLen,int iParamLen)
{
TreeList	   *pTree;
TreeScan	   *pScan;
int				iPos;




	   pScan = (TreeScan*)HttpGetUserPointer(hConn,iUserParamId+1);
	if(pScan)
		{
		HttpSetUserPointer(hConn,iUserParamId+1,0);
		DeallocScanList(pScan);
		}



	if(iCommand!=HTTP_CMD_GET )								// check http command
	if(iCommand!=HTTP_CMD_HEAD)
		{
		goto NextPlugin;
		}

	for(iPos=0;;iPos++)										// find url of a tree entry
		{
		if(iPos>=TP_COUNT)goto NextPlugin;

		pTree = aTreeArray[iPos];

		if(pTree==0)continue;

		if(iUrlLen==(int)pTree->uUrlLength)
		if(stricmp( pUrl,pTree->cUrl) == 0)
			{
			break;
			}
		}

	HttpSetUserValue(hConn,iUserParamId,iPos);
	HttpSetCacheMode(hConn,HTTP_CACHE_NOCACHE);

	return HttpTransmitCallback(hConn,HttpTreeData);


NextPlugin:


	if(pOldRequest)											// is an other plugin active
		{
		return pOldRequest(hConn,iCommand,pHost,pUrl,pParams,iUrlLen,iParamLen);
		}


return 0;
}

//*****************************************************************************
//*
//*		HttpTreeDisconnect
//*
//*****************************************************************************
//	Callback for dissconnecting
//	hConn		: handle of the http connection
//	pIpAddr		: is the ip address of the client
//	usPort		: is the tcp port of the connection
//	iReason		: is the dissconnect reason (see HTTP_MSG_???)
HTTP_API void _cdecl HttpTreeDisconnect(HttpConnection hConn,const char *pIpAddr,unsigned short usPort,int iReason)
{
TreeScan	   *pScan;


	   pScan = (TreeScan*)HttpGetUserPointer(hConn,iUserParamId+1);
	if(pScan)
		{
		HttpSetUserPointer(hConn,iUserParamId+1,0);
		DeallocScanList(pScan);
		}

	if(pOldDisconnect)
		{
		pOldDisconnect(hConn,pIpAddr,usPort,iReason);
		}

}


//*****************************************************************************
//*
//*		HttpRegisterTree
//*
//*****************************************************************************
//	Registers the tree plugin. This function must called directly
//	before HttpDaenon(...) is called.
//	pConfig		: config data of the server
//	Returns 0 if the plugin was installed or an error code
//					1 = no more user parameters
HTTP_API int HttpRegisterTree(HttpConfigData *pConfig)
{


		iUserParamId = HttpReserveUserParams(pConfig,2);
	if(!iUserParamId)return 1;
	if(!hTreeSem    )hTreeSem    =SysSemaphoreCreate();
	if(!hTreeSemNum )hTreeSemNum =SysSemaphoreCreateEx(TP_COUNT);
	if(!hTreeSemScan)hTreeSemScan=SysSemaphoreCreateEx(TP_SCANS);


	pOldRequest				= pConfig->pCbRequest;			// next callback in chain
	pConfig->pCbRequest		= HttpTreeRequest;
	pConfig->pCbDisconnect	= HttpTreeDisconnect;
 	pConfig->pCbDisconnect	= HttpTreeDisconnect;




return 0;
}


//*****************************************************************************
//*
//*		HttpTreeAddPath
//*
//*****************************************************************************
//	Inserts a tree page for a file path in a tree
//	hTree		: is the tree handle or the url for the tree
//				  page if HTTP_TREE_CREATENEW was used
//	pPath		: is the the path for the folder
//	pFilter		: is the the type filter for the files (i.e. "html;htm;txt")
//	uFlags		: several flags
//					HTTP_TREE_CREATENEW		: create new tree with url
//					HTTP_TREE_EXPANDALL		: no folder view for the root path
//					HTTP_TREE_TYPESORT		: sort file entries by type
//					HTTP_TREE_SCANROOT		: insert files from the root folter too
//					HTTP_TREE_EXPANDALL		: no folder view for the root path
//					HTTP_TREE_NOLINES		: no lines in the tree
//					HTTP_TREE_NOSUFFIX		: don't show file suffixes
//					HTTP_TREE_NOROOTURL		: don't insert href urls from the root
//					HTTP_TREE_NOTREEURL		: don't insert the url of the tree
//					HTTP_TREE_ADDFOLDER		: add the scan path as folder
//					HTTP_TREE_ADDABOVE?		: insert the scan n levels above
//	Returns 0 if the page was created or an error number
//					1 = url allready exist
//					2 = can't create tree memory
//					3 = path not exist
//					4 = url too long
//					5 = path too long
//					6 = tree contains allready a path
HTTP_API int _cdecl HttpTreeAddPath(HttpTreeHandle hTree,const char *pPath,const char *pFilter,unsigned uFlags)
{
const char     *pUrl=(const char*)hTree;
char			cPath[256];
unsigned		aSize[8];
SysFind			hFind;
SysFindStruct	sData;
TreeList	   *pTree;
TreeEntry	   *pEntry;
TreeEntry	   *pBefore;
unsigned		uParent;
unsigned		uLevel;
unsigned		uTurn;
unsigned 		uLen;
unsigned		uMax;
unsigned		uOff;
int				iPos;
int				iCnt;
int				iNum;
int				iEnd;



	if(!hTreeSem    )hTreeSem    =SysSemaphoreCreate();
	if(!hTreeSemNum )hTreeSemNum =SysSemaphoreCreateEx(TP_COUNT);
	if(!hTreeSemScan)hTreeSemScan=SysSemaphoreCreateEx(TP_SCANS);


	   uLen = strlen(pUrl);
	if(uLen>=TP_MAX_URL)return 4;

	   uMax = strlen(pPath);
	if(uMax>=TP_MAX_PATH-8)return 5;

	uMax = SysGetFullPath(pPath,cPath,sizeof(cPath));
															// add slash at path end

	if(uMax==0 || (cPath[uMax-1]!='/' && cPath[uMax-1]!='\\'))
		{
		cPath[uMax] = '/';	uMax++;
		cPath[uMax] = 0;
		}

	if(uFlags&HTTP_TREE_ADDFOLDER)							// add the scan path as folder
		{
		uMax--;
		cPath[uMax] = 0;

		while(uMax>0)
			{
			if(cPath[uMax-1]==':' )break;
			if(cPath[uMax-1]=='/' )break;
			if(cPath[uMax-1]=='\\')break;
			uMax--;
			}
		}



	if(uFlags&HTTP_TREE_CREATENEW)							// create new tree list
		{
		SysSemaphoreLock(hTreeSem);

		for(iPos=0;iPos<TP_COUNT;iPos++)					// exists the url
			{
			if(aTreeArray[iPos]            == 0  )continue;
			if(aTreeArray[iPos]->uLocked   == 0  )continue;
			if(aTreeArray[iPos]->uUrlLength!=uLen)continue;
			if(strnicmp(pUrl,aTreeArray[iPos]->cUrl,uLen))continue;

			SysSemaphoreUnlock(hTreeSem);

			return 1;
			}

		SysSemaphoreUnlock(hTreeSem);


			pTree = AllocTreeList();
		if(!pTree)return 2;

		memcpy(pTree->cUrl ,pUrl ,uLen+1);
		pTree->uUrlLength		= uLen;
		pTree->uEntryCount		= 0;
		pTree->uBodyLength		= 0;
		pTree->uTargetLength	= 0;
		pTree->uSuffixLength	= 0;
		pTree->uPrefixLength	= 0;
		pTree->ucNoFolderUrl	= (uFlags&HTTP_TREE_NOFOLDERURL)? 1:0;
		pTree->ucExpandAll		= (uFlags&HTTP_TREE_EXPANDALL  )? 1:0;
	 	pTree->ucScanRoot		= (uFlags&HTTP_TREE_SCANROOT   )? 1:0;
		pTree->ucNoSuffux		= (uFlags&HTTP_TREE_NOSUFFIX   )? 1:0;
		pTree->ucNoLines		= (uFlags&HTTP_TREE_NOLINES    )? 1:0;
		pTree->ucNoRootUrl		= (uFlags&HTTP_TREE_NOROOTURL  )? 1:0;
		pTree->ucNoTreeUrl		= (uFlags&HTTP_TREE_NOTREEURL  )? 1:0;
		uParent					= 0;
		uLevel					= 1;
		}
	else{													// add to existing tree
			pTree = (TreeList*)hTree;
		if(!pTree || !pTree->uLocked)
			{
			return 2;
			}

		if(pTree->cPath[0])
			{
			return 6;
			}

		if(pTree->uEntryCount>0)
			{
			pEntry  = pTree ->aEntries+pTree->uEntryCount-1;
			uParent = pTree ->uEntryCount;
			uLevel  =(pEntry->ucLevel&0x7F)+1;
			pBefore = 0;

			if(uFlags&HTTP_TREE_ADDABOVEX)
				{
				uTurn = uFlags&HTTP_TREE_ADDABOVEX;

				for(;uTurn>0;uTurn--)
					{
					if(!uParent)break;

					pBefore = pEntry;
					uParent = pEntry->uParent;
					pEntry  = pTree ->aEntries+uParent-1;
					uLevel  =(pEntry->ucLevel&0x7F)+1;
					}
				}

			if(pBefore)
				{
				pBefore->ucLevel &= 0x7F;
				}
			}
		else{
			uParent = 0;
			uLevel  = 1;
			}

		for(uOff=0;uOff<=uLevel;uOff++)
			{
			aSize[uOff] = uMax;
			}
		}



	pTree->pSuffixFilter		= pFilter;
	pTree->uPathLength			= uMax;
	pTree->ucSortByType			= (uFlags&HTTP_TREE_TYPESORT )? 1:0;


	memcpy(pTree->cPath,cPath,uMax);
	pTree->cPath[uMax]=0;

	if(!(uFlags&HTTP_TREE_ADDFOLDER))
		{
		aSize[0     ] = uMax;
		cPath[uMax  ] = '*';
		cPath[uMax+1] = 0;
		}



	   hFind = SysFindFirst(cPath,&sData);
	if(hFind<0)												// path not exist
		{
		if(uFlags&HTTP_TREE_CREATENEW)
			{
			pTree->cUrl[0]=0;
			DeallocTreeList(pTree);
			}

		return 3;
		}


	iCnt	= pTree->uEntryCount;
	iEnd	= TP_MAX_LIST;
	iPos	= TP_MAX_LIST-1;


	for(;;)													// insert folders
		{
		if(sData.dwAttr&SYS_DIR)
		if(sData.cName[0]!='.' || (sData.cName[1]!=0 && sData.cName[1]!='.') || (sData.cName[2]!=0 && sData.cName[1]!=0))
			{
			uLen = strlen(sData.cName);

			for(uOff=uLen;;uOff--)							// get suffix position
				{
				if(uOff<=0){uOff=uLen;break;}
				if(sData.cName[uOff-1]=='.')break;
				}

			if(!pFilter || CmpFileSuffix(sData.cName+uOff,pFilter,uLen-uOff))
				{
				pEntry	= pTree->aEntries+iPos;

				if(iPos<=iCnt)								// too much entries
					{
					pEntry++;

					memcpy(pEntry->cName,"...",4);
					pEntry->ucSuffixPos	= 3;
					pEntry->ucNameLen	= 3;
					pEntry->uParent		= uParent;
					pEntry->ucLevel		= uLevel;
					pEntry->ucType		= 1;
					}
				else{										// insert entry
					memcpy(pEntry->cName,sData.cName,uLen+1);
					pEntry->ucSuffixPos	= uOff;
					pEntry->ucNameLen	= uLen;
					pEntry->uParent		= uParent;
					pEntry->ucLevel		= uLevel;
					pEntry->ucType		= 2;

					iPos--;
					}
				}
			}


		if(iPos<=iCnt || SysFindNext(hFind,&sData)<0)		// last entry in folder
			{
			SysFindClose(hFind);

			pEntry	= pTree->aEntries+iPos+1;
			iNum	= iEnd-iPos-1;

			qsort(pEntry,iNum,sizeof(TreeEntry),CmpByName);

			for(iPos++;;pEntry++,iPos++)
				{
				if(iPos>=TP_MAX_LIST)goto Exit;

				uLevel = pEntry->ucLevel;

				if(iPos==TP_MAX_LIST-1 || uLevel!=pEntry[1].ucLevel)
					{
					pEntry[0].ucLevel |= 0x80;
					}

				memcpy(pTree->aEntries+iCnt,pEntry,sizeof(TreeEntry));
				iCnt++;

				if(pEntry->ucType!=2)						// find next level
					{
					continue;
					}


				uMax			= aSize[uLevel-1];
				memcpy(cPath+uMax,pEntry->cName,pEntry->ucNameLen);
				uMax           += pEntry->ucNameLen;
				cPath[uMax  ]	= '/'; uMax++;
				cPath[uMax  ]	= '*';
				cPath[uMax+1]	=  0;

				   hFind = SysFindFirst(cPath,&sData);
				if(hFind>= 0 && uLevel<8)
					{
					aSize[uLevel]=uMax;
					uParent=iCnt;
					iEnd=iPos+1;
					uLevel++;
					break;
					}
				}
			}
		}




Exit:

	pTree->uEntryCount	= iCnt;
	pTree->aEntries[iCnt].ucLevel		= 0x80;
	pTree->aEntries[iCnt].ucType		= 1;
	pTree->aEntries[iCnt].ucNameLen		= 0;
	pTree->aEntries[iCnt].ucSuffixPos	= 0;
	pTree->aEntries[iCnt].cName[0]		= 0;



return 0;
}

//*****************************************************************************
//*
//*		HttpTreeCreate
//*
//*****************************************************************************
//	Create a new tree page
//	pUrl		: is the url for the page
//	pHandle		: here the tree handel will be stored
//	uFlags		: several flags
//					HTTP_TREE_EXPANDALL		: no folder view for the root path
//					HTTP_TREE_TYPESORT		: sort file entries by type
//					HTTP_TREE_SCANROOT		: insert files from the root folter too
//					HTTP_TREE_EXPANDALL		: no folder view for the root path
//					HTTP_TREE_NOLINES		: no lines in the tree
//					HTTP_TREE_NOSUFFIX		: don't show file suffixes
//					HTTP_TREE_NOFOLDERURL	: don't insert href urls for folders
//					HTTP_TREE_NOROOTURL		: don't insert href urls from the root
//					HTTP_TREE_NOTREEURL		: don't insert the url of the tree
//	Returns 0 if the page was created or an error number
//					1 = url allready exist
//					2 = can't create tree memory
//					3 = path not exist
//					4 = url too long
//					5 = path too long
//					6 = tree contains allready a path
HTTP_API int _cdecl HttpTreeCreate(const char *pUrl,HttpTreeHandle *pHandle,unsigned uFlags)
{
TreeList	   *pTree;
unsigned 		uLen;
int				iPos;



	*pHandle = 0;

	if(!hTreeSem    )hTreeSem    =SysSemaphoreCreate();
	if(!hTreeSemNum )hTreeSemNum =SysSemaphoreCreateEx(TP_COUNT);
	if(!hTreeSemScan)hTreeSemScan=SysSemaphoreCreateEx(TP_SCANS);


	   uLen = strlen(pUrl);
	if(uLen>=TP_MAX_URL)return 4;

	SysSemaphoreLock(hTreeSem);

	for(iPos=0;iPos<TP_COUNT;iPos++)						// exists the url
		{
		if(aTreeArray[iPos]            == 0  )continue;
		if(aTreeArray[iPos]->uLocked   == 0  )continue;
		if(aTreeArray[iPos]->uUrlLength!=uLen)continue;
		if(strnicmp(pUrl,aTreeArray[iPos]->cUrl,uLen))continue;

		SysSemaphoreUnlock(hTreeSem);

		return 1;
		}

	SysSemaphoreUnlock(hTreeSem);


		pTree = AllocTreeList();
	if(!pTree)return 2;

	memcpy(pTree->cUrl ,pUrl ,uLen+1);

	pTree->uUrlLength		= uLen;
	pTree->pSuffixFilter	= 0;
	pTree->uPathLength		= 0;
	pTree->uBodyLength		= 0;
	pTree->uEntryCount		= 0;
	pTree->uTargetLength	= 0;
	pTree->uSuffixLength	= 0;
	pTree->uPrefixLength	= 0;
	pTree->cPath[0]			= 0;
 	pTree->ucScanRoot		= 0;
	pTree->ucNoFolderUrl	= (uFlags&HTTP_TREE_NOFOLDERURL)? 1:0;
	pTree->ucExpandAll		= (uFlags&HTTP_TREE_EXPANDALL  )? 1:0;
	pTree->ucSortByType		= (uFlags&HTTP_TREE_TYPESORT   )? 1:0;
	pTree->ucNoSuffux		= (uFlags&HTTP_TREE_NOSUFFIX   )? 1:0;
	pTree->ucNoLines		= (uFlags&HTTP_TREE_NOLINES    )? 1:0;
	pTree->ucNoRootUrl		= (uFlags&HTTP_TREE_NOROOTURL  )? 1:0;
	pTree->ucNoTreeUrl		= (uFlags&HTTP_TREE_NOTREEURL  )? 1:0;



	pTree->aEntries[0].ucLevel		= 0x80;
	pTree->aEntries[0].ucType		= 1;
	pTree->aEntries[0].ucNameLen	= 0;
	pTree->aEntries[0].ucSuffixPos	= 0;
	pTree->aEntries[0].cName[0]		= 0;

	*pHandle = (HttpTreeHandle)pTree;



return 0;
}


//*****************************************************************************
//*
//*		HttpTreeAddEntry
//*
//*****************************************************************************
//	Adds an entry to an existing tree
//	hTree			: handel of the tree
//	pName			: is the name of the entry , it is possible to add an url to
//					  the name. (i.e. "test.html" or "test.html#/my_test.html")
//	iMode			: is the insert mode
//						HTTP_TREE_ENTRY	= insert common entry on the current level
//						+1...0...-7		= insert a folder in on the delta level
//	pUrl			: is the url for the entry. if pUrl is zero the url will be
//					  taken from pName.
//	Returns 0 if the entry was inserted or an error number
//					2 = tree not exist
//					4 = url too long
//					5 = name too long
//					7 = wrong mode
//					8 = tree data is full
int _cdecl HttpTreeAddEntry(HttpTreeHandle hTree,const char *pName,int iMode,const char *pUrl)
{
TreeList	   *pTree=(TreeList*)hTree;
TreeEntry	   *pEntry;
TreeEntry	   *pBefore;
unsigned		uParent;
unsigned		uLevel;
unsigned		uHigh;
unsigned		uType;
unsigned 		uLen;
unsigned		uOff;
unsigned		uUrl;
unsigned		uEnd;
unsigned		uAdd;



	if(!pTree || !pTree->uLocked)
		{
		return 2;
		}

	if(pTree->uEntryCount>=TP_MAX_LIST)						// no more tree slots
		{
		return 8;
		}


	for(uLen=0,uUrl=0;;uLen++)								// get length and url
		{
		if(!pName[uLen])
			{
			uEnd=uLen;
			break;
			}

		if(uLen>=TP_MAX_NAME)
			{
			return 5;
			}

		if(pName[uLen]=='#')
			{
			   uUrl = uLen+1;
			   uEnd = strlen(pName+uUrl)+uUrl;
			if(uEnd>=TP_MAX_NAME-1)
				{
				return 5;
				}

			break;
			}
		}


	if(pUrl && !uUrl && *pUrl)								// is url paramater valid
		{
		uAdd = strlen(pUrl);
		uUrl = uLen+1;

		if(uAdd+uLen>=TP_MAX_NAME-1)
			{
			return 4;
			}
		}
	else{													// use name as url
		pUrl = pName+uUrl;
		uAdd = uEnd -uUrl;
		}

	for(uOff=uAdd;;uOff--)									// get suffix position
		{
		if(uOff<=0){uOff=uAdd;break;}
		if(pUrl[uOff-1] =='.')break;
		}


	if(pTree->uEntryCount>0)
		{
		pEntry  = pTree->aEntries+pTree->uEntryCount;
		pBefore	= pTree->aEntries+pTree->uEntryCount-1;
		uLevel  = pBefore->ucLevel&0x7F;
		}
	else{
		uLevel  = 1;
		pEntry  = pTree->aEntries;
		pBefore	= pEntry+4;
		}



	if(iMode<=HTTP_TREE_ENTRY+8 && iMode>=HTTP_TREE_ENTRY-8)// insert a common entry
		{
		iMode -= HTTP_TREE_ENTRY;
		uType  = 0;

		if(iMode>0)
		if(pTree->uEntryCount==0 || !pBefore->ucType)
			{
			return 7;
			}
		}
	else{													// insert a folder
		uType  = 3;
		}

	   uHigh = uLevel+iMode;
	if(uHigh<=0 || uHigh>=7 || iMode>1)
		{
		return 7;
		}



	if(uHigh==uLevel)										// insert on same level
		{
		if(pTree->uEntryCount>0)
			{
			pBefore->ucLevel &= 0x7F;
			uParent = pBefore->uParent;
			}
		else{
			uParent = 0;
			uLevel  = 1;
			}
		}
	else if(uHigh>uLevel)									// insert as child
		{
		uParent = pTree->uEntryCount;
		}
	else{													// insert above
		uParent = pBefore->uParent;

		while(uParent)
			{
			pBefore = pTree->aEntries+uParent-1;
			uLevel  = pBefore->ucLevel&0x7F;
			uParent	= pBefore->uParent;

			if(uLevel<=uHigh)
				{
				pBefore->ucLevel &= 0x7F;
				break;
				}
			}
		}


	pEntry->ucType		= uType;
	pEntry->uParent		= uParent;
	pEntry->ucLevel		= uHigh|0x80;
	pEntry->ucNameLen	= uLen;
	pEntry->ucSuffixPos	= uUrl+uOff;

	memcpy(pEntry->cName,pName,uLen+1);

	if(uUrl)												// add url to name
		{
		pEntry->cName[uUrl-1]='#';
		memcpy(pEntry->cName+uUrl,pUrl,uAdd+1);
		}

	pTree->uEntryCount++;

	pEntry++;
 	pEntry->ucLevel		= 0x80;
	pEntry->ucType		= 1;
	pEntry->ucNameLen	= 0;
	pEntry->ucSuffixPos	= 0;
	pEntry->cName[0]	= 0;



return 0;
}

//*****************************************************************************
//*
//*		HttpTreeBackground
//*
//*****************************************************************************
//	Sets the backgrond off the tree page
//	pTree			: handel of the tree
//	uColor			: is the backgrond color  (0x00RRGGBB, or 0xFFFFFFFF for default)
//	pFile			: is the url of the picture for the background
//						0 means no background
//	Returns 0 if the entry was inserted or an error number
//					2 = tree not exist
//					5 = name too long
//					8 = tree data is full
HTTP_API int _cdecl HttpTreeBackground(HttpTreeHandle hTree,unsigned uColor,const char *pFile)
{
TreeList	   *pTree=(TreeList*)hTree;




	if(!pTree || !pTree->uLocked)
		{
		return 2;
		}

	if(pTree->uEntryCount>=TP_MAX_LIST)
		{
		return 8;
		}


	if(uColor!=0xFFFFFFFF)
		{
		pTree->uBodyLength = sprintf(pTree->cBody," bgcolor=\"#%06X\"",uColor);
		}
	else{
		pTree->uBodyLength = 0;
		}


	if(pFile && *pFile)
		{
		pTree->uBodyLength += sprintf(pTree->cBody+pTree->uBodyLength," background=");
		pTree->uBodyLength += HttpEncodeString(pFile,pTree->cBody+pTree->uBodyLength,sizeof(pTree->cBody)-pTree->uBodyLength-2);
		}


return 0;
}



//*****************************************************************************
//*
//*		HttpTreeSetUrlFrame
//*
//*****************************************************************************
//	Sets the suffix and prefix for the url's
//			(...href=<prefix>abc<suffix>...target=xyz..)
//	pTree			: handel of the tree
//	pPrefix			: is the prefix for the url
//	pSuffix			: is the suffix for the url
//	Returns 0 if the entries will be inserted or an error number
//					2 = tree not exist
//					5 = name too long
HTTP_API int _cdecl HttpTreeSetUrlFrame(HttpTreeHandle  hTree,const char *pPrefix,const char *pSuffix)
{
TreeList	   *pTree=(TreeList*)hTree;
unsigned		uSuffix;
unsigned		uPrefix;



	if(!pTree || !pTree->uLocked)
		{
		return 2;
		}

	uSuffix = (pSuffix)? strlen(pSuffix):0;
	uPrefix = (pPrefix)? strlen(pPrefix):0;

	if(uSuffix>TP_MAX_SUFFIX)return 5;
	if(uPrefix>TP_MAX_PREFIX)return 5;

	memcpy(pTree->cSuffix,pSuffix,uSuffix);
	memcpy(pTree->cPrefix,pPrefix,uPrefix);

	pTree->uSuffixLength = uSuffix;
	pTree->uPrefixLength = uPrefix;


return 0;
}

